!***********************************************************************
  module body_force
!***********************************************************************
! Module providing functionality for external body forces.
!23456789012345678901234567890123456789012345678901234567890123456789012

  use simulation_box

  implicit none

  private

! Public procedures
  public :: calc_body_force, init_body_force, destroy_body_force, &
          & rdinput_body_force, wr_body_force

! Public data types
  public :: body_force_data_type

  type body_force_data_type
   double precision, dimension(:,:), allocatable :: bf  ! body force
   double precision, dimension(:), allocatable :: sbf   ! specific body force
   double precision :: wvec  ! wavevector for spatially sinusoidal body force
   double precision :: omega ! angular frequency for temporally sinusoidal body force
   integer :: bforcetype
   integer :: mdnumber ! mode number used to calculate wavevector
   integer :: psteps ! number of timesteps in period
  end type body_force_data_type

! Note: specific body force means force/mass, i.e. acceleration
! Values of constants
! bforcetype = 0 none
!            = 1 constant gravitational acceleration field
!            = 2 STF - sinusoidal transverse force
!            = 3 SLF - sinusoidal longitudinal force
  contains
!***********************************************************************
  subroutine calc_body_force(bfd, mass, ry, nsteps)
! Calculate body force - may be position and time dependent.
!***********************************************************************
  implicit none

  type(body_force_data_type) :: bfd
  double precision, dimension(:) :: mass
  double precision, dimension(:) :: ry
  double precision :: amp
  integer :: i, nsteps

  bfd%bf = 0.d0

  if(bfd%psteps .ne. 0) then
   amp = sin(bfd%omega*nsteps)
  else
   amp = 1.d0
  endif

  select case (bfd%bforcetype)
  case(1)
   do i = 1, size(bfd%bf,2)
    bfd%bf(:,i) = amp*mass*bfd%sbf(i)
   enddo

  case(2)
   bfd%bf(:,1) = amp*mass*bfd%sbf(1)*sin(bfd%wvec*ry)
 
  case(3)
   bfd%bf(:,2) = amp*mass*bfd%sbf(2)*sin(bfd%wvec*ry)

  case default
   stop 'calc_body_force : invalid body force type'

  end select

  end subroutine calc_body_force
!***********************************************************************
  subroutine destroy_body_force(bfd)
!***********************************************************************
  implicit none

  type(body_force_data_type) :: bfd

  if(allocated(bfd%sbf)) deallocate(bfd%sbf)
  if(allocated(bfd%bf)) deallocate(bfd%bf)

  end subroutine destroy_body_force
!***********************************************************************
  subroutine init_body_force(bfd, np, box)
! Initialize all body force variables.
!***********************************************************************
  implicit none

  type(body_force_data_type) :: bfd
  type(box_data_type) :: box
  double precision :: pi
  integer :: np, error

   pi = 2.d0*acos(0.d0)

  allocate(bfd%bf(np,box%ndim), stat=error)
  if(error .ne. 0) stop 'body_force.init_body_force: allocation error'
  bfd%bf = 0.d0

  if(bfd%bforcetype .eq. 2 .or. bfd%bforcetype .eq. 3) then
   bfd%wvec = 2.d0*pi*bfd%mdnumber/box%boxvec(2,2)
  else
   bfd%wvec = 0.d0
  endif
  
  if(bfd%psteps .ne. 0) then
   bfd%omega = 2.d0*pi/bfd%psteps
  else
   bfd%omega = 0.d0
  endif

  end subroutine init_body_force
!***********************************************************************
  subroutine rdinput_body_force(bfd, dim, iunit)
!***********************************************************************
  implicit none

  type (body_force_data_type) :: bfd
  integer :: iunit, i, dim, error
    
  read(iunit,*) bfd%bforcetype
  allocate(bfd%sbf(dim), stat=error)
  if(error .ne. 0) stop 'body_force.rdinput_body_force: allocation error'
  read(iunit,*) (bfd%sbf(i), i = 1, size(bfd%sbf))
  read(iunit,*) bfd%mdnumber
  read(iunit,*) bfd%psteps

  end subroutine rdinput_body_force
!***********************************************************************
  subroutine wr_body_force(bfd, ounit)
! writes input data to the output file
!***********************************************************************
  implicit none

  type (body_force_data_type), intent(in) :: bfd
  integer :: i, ounit

  write (ounit,*) 'Body force type                  = ', bfd%bforcetype
  write (ounit,*) 'Specific body force              = ', &
                & (bfd%sbf(i), i = 1, size(bfd%sbf))
  
  end subroutine wr_body_force
!***********************************************************************
  end module body_force
