  module bond_bend
!***********************************************************************
! Calculates the bond angle bending force and energy.
!
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************
  use atoms
  use molecules
  use species
  use neighbour_list

  implicit none

  private

! Public procedures
  public :: add_bondbend_forces, destroy_bondbend, init_imask_bondbend, &
           & rd_bondbend_input, wr_bondbend

! Public data types
  public :: bondbend_data_type, bondbend_set

  type :: bondbend_data_type
   integer :: na ! number of bond angles
   integer, dimension(:,:), allocatable :: nalst ! bond bend interaction list
  end type bondbend_data_type

  type :: bondbend_set
   type(bondbend_data_type), dimension(:), allocatable :: bbset
   double precision, dimension(:), allocatable :: bc ! bond potential constants
   integer :: bbtype ! = 0 for no bond bend forces
                     ! = 1 for Hookean theta
                     ! = 2 for Hookean cos(theta)
    integer :: bbarch ! = 1 linear chain
                      ! = 2 for arbitrary architecture, nalst is in a file
  end type bondbend_set

  contains
!***********************************************************************
  subroutine add_bondbend_forces(atd,bb,md,ubb,ava,pta)
! Calculate bond bending forces and energies.
!***********************************************************************
  implicit none

  type(atom_type) :: atd
  type (bondbend_set) :: bb
  type(mol_set) :: md

  double precision :: ubb, ava
  double precision, dimension(size(atd%r,2),size(atd%r,2)) :: pta
  double precision, dimension(size(atd%r,1),size(atd%r,2)) :: fbi
  integer :: ndim, i, isf, isl, np, ia

  ndim = size(atd%r,2)

  ubb = 0.0   ! initialize total bond bend energy
  ia = 0      ! initialize counter for average bond angle calculation
  ava = 0.0   ! initialize accumulator for average bond angle

! Calculate bond bend forces for each fraction of molecules in sequence.

  isl = 0

  do i = 1, md%nf

   np = md%nsf(i)*md%nmf(i)   ! total number of sites in this fraction
   isf = isl + 1              ! first site in next fraction
   isl = isf - 1 + np         ! last site in this fraction

   if(bb%bbset(i)%na .eq. 0) cycle ! if no bonds in this fraction, skip

   fbi = 0.d0

   call bondbend_force_energy(atd%r(isf:isl,:), md%nsf(i), md%nmf(i), &
    & bb%bbset(i)%nalst, bb%bc, bb%bbtype, fbi(isf:isl,:), &
    & atd%eint(isf:isl), ubb, ava, ia, pta)

   atd%f(isf:isl,:) = atd%f(isf:isl,:) + fbi(isf:isl,:)

  enddo
  
  ava = ava/ia

  end subroutine add_bondbend_forces
!***********************************************************************
  subroutine bondbend_force_energy(ra,ns,nmf,nalst,bc,bbtype,fi,ei,ubb,ava,ia,pta)
!***********************************************************************
  implicit none

  double precision, dimension(:,:), intent(inout) :: ra, fi
  double precision, dimension(:), intent(inout) :: ei
  double precision, dimension(:), intent(in) :: bc
  double precision, dimension(:,:), intent(inout) :: pta
  double precision, dimension(size(ra,2)) :: r12, r23, ur12, ur23, &
  & cr1, cr2, cr3, cr4
  double precision :: dsq12, dsq23, r12dot23, ua=0.d0, ubb, &
  & ava, ctheta, theta, d12, d23, tb1, fb1
  integer, dimension(:,:), intent(in) :: nalst
  integer, intent(in) :: ns, nmf, bbtype
  integer :: i, ia, im, j, k, ism, i1, i2, i3, ndim

  ndim = size(ra,2)

  ism = -ns ! initialize site offset so that it starts at zero

! Calculate bond bending forces

  do im = 1, nmf    ! loop over all molecules in this fraction

   ism = ism + ns   ! calculate site offset for molecule im, start at zero

   do k = 1, size(nalst,1)  ! loop over all bonds of each molecule

    i1 = nalst(k,1) + ism
    i2 = nalst(k,2) + ism
    i3 = nalst(k,3) + ism

    dsq12 = 0.d0
    dsq23 = 0.d0
    r12dot23 = 0.d0

    do i = 1, ndim
     r12(i) = ra(i2,i) - ra(i1,i)
     r23(i) = ra(i3,i) - ra(i2,i)
     dsq12 = dsq12 + r12(i)*r12(i)
     dsq23 = dsq23 + r23(i)*r23(i)
     r12dot23 = r12dot23 + r12(i)*r23(i)
    enddo

! bond lengths
    d12 = sqrt(dsq12)
    d23 = sqrt(dsq23)

! calculate cos(bond angle) and bond angle
    ctheta = -r12dot23/(d12*d23)
    theta = acos(ctheta)
! accumulate sum and number of bond angles to use in average
    ava = ava + theta
    ia = ia + 1

    do i = 1, ndim 
     ur12(i) = r12(i)/d12
     ur23(i) = r23(i)/d23
     cr1(i)  = ur23(i)/d12
     cr2(i)  = ur12(i)*ctheta/d12
     cr3(i)  = ur12(i)/d23
     cr4(i)  = ur23(i)*ctheta/d23
    enddo

    select case (bbtype)

     case (1)                           ! first case theta potential
      tb1 = theta - bc(1)
      ua = 0.5d0*bc(2)*tb1**2
      fb1 = bc(2)*tb1/sin(theta)
      do i = 1, ndim
       fi(i1,i) = fi(i1,i) + fb1*(cr1(i) + cr2(i))
       fi(i2,i) = fi(i2,i) - fb1*(cr1(i) + cr2(i) - cr3(i) - cr4(i))
       fi(i3,i) = fi(i3,i) - fb1*(cr3(i) + cr4(i))
      enddo

     case (2)                            ! second case cos(theta) potential
      tb1 = ctheta - bc(1)
      ua = 0.5d0*bc(2)*tb1**2
      fb1 = bc(2)*tb1
      do i = 1, ndim
       fi(i1,i) = fi(i1,i) - fb1*(cr1(i) + cr2(i))
       fi(i2,i) = fi(i2,i) + fb1*(cr1(i) + cr2(i) - cr3(i) - cr4(i))
       fi(i3,i) = fi(i3,i) + fb1*(cr3(i) + cr4(i)) 
      enddo

    end select

    ei(i1) = ua/3.0
    ei(i2) = ua/3.0
    ei(i3) = ua/3.0

    do j = 1, ndim
     do i = 1, ndim
      pta(j,i) = pta(j,i) - ra(i1,j)*fb1*(cr1(i) + cr2(i))
      pta(j,i) = pta(j,i) + ra(i2,j)*fb1*(cr1(i) + cr2(i) - cr3(i) - cr4(i))
      pta(j,i) = pta(j,i) + ra(i3,j)*fb1*(cr3(i) + cr4(i)) 
     enddo
    enddo

    ubb = ubb + ua

   enddo
  enddo

  end subroutine bondbend_force_energy
!***********************************************************************
  subroutine init_imask_bondbend(bb, md)
! Bond angle bend interacting sites do not interact via a LJ potential.
! Set the interaction mask to false for pairs of sites that interact
! via a bond angle bending potential.
!***********************************************************************
  implicit none

  type (bondbend_set) :: bb
  type(mol_set), intent(inout) :: md
  integer :: j, k, ii, ij, ik

! new for fraction masking
  do j = 1, md%nf
   do k = 1, bb%bbset(j)%na
    ii = bb%bbset(j)%nalst(k,1)
    ij = bb%bbset(j)%nalst(k,2)
    ik = bb%bbset(j)%nalst(k,3)
    
    md%frac_mask(j)%fmask(ii,ij) = .false.
    md%frac_mask(j)%fmask(ij,ii) = .false.

    md%frac_mask(j)%fmask(ij,ik) = .false.
    md%frac_mask(j)%fmask(ik,ij) = .false.

    md%frac_mask(j)%fmask(ii,ik) = .false.
    md%frac_mask(j)%fmask(ik,ii) = .false.
   enddo
  enddo

  end subroutine init_imask_bondbend
!***********************************************************************
  subroutine create_bondbend_set(bb, nf)
!***********************************************************************
  implicit none

  type (bondbend_set), intent(inout) :: bb
  integer, intent(in) :: nf
  integer :: error=0

  allocate(bb%bbset(nf),stat = error)
  if(error .ne. 0) stop 'bond_bend.create_bondbend: allocation error'

  select case (bb%bbtype)
  case (1)  ! theta
   allocate(bb%bc(2), stat = error)
   if(error .ne. 0) stop 'bond_bend.create_bondbend: allocation error'
  case (2)  ! cos(theta)
   allocate(bb%bc(2), stat = error)
   if(error .ne. 0) stop 'bond_bend.create_bondbend: allocation error'
  case default
   stop 'bond_bend:create_bondbend_set: invalid bond bend type'
  end select

  end subroutine create_bondbend_set
!***********************************************************************
  subroutine create_bondbend(bb, md)
!***********************************************************************
  implicit none

  type (bondbend_set), intent(inout) :: bb
  type(mol_set), intent(in) :: md
  integer :: i, error, na

   do i = 1, md%nf
    na = bb%bbset(i)%na
    if(na .eq. 0) return
    allocate(bb%bbset(i)%nalst(na,3), stat = error)
    if(error .ne. 0) stop 'bond_bend:create_bondbend: allocation error'
   enddo

  end subroutine create_bondbend
!***********************************************************************
  subroutine destroy_bondbend(bb)
!***********************************************************************
  implicit none

  type (bondbend_set), intent(inout) :: bb
  integer :: i

  do i = 1, size(bb%bbset)
   if(allocated(bb%bbset(i)%nalst)) deallocate(bb%bbset(i)%nalst)
  enddo

  if(allocated(bb%bbset)) deallocate(bb%bbset)
  if(allocated(bb%bc)) deallocate(bb%bc)

  end subroutine destroy_bondbend
!***********************************************************************
  subroutine rd_bondbend_input(bb,md,header,ifile,iunit)
!***********************************************************************
! Reads bond bend potential variables from input file.
! 
!***********************************************************************
  implicit none

  type (bondbend_set), intent(inout) :: bb
  type(mol_set), intent(in) :: md
  
  integer :: i, j, k, i1, iunit
  
  character(len=*) :: header
  character(len=32) :: rdstring
  character(len=32) :: ifile        ! input file names

  rdstring = ''
  i = 0
  open(iunit,file=ifile,status='old',form='formatted', &
     & position='rewind')
  do j = 1, 1000
   read(iunit,*, iostat=i) rdstring
   if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 999) exit
  enddo

  if (rdstring .eq. header) then
   read(iunit,*) bb%bbtype           ! read the bond bend type
   if(bb%bbtype .eq. 0) then
    close(iunit)
    return                           ! no bond bend forces - return
   endif
   read(iunit,*) bb%bbarch           ! read the bond bend architecture
   call create_bondbend_set(bb, md%nf) 
   read(iunit,*) (bb%bbset(i)%na, i=1, md%nf) ! number of bonds
   read(iunit,*) (bb%bc(i), i=1, size(bb%bc)) ! bond bend constants

   call create_bondbend(bb, md)

   do k = 1, md%nf
    select case(bb%bbarch)
    case(1)            ! linear chains
     call init_bondbend_linear(bb, md)
    case(2)            ! arbitrary molecular architecture
     do j = 1, bb%bbset(k)%na
      read(iunit,*) (bb%bbset(k)%nalst(j,i1), i1 = 1, 3)
     enddo
    case default
     stop 'bond_bend:rd_bondbend_input: Invalid molecular architecture'
    end select
   enddo

   close(iunit)
  else
   close(iunit)
   stop 'bond_bend:rd_bondbend_input: Bond bend input data not found'
  endif

  end subroutine rd_bondbend_input
!***********************************************************************
  subroutine init_bondbend_linear(bb, md)
!***********************************************************************
  implicit none
 
  type (bondbend_set), intent(inout) :: bb
  type(mol_set), intent(in) :: md
  integer :: i, j

! Simple bond bend matrix initialization for linear chain molecules

  do i = 1, md%nf
   if(bb%bbset(i)%na .ne. md%nsf(i)-2) &
   & stop 'bond_bend:init_bondbend_linear: error in number of bonds'
   bb%bbset(i)%nalst = 0
   do j = 1, md%nsf(i)-2
    bb%bbset(i)%nalst(j,1) = j
    bb%bbset(i)%nalst(j,2) = j+1
    bb%bbset(i)%nalst(j,3) = j+2
   enddo
  enddo

  end subroutine init_bondbend_linear

!***********************************************************************
  subroutine wr_bondbend(bb,ofile,ounit)
! writes input data to the output file
!***********************************************************************
  type (bondbend_set), intent(in) :: bb
  integer :: ounit
  character(len=32) :: ofile      !output file name

  open(ounit,file=ofile,status='old',form='formatted',position='append')
  write (ounit,*) 'Initial bond bend force values'
  write (ounit,*)
  write (ounit,*) 'bond bend force type       = ', bb%bbtype
  write (ounit,*) 'bond bend architecture     = ', bb%bbarch
  write (ounit,*)
  close(ounit)
  
  end subroutine wr_bondbend
!***********************************************************************
  end module bond_bend
