  module bond_stretch
!***********************************************************************
! Calculates the bond length, stretching force and energy.
!
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************
  use atoms
  use molecules
  use species
  use neighbour_list
!  use constraints

  implicit none

  private

! Public procedures
  public :: add_bondstr_forces, destroy_bondstr, init_imask_bondstr, &
           & rd_bondstr_input, wr_bondstr

! Public data types
  public :: bondstr_data_type, bondstr_set

  type :: bondstr_data_type
   integer :: nb ! number of bonds
   integer, dimension(:,:), allocatable :: nblst ! bond stretch interaction list
  end type bondstr_data_type

  type :: bondstr_set
   type(bondstr_data_type), dimension(:), allocatable :: bset
   double precision, dimension(:), allocatable :: bc ! bond potential constants
   integer :: bstype ! = 0 for no bond stretch forces
                     ! = 1 for Hookean
                     ! = 2 for FENE
   integer :: bsarch ! = 1 linear chain
                     ! = 2 for arbitrary architecture, nblst is in a file
  end type bondstr_set

  contains
!***********************************************************************
  subroutine add_bondstr_forces(atd,bs,md,ubs,avb,pta)
! Calculate bond stretching forces and energies.
!***********************************************************************
  implicit none

  type(atom_type) :: atd
  type (bondstr_set) :: bs
  type(mol_set) :: md

  double precision :: ubs, avb
  double precision, dimension(size(atd%r,2),size(atd%r,2)) :: pta
  double precision, dimension(size(atd%r,1),size(atd%r,2)) :: fbi
  integer :: ndim, i, isf, isl, np, ib

  ndim = size(atd%r,2)

  ubs = 0.d0   ! initialize total bond stretch energy
  ib = 0      ! initialize counter for average bond length calculation
  avb = 0.d0   ! initialize accumulator for average bond length

! Calculate bond stretch forces for each fraction of molecules in sequence.

  isl = 0

  do i = 1, md%nf

   np = md%nsf(i)*md%nmf(i)   ! total number of sites in this fraction
   isf = isl + 1              ! first site in next fraction
   isl = isf - 1 + np         ! last site in this fraction

   if(bs%bset(i)%nb .eq. 0) cycle ! if no bonds in this fraction, skip

   fbi = 0.d0

   call bondstr_force_energy(atd%r(isf:isl,:), md%nsf(i), md%nmf(i), &
    & bs%bset(i)%nblst, bs%bc, bs%bstype, fbi(isf:isl,:), &
    & atd%eint(isf:isl), ubs, avb, ib, pta)

   atd%f(isf:isl,:) = atd%f(isf:isl,:) + fbi(isf:isl,:)

  enddo
  
  avb = avb/ib

  end subroutine add_bondstr_forces
!***********************************************************************
  subroutine bondstr_force_energy(ra,ns,nmf,nblst,bc,bstype,fi,ei,ubs,avb,ib,pta)
!***********************************************************************
  implicit none

  double precision, dimension(:,:), intent(inout) :: ra, fi
  double precision, dimension(:), intent(in) :: bc
  double precision, dimension(:), intent(inout) :: ei
  double precision, dimension(:,:), intent(inout) :: pta
  double precision, dimension(size(ra,2)) :: r12, fb
  double precision :: dsq, db=0.d0, ui=0.d0, ubs, xb, avb, bl
  integer, dimension(:,:), intent(in) :: nblst

  integer, intent(in) :: ns, nmf, bstype
  integer :: i, im, j, k, ism, i1, i2, ndim, ib

  ndim = size(ra,2)

  ism = -ns ! initialize site offset so that it starts at zero

! Calculate bond stretch forces

  do im = 1, nmf    ! loop over all molecules in this fraction

   ism = ism + ns   ! calculate site offset for molecule im, start at zero

   do k = 1, size(nblst,1)  ! loop over all bonds of each molecule

    i1 = nblst(k,1) + ism
    i2 = nblst(k,2) + ism

    dsq = 0.d0
    do i = 1, ndim
     r12(i) = ra(i2,i) - ra(i1,i)
     dsq = dsq + r12(i)*r12(i)
    enddo

    bl = sqrt(dsq)  ! bond length
    avb = avb + bl  ! average bond length
    ib = ib + 1

    select case (bstype)
    case (1)                           ! first case Hookean springs
    xb = bl - bc(1)
    ui = 0.5d0*bc(2)*xb**2             ! potential energy
    db = bc(2)*xb/dsq

    case (2)                            ! second case FENE springs
    xb = 1.d0 - dsq/bc(1)**2
    if(xb .le. 0.d0) then
     ui = 0.d0
    else
     ui = -0.5d0*bc(2)*bc(1)**2*log(xb)   ! potential energy
    endif
    db = bc(2)/xb
    end select

! bond stretch forces - fb is F12 = force on 1 due to 2
    do i = 1, ndim
     fb(i) = db*r12(i)
     fi(i1,i) = fi(i1,i) + fb(i)
     fi(i2,i) = fi(i2,i) - fb(i)
    enddo

    ei(i1) = ui*0.5
    ei(i2) = ui*0.5

    do j = 1, ndim
     do i = 1, ndim
      pta(j,i) = pta(j,i) + ra(i1,j)*fb(i)
      pta(j,i) = pta(j,i) - ra(i2,j)*fb(i)
     enddo
    enddo

    ubs = ubs + ui

   enddo
  enddo

  end subroutine bondstr_force_energy
!***********************************************************************
  subroutine init_imask_bondstr(bs, md)
! Dihedral interacting sites do not interact via a LJ potential.
! Set the interaction mask to false for pairs of sites that interact
! via a dihedral potential.
! Assume that sites connected by a bond or a bond angle constraint are
! already masked out. This procedure only masks out LJ interactions between
! sites 1 and 4 for each dihedral interaction.
!***********************************************************************
  implicit none

  type (bondstr_set) :: bs
  type(mol_set), intent(inout) :: md
  integer :: j, k, ii, ij

! new for fraction masking
  do j = 1, md%nf
   do k = 1, bs%bset(j)%nb
    ii = bs%bset(j)%nblst(k,1)
    ij = bs%bset(j)%nblst(k,2)
    md%frac_mask(j)%fmask(ii,ij) = .false.
    md%frac_mask(j)%fmask(ij,ii) = .false.
   enddo
  enddo

  end subroutine init_imask_bondstr
!***********************************************************************
  subroutine create_bondstr_set(bs, nf)
!***********************************************************************
  implicit none

  type (bondstr_set), intent(inout) :: bs
  integer, intent(in) :: nf
  integer :: error=0

  allocate(bs%bset(nf),stat = error)
  if(error .ne. 0) stop 'bond_stretch:create_bondstr_set: allocation error'

  select case (bs%bstype)
  case (1)  ! Hookean
   allocate(bs%bc(2), stat = error)
   if(error .ne. 0) stop 'bond_stretch:create_bondstr_set: allocation error'
  case (2)  ! FENE
   allocate(bs%bc(2), stat = error)
   if(error .ne. 0) stop 'bond_stretch:create_bondstr_set: allocation error'
  case default
   stop 'bond_stretch:create_bondstr_set: invalid bond stretch type'
  end select

  end subroutine create_bondstr_set
!***********************************************************************
  subroutine create_bondstr(bs, md)
!***********************************************************************
  implicit none

  type (bondstr_set), intent(inout) :: bs
  type(mol_set), intent(in) :: md
  integer :: i, error, nb

   do i = 1, md%nf
    nb = bs%bset(i)%nb
    if(nb .eq. 0) return
    allocate(bs%bset(i)%nblst(nb,2), stat = error)
    if(error .ne. 0) stop 'bond_stretch:create_bondstr: allocation error'
   enddo

  end subroutine create_bondstr
!***********************************************************************
  subroutine destroy_bondstr(bs)
!***********************************************************************
  implicit none

  type (bondstr_set), intent(inout) :: bs
  integer :: i

  do i = 1, size(bs%bset)
   if(allocated(bs%bset(i)%nblst)) deallocate(bs%bset(i)%nblst)
  enddo

  if(allocated(bs%bset)) deallocate(bs%bset)
  if(allocated(bs%bc)) deallocate(bs%bc)

  end subroutine destroy_bondstr
!***********************************************************************
  subroutine rd_bondstr_input(bs,md,header,ifile,iunit)
!***********************************************************************
! Reads bond stretch potential variables from input file.
! 
!***********************************************************************
  implicit none

  type (bondstr_set), intent(inout) :: bs
  type(mol_set), intent(in) :: md
  
  integer :: i, j, k, i1, iunit
  
  character(len=*) :: header
  character(len=32) :: rdstring
  character(len=32) :: ifile        ! input file names

  rdstring = ''
  i = 0
  open(iunit,file=ifile,status='old',form='formatted', &
     & position='rewind')
  do j = 1, 1000
   read(iunit,*, iostat=i) rdstring
   if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 999) exit
  enddo

  if (rdstring .eq. header) then
   read(iunit,*) bs%bstype           ! read the bond stretch type
   if(bs%bstype .eq. 0) then
    close(iunit)
    return                           ! no bond stretch forces - return
   endif
   read(iunit,*) bs%bsarch           ! read the bond stretch architecture
   call create_bondstr_set(bs, md%nf) 
   read(iunit,*) (bs%bset(i)%nb, i=1, md%nf) ! number of bonds
   read(iunit,*) (bs%bc(i), i=1, size(bs%bc)) ! bond stretch constants

   call create_bondstr(bs, md)

   do k = 1, md%nf
    select case(bs%bsarch)
    case(1)            ! linear chains
     call init_bondstr_linear(bs, md)
    case(2)            ! arbitrary molecular architecture
     do j = 1, bs%bset(k)%nb
      read(iunit,*) (bs%bset(k)%nblst(j,i1), i1 = 1, 2)
     enddo
    case default
     stop 'bond_stretch:rd_bondstr_input: Invalid molecular architecture'
    end select
   enddo

   close(iunit)
  else
   close(iunit)
   stop 'bond_stretch:rd_bondstr_input: Bond stretch input data not found'
  endif

  end subroutine rd_bondstr_input
!***********************************************************************
  subroutine init_bondstr_linear(bs, md)
!***********************************************************************
  implicit none
 
  type (bondstr_set), intent(inout) :: bs
  type(mol_set), intent(in) :: md
  integer :: i, j

! Simple bond stretch matrix initialization for linear chain molecules

  do i = 1, md%nf
   if(bs%bset(i)%nb .ne. md%nsf(i)-1) &
   & stop 'bond_stretch:init_bondstr_linear: error in number of bonds'
   bs%bset(i)%nblst = 0
   do j = 1, (md%nsf(i)-1)
    bs%bset(i)%nblst(j,1) = j
    bs%bset(i)%nblst(j,2) = j+1
   enddo
  enddo

  end subroutine init_bondstr_linear
!***********************************************************************
  subroutine wr_bondstr(bs,ofile,ounit)
! writes input data to the output file
!***********************************************************************
  type (bondstr_set), intent(in) :: bs
  integer :: ounit
  character(len=32) :: ofile      !output file name

  open(ounit,file=ofile,status='old',form='formatted',position='append')
  write (ounit,*) 'Initial bond stretch force values'
  write (ounit,*)
  write (ounit,*) 'bond stretch force type       = ', bs%bstype
  write (ounit,*) 'bond stretch architecture     = ', bs%bsarch
  write (ounit,*)
  close(ounit)
  
  end subroutine wr_bondstr
!***********************************************************************
  end module bond_stretch
