  module constraints
!***********************************************************************
! Module providing functionality for holonomic bond constraints.
!
!23456789012345678901234567890123456789012345678901234567890123456789012

  use v_sub_library
  use atoms
  use species
  use molecules
  use neighbour_list

  implicit none
 
  private
 
! Public procedures
  public :: add_constraint_forces, check_primary_constraints, &
          & create_constraints, create_constraint_set, &
          & fjc_molecule_maker, init_constr_fjc, init_constr_linalk, &
          & init_imask_constr, linalk_molecule_maker, rd_constraints, &
          & init_constr, update_secondary

! Public data types
  public :: constr_data_type, constraint_set

  type constr_data_type
   double precision, dimension(:,:), allocatable :: ksp ! constraint matrix M sparse format
   double precision, dimension(:,:), allocatable :: ka  ! constraint diff matrix L (nc,nc)
   double precision, dimension(:,:), allocatable :: xprev ! previous bond multipliers (nc)
   double precision, dimension(:,:), allocatable :: mol_template ! atom coordinates for molecule
   double precision, dimension(:,:), allocatable :: cai  ! c matrix for secondary particles (nsec,nbas)
   double precision, dimension(:,:), allocatable :: aabi  ! inverse of a matrix for secondary particles (nsec,nsec)
   double precision, dimension(:,:), allocatable :: nij  ! constraint matrix N (nbas,nc)
   double precision, dimension(:), allocatable :: blength ! bond lengths(nc)
   double precision, dimension(:), allocatable :: ndf ! no. of degrees of freedom for each atom type for each fraction
   double precision :: rmsberr ! rms bond length error per bond
   double precision :: rmsverr ! rms velocity error per bond
   integer, dimension(:,:), allocatable :: nclst    ! constraint list(nc,2)
   integer, dimension(:), allocatable :: ind     ! index vector for sparse storage
   integer, dimension(:), allocatable :: nblst   ! index vector for basic particles (nbas)
   integer, dimension(:), allocatable :: nslst   ! index vector for secondary particles (nsec)
   logical, dimension(:), allocatable :: is_basic_fr ! designate basic (TRUE) and secondary (FALSE) particles for this fraction (ns)
   integer :: mol_arch                     ! molecular architecture
                                           ! 0 = none - atomic
                                           ! 1 = arbitrary
                                           ! 2 = linear fjc
                                           ! 3 = linear alkane
   integer :: nc                           ! number of basic constraints
   integer :: its                          ! number of iterations required for iterative solver
   integer :: nbas                         ! number of basic particles per molecule
   integer :: nsec                         ! number of secondary particles (zero means no secondary particles)
  end type constr_data_type

  type constraint_set
   type (constr_data_type), dimension(:), allocatable :: cset  ! one for every fraction (md%nf)
   double precision :: cfb, dfb ! position and momentum feedback strength
   double precision :: stol ! tolerance for iterative solvers
   logical, dimension(:), allocatable :: is_basic ! mask array for basic particles
   integer :: csolvertype  ! selects the way to solve constraint equations
                           ! =0 none;
                           ! =1 for lussolv (direct);
                           ! =2 for lusolv first step, then gssolv (iterative)
                           ! =3 for gssolv (iterative) only
                           ! =4 for sparse, direct (lusolv_trid - linear fjc only)
                           ! =5 for sparse, iterative (gssolv_sp)
   logical :: cflag   ! true if good initial guesses for gssolv are available
   logical :: any_secondary=.FALSE.   ! true if any molecules with secondary particles are present
  end type constraint_set

  contains
!***********************************************************************
  subroutine add_constraint_forces(atd,md,spd,ca,reqm,peqm,pta)
! Provides a cleaner public interface to the old, ugly
! private constraint force procedure.
!***********************************************************************
  implicit none
 
  type(atom_type) :: atd
  type (constraint_set) :: ca
  type(mol_set) :: md
  type(species_set) :: spd
  double precision, dimension(:,:), allocatable :: fc, rfb, pfb
  double precision, dimension(:,:), intent(inout) :: reqm, peqm
  double precision, dimension(size(atd%r,2),size(atd%r,2)) :: pta
  integer :: i, j, k, isf, jsf, ndim, np, error

  ndim = size(atd%r,2)  

! Calculate constraints for each fraction of molecules in sequence.

  jsf = 0

! Index i in this loop denotes fraction number
  do i = 1, md%nf
   isf = jsf + 1              ! first atom for molecules of this fraction
   np = md%nsf(i)*md%nmf(i)   ! total number of particles in this fraction
   jsf = jsf + np             ! last atom for molecules of this fraction

! Go direct to the next fraction if there are no constraints for this fraction
   if(ca%cset(i)%nc .eq. 0) cycle

   allocate(fc(np,ndim),rfb(np,ndim),pfb(np,ndim), stat=error)
   if(error .ne. 0) stop 'constraints:add_constraint_forces: allocation error'

! Do calculations for secondary constraints if they are present
   if(ca%cset(i)%nsec .ne. 0) then
    call calculate_constraints_secondary( &
    & atd%r(isf:jsf,:),atd%p(isf:jsf,:),atd%f(isf:jsf,:),spd%mass(md%fr(i)%spid), &
    & ca%cset(i),md%nmf(i),md%nsf(i),fc,ca%csolvertype,ca%cflag, &
    & rfb,pfb,ca%stol)

   else

! Otherwise, calculate constraint and feedback forces for all constrained particles in this fraction
   call calculate_constraint_forces( &
    & atd%r(isf:jsf,:),atd%p(isf:jsf,:),atd%f(isf:jsf,:), &
    & ca%cset(i)%nclst,ca%cset(i)%ka,ca%cset(i)%ksp,ca%cset(i)%nc, &
    & md%nmf(i),md%nsf(i),fc,ca%csolvertype,ca%cset(i)%xprev,ca%cflag, &
    & ca%cset(i)%blength,rfb,pfb,ca%cset(i)%rmsberr,ca%cset(i)%rmsverr,ca%cset(i)%its, &
    & ca%stol,ca%cset(i)%ind,spd%mass(md%fr(i)%spid))

   endif

! Sum constraint forces and feedback into equations of motion
   reqm(isf:jsf,:) = reqm(isf:jsf,:) - ca%cfb*rfb
   peqm(isf:jsf,:) = peqm(isf:jsf,:) + fc - ca%dfb*pfb

! Add constraint force contribution to atomic pressure tensor
   do j = 1, ndim
    do k = 1, ndim
     pta(j,k) = pta(j,k) + sum(atd%r(isf:jsf,j)*fc(:,k))
    enddo
   enddo

   deallocate(pfb,rfb,fc)

  enddo

  end subroutine add_constraint_forces
!***********************************************************************
  subroutine calculate_constraint_forces(r,p,f,nclst,ka,ksp,nc,nm,ns,fc,&
    & cstype,xprev,cflag,bl,rfb,pfb,rmsberr,rmsverr,its,tol,isp,mass)
! Note: nij is the M matrix in the EEM papers, ka is the L matrix.
!***********************************************************************
  implicit none
 
  double precision, dimension(:,:), intent(in) :: r,p,f,ksp,ka
  double precision, dimension(:,:) :: rfb,pfb,fc
  double precision, dimension(:) :: bl, mass
  double precision, dimension(size(r,1),size(r,2)) :: ptmp, ftmp
  double precision, dimension(nc,nm) :: x12,y12,z12,px12,py12,pz12, &
                                      & fx12,fy12,fz12,b,ga,xprev
  double precision, dimension(nc,nm) :: afb, bfb
  double precision, allocatable, dimension(:,:,:) :: a
  double precision, allocatable, dimension(:,:) :: asp, a1, a2, a3
  double precision, dimension(nm) :: rsq
  double precision :: rmsberr, rmsverr, tol
  integer, dimension(:,:), intent(in) :: nclst
  integer, dimension(:), intent(in) :: isp
  integer :: i,j,k,nc,nm,ns,np,cstype,its,i1, i2, ndim, error
  logical :: cflag

! compute quantities r.f and (rdot)**2
! for all constrained distances within
! each molecule

  np = ns*nm
  ndim = size(r,2)
  
  do j = 1, ndim
   do i = 1, nm
    do k = 1, ns
     i1 = (i-1)*ns + k
     ptmp(i1,j) = p(i1,j)/mass(k)
     ftmp(i1,j) = f(i1,j)/mass(k)
    enddo
   enddo
  enddo

  do i = 1, nc
   x12(i,:) = r(nclst(i,2):np:ns,1) - r(nclst(i,1):np:ns,1)
   y12(i,:) = r(nclst(i,2):np:ns,2) - r(nclst(i,1):np:ns,2)
   z12(i,:) = r(nclst(i,2):np:ns,3) - r(nclst(i,1):np:ns,3)
   px12(i,:) = ptmp(nclst(i,2):np:ns,1) - ptmp(nclst(i,1):np:ns,1)
   py12(i,:) = ptmp(nclst(i,2):np:ns,2) - ptmp(nclst(i,1):np:ns,2)
   pz12(i,:) = ptmp(nclst(i,2):np:ns,3) - ptmp(nclst(i,1):np:ns,3)
   fx12(i,:) = ftmp(nclst(i,2):np:ns,1) - ftmp(nclst(i,1):np:ns,1)
   fy12(i,:) = ftmp(nclst(i,2):np:ns,2) - ftmp(nclst(i,1):np:ns,2)
   fz12(i,:) = ftmp(nclst(i,2):np:ns,3) - ftmp(nclst(i,1):np:ns,3)
  enddo

  b = -(fx12*x12 + fy12*y12 + fz12*z12 &
     &  + px12**2 + py12**2 + pz12**2)

! calculate symmetric dot product
! matrix a (size nc x nc)

  if(cstype .lt. 4) then ! calculate the full a matrix, including zeros
  allocate(a(nc,nc,nm), stat=error)
  if(error .ne. 0) stop 'constraints:calculate_constraint_forces: allocation error'

! Note: loop modified to go over all elements to allow loop transformation
  do i = 1, nc
   do j = 1, nc
    do k = 1, nm
     a(i,j,k) = (x12(i,k)*x12(j,k) + y12(i,k)*y12(j,k) &
      &       + z12(i,k)*z12(j,k))*ka(i,j)
    enddo
   enddo
  enddo

! Old version - Intel compiler optimises incorrectly by putting loop over i as inner loop
!  do i = 1, nc
!   do j = i, nc
!    do k = 1, nm
!     a(i,j,k) = (x12(i,k)*x12(j,k) + y12(i,k)*y12(j,k) &
!      &       + z12(i,k)*z12(j,k))*ka(i,j)
!     a(j,i,k) = a(i,j,k)
!    enddo
!   enddo
!  enddo

  elseif(cstype .eq. 4) then ! assume ka (L) is a tridiagonal matrix
   allocate(a1(nc,nm), a2(nc,nm), a3(nc,nm), stat=error)
   if(error .ne. 0) stop 'constraints:calculate_constraint_forces: allocation error'
   do k = 1, nm
    do i = 2, nc
     a1(i,k) = (x12(i,k)*x12(i-1,k) + y12(i,k)*y12(i-1,k) &
        &       + z12(i,k)*z12(i-1,k))*ka(i,i-1)
    enddo
    do i = 1, nc
     a2(i,k) = (x12(i,k)*x12(i,k) + y12(i,k)*y12(i,k) &
        &       + z12(i,k)*z12(i,k))*ka(i,i)
    enddo
    do i = 1, nc-1
     a3(i,k) = (x12(i,k)*x12(i+1,k) + y12(i,k)*y12(i+1,k) &
        &       + z12(i,k)*z12(i+1,k))*ka(i,i+1)
    enddo
   enddo
  elseif(cstype .eq. 5) then ! assume ka (L) is sparse
   k = size(isp)
   allocate(asp(k,nm), stat=error)
   if(error .ne. 0) stop 'constraints:calculate_constraint_forces: allocation error'
   do j = 1, nm
    i1 = nc
    do i = 1, nc
     asp(i,j) = (x12(i,j)*x12(i,j) + y12(i,j)*y12(i,j) &
      &       + z12(i,j)*z12(i,j))*ka(i,i)
     do k = 1, isp(i)
      i1 = i1 + 1
      asp(i1,j) = (x12(i,j)*x12(isp(i1),j) + y12(i,j)*y12(isp(i1),j) &
       &        +  z12(i,j)*z12(isp(i1),j))*ka(i,isp(i1))
     enddo
    enddo
   enddo
  else
   stop 'invalid value of constraint solver type'
  endif

! solve matrix equation for multipliers
! ga (array of size nc,nm)

  if(all(b .eq. 0.d0)) then
   ga = 0.d0   ! the trivial solution to the constraint equations
   xprev = 0.d0

  else

  ga = b

! call linear equation solver
  select case(cstype)
  case(0) ! set constraint forces to zero
   ga = 0.d0
  case(1) ! use direct solver - not sparse
   call lusolv(a, ga, nc, nm)
  case(2) ! use direct solver for first step, then iterative solver
   if(.not. cflag) then
    call lusolv(a, ga, nc, nm)
    xprev = ga
    cflag = .true.
   else
    call gssolv_1(a, ga, xprev, tol, nc, nm, its)
    xprev = ga
   endif
  case(3) ! use only the iterative solver
   call gssolv_1(a, ga, xprev, tol, nc, nm, its)
   xprev = ga
  case(4) ! use the tridiagonal solver (linear molecules, bond length constraints)
   call tridiag(a1, a2, a3, b, ga, nc, nm)
  case(5) ! use the sparse iterative solver
   call gssolv_sp_1(asp, isp, b, xprev, tol, its)
   ga = b
   xprev = b
  case default
   stop 'invalid value of constraint solver type'
  end select

 endif

! calculate constraint errors
  rmsberr = 0.d0
  rmsverr = 0.d0
  do j = 1, nc
   rsq = x12(j,:)*x12(j,:) + y12(j,:)*y12(j,:) + z12(j,:)*z12(j,:)
   afb(j,:) = sqrt(rsq) - bl(j)
   rmsberr = rmsberr + sum(afb(j,:)*afb(j,:))
   afb(j,:) = afb(j,:)/sqrt(rsq)
   bfb(j,:) = (px12(j,:)*x12(j,:) + py12(j,:)*y12(j,:) + pz12(j,:)*z12(j,:))/sqrt(rsq)
   rmsverr = rmsverr + sum(bfb(j,:)*bfb(j,:))
!   bfb(j,:) = (px12(j,:)*x12(j,:) + py12(j,:)*y12(j,:) + pz12(j,:)*z12(j,:))/rsq
  enddo
!do i = 1, nm
! print *, i, afb(1,i), afb(2,i), afb(3,i), afb(4,i), afb(5,i), bfb(1,i), afb(2,i), afb(3,i), afb(4,i), afb(5,i)
!enddo
  rmsberr = sqrt(rmsberr/(nm*nc))
  rmsverr = sqrt(rmsverr/(nm*nc))

! calculate constraint and feedback forces
! sum and group constaint and feedback forces for each
! site within the molecule

  rfb = 0.d0
  pfb = 0.d0

  fc = 0.d0

  do k = 1, nm
   i1 = (k-1)*ns ! site index offset for molecule k
   do i2 = 1, size(ksp,1) ! number of non-zero elements in constraint matrix
    i = ksp(i2,1)  ! index of site
    j = ksp(i2,2)  ! index of constraint
    rfb(i+i1,1) = rfb(i+i1,1) + afb(j,k)*ksp(i2,3)*x12(j,k)
    rfb(i+i1,2) = rfb(i+i1,2) + afb(j,k)*ksp(i2,3)*y12(j,k)
    rfb(i+i1,3) = rfb(i+i1,3) + afb(j,k)*ksp(i2,3)*z12(j,k)
    pfb(i+i1,1) = pfb(i+i1,1) + bfb(j,k)*ksp(i2,3)*x12(j,k)
    pfb(i+i1,2) = pfb(i+i1,2) + bfb(j,k)*ksp(i2,3)*y12(j,k)
    pfb(i+i1,3) = pfb(i+i1,3) + bfb(j,k)*ksp(i2,3)*z12(j,k)
    fc(i+i1,1) = fc(i+i1,1) + x12(j,k)*ga(j,k)*ksp(i2,3)
    fc(i+i1,2) = fc(i+i1,2) + y12(j,k)*ga(j,k)*ksp(i2,3)
    fc(i+i1,3) = fc(i+i1,3) + z12(j,k)*ga(j,k)*ksp(i2,3)
   enddo
  enddo
  if(allocated(a)) deallocate(a)
  if(allocated(asp)) deallocate(asp)
  if(allocated(a1)) deallocate(a1)
  if(allocated(a2)) deallocate(a2)
  if(allocated(a3)) deallocate(a3)

  end subroutine calculate_constraint_forces
!***********************************************************************
  subroutine calculate_constraints_secondary(ri,pi,fi,mass,c, &
  & nm,ns,fc,csolvertype,cflag,rfb,pfb,stol)
! Inputs: forces on atoms and 
!***********************************************************************
  implicit none
  type (constr_data_type) :: c
  integer :: nm
  double precision, dimension(:,:) :: ri, pi, fi, rfb, pfb, fc
  double precision, dimension(c%nsec*nm,size(fi,2)) :: tij, mu
  double precision, dimension(c%nbas*nm,size(fi,2)) :: rb, pb, fb, fcb, rfbb, pfbb
  double precision, dimension(:) :: mass
  double precision, dimension(c%nbas) :: massb
  double precision :: stol ! tolerance for iterative solvers
  double precision :: t1
  integer, dimension(c%nbas*nm) :: ibvec
  integer, dimension(c%nsec*nm) :: isvec
  integer :: im, ism, is, isec, ibas, ip, idim, ndim, ns, csolvertype
  integer :: i, j, k, ia, ib
  logical :: cflag
  
  ndim = size(ri,2)

! Set up array of masses of basic particles
  ibas = 0
  do i = 1, ns
   if(.not. c%is_basic_fr(i)) cycle
   ibas = ibas + 1
   massb(ibas) = mass(i)
  enddo

! Set up the vectors of indices of basic and secondary particles for this fraction
! Note: later - use the molecule index vectors that have already been calculated c%nblst and c%nslst
  ia = 0
  ib = 0
  ip = 0
  do i = 1, nm
   do is = 1, ns
    ip = ip + 1
    if(c%is_basic_fr(is)) then
     ib = ib + 1
     ibvec(ib) = ip
    else
     ia = ia + 1
     isvec(ia) = ip
    endif
   enddo
  enddo

! calculate the T matrix
  tij = 0.d0
  isec = 0
  do im = 1, nm
   ism = 0
   do is = 1, ns
    if(c%is_basic_fr(is)) cycle
    isec = isec + 1    ! index of secondary particles for this fraction
    ism = ism + 1      ! index of secondary particles for this molecule
    ibas = 0
    do idim = 1, ndim
     tij(isec,idim) = -fi((im-1)*ns + is,idim)/mass(is)
    enddo
    do j = 1, ns
     ip = (im-1)*ns + j
     if(.not. c%is_basic_fr(j)) cycle
     ibas = ibas + 1   ! index of basic particles for this molecule
     do idim = 1, ndim
      tij(isec,idim) = tij(isec,idim) + c%cai(ism,ibas)*fi(ip,idim)/mass(j)
     enddo
    enddo
   enddo
  enddo

! Calculate the script F matrix for basic particles
  fb = fi(ibvec,:)

  i = 0
  do im = 1, nm
   ibas = 0
   do is = 1, ns
    if(.not. c%is_basic_fr(is)) cycle
    i = i+1
    ibas = ibas + 1
    ip = (im-1)*c%nsec
    ia = 0
    do j = 1, ns
     if(c%is_basic_fr(j)) cycle
     ia = ia + 1
     ib = 0
     do k = 1, ns
      if(c%is_basic_fr(k)) cycle
      ib = ib + 1
      do idim = 1, ndim
       fb(i,idim) = fb(i,idim) - c%cai(ib,ibas)*c%aabi(ia,ib)*tij(ip+ia,idim)
      enddo
     enddo  
    enddo
   enddo  
  enddo
  
  rb = ri(ibvec,:)
  pb = pi(ibvec,:)
  rfbb = 0.d0
  pfbb = 0.d0

  fcb = 0.d0

! Calculate constraint forces on basic particles
  call calculate_constraint_forces( &
  & rb,pb,fb,c%nclst,c%ka,c%ksp,c%nc, &
  & nm,c%nbas,fcb,csolvertype,c%xprev,cflag, &
  & c%blength,rfbb,pfbb,c%rmsberr,c%rmsverr,c%its, &
  & stol,c%ind,massb)

  rfb(ibvec,:) = rfbb
  pfb(ibvec,:) = pfbb

! Calculate secondary constraint forces
  mu = 0.d0
  do idim = 1, ndim
  ip = 0
   do im = 1, nm
    do is = 1, c%nsec
     ip = ip + 1
     do ia = 1, c%nsec
      ib = (im-1)*c%nsec + ia
      t1 = 0.d0
      do ibas = 1, c%nbas
       k = (im-1)*c%nbas + ibas
       t1 = t1 + c%cai(ia,ibas)*fcb(k,idim)/massb(ibas)
      enddo
      mu(ip,idim) = mu(ip,idim) + c%aabi(is,ia)*(tij(ib,idim) + t1)
     enddo
    enddo
   enddo
  enddo

! Add secondary constraint force contributions to basic particles
  do idim = 1, ndim
  ip = 0
   do im = 1, nm
    do is = 1, c%nbas
     ip = ip + 1
     do ia = 1, c%nsec
      ib = (im-1)*c%nsec + ia
      fcb(ip,idim) = fcb(ip,idim) - c%cai(ia,is)*mu(ib,idim)
     enddo
    enddo
   enddo
  enddo

  fc(ibvec,:) = fcb

  fc(isvec,:) = mu
  
  end subroutine calculate_constraints_secondary
!***********************************************************************
  subroutine check_primary_constraints(ca, md, spd)
! Check input data to ensure that a massless particle (interaction site)
! has not been designated as primary.
!***********************************************************************
  implicit none
 
  type (constraint_set) :: ca
  type(mol_set), intent(in) :: md
  type(species_set) :: spd
  integer :: i, j

  do i = 1, md%nf
   do j = 1, md%nsf(i)
    if(ca%cset(i)%is_basic_fr(j) .and. .not. spd%has_mass(md%fr(i)%spid(j))) &
     & stop 'constraints:check_primary_constraints: A site cannot be primary and massless.'
   enddo
  enddo

  end subroutine check_primary_constraints
!***********************************************************************
  subroutine create_constraints(cset, md, i)
!***********************************************************************
  implicit none
 
  type (constr_data_type) :: cset
  type(mol_set), intent(in) :: md
  integer :: i, error, nbas, ns, nc, nm

  ns = md%nsf(i)
  nbas = cset%nbas
  nc = cset%nc
  nm = md%nmf(i)
  allocate(cset%is_basic_fr(ns), stat = error)
  if(error .ne. 0) stop 'constraints:create_constraints: allocation error'
  if(nc .eq. 0) return
  allocate(cset%nij(nbas,nc), cset%ka(nc,nc), cset%nclst(nc,2), &
          & cset%blength(nc), cset%xprev(nc,nm), stat = error)
  if(error .ne. 0) stop 'constraints:create_constraints: allocation error'

  end subroutine create_constraints
!***********************************************************************
  subroutine create_constraint_set(ca, nf)
!***********************************************************************
  implicit none
 
  type (constraint_set) :: ca
  integer, intent(in) :: nf
  integer :: error

  allocate(ca%cset(nf),stat = error)
  if(error .ne. 0) stop 'constraints:create_constraint_set: allocation error'

  end subroutine create_constraint_set
!***************************************************************************** 
  subroutine fjc_molecule_maker(rt, ns, bond_length)
!*****************************************************************************
  implicit none
 
  double precision, dimension(:,:) :: rt
  double precision, allocatable, dimension(:,:) :: rt_temp
  double precision :: bond_length
  integer :: ns, ppc, ppe, ppp, index, i, j, k, error

  ppe = int(ns**(1.d0/3.d0)) + 1
  ppp = ppe * ppe
  ppc = ppp * ppe
  index = 1

  rt = 0.d0

  allocate(rt_temp(ppc,3), stat=error)
  if(error .ne. 0) stop 'constraints:fjc_molecule_maker: allocation error'

  rt_temp = 0.d0
 
  do i = 0, (ppe - 1)
   rt_temp(index,1) = real(i)
   index = index +1
  enddo

  do j = 1, (ppe - 1)
   rt_temp(index:(index + ppe - 1),1) = rt_temp((index - 1):(index - ppe):-1,1)
   rt_temp(index:(index + ppe - 1),2) = real(j)
   index = index + ppe
  enddo

  do k = 1, (ppe - 1) 
   rt_temp(index:(index + ppp - 1),1) = rt_temp((index - 1):(index - ppp):-1,1)
   rt_temp(index:(index + ppp - 1),2) = rt_temp((index - 1):(index - ppp):-1,2)
   rt_temp(index:(index + ppp - 1),3) = real(k)
   index = index + ppp
  enddo

  rt = rt_temp(1:ns,:)

  deallocate(rt_temp)

  rt = rt*bond_length

  end subroutine fjc_molecule_maker
!***********************************************************************
  subroutine init_constr(ca,md,spd,header,ifile,iunit)
!***********************************************************************
!Initialize constraint matrices.
! 
!***********************************************************************
  implicit none
 
  type(constraint_set) :: ca
  type(mol_set), intent(in) :: md
  type(species_set) :: spd

  integer :: i, j, k, iunit, i1, i2, error
  
  character(len=*) :: header
  character(len=32) :: rdstring
  character(len=32) :: ifile      !input file name

  ca%any_secondary = .FALSE.

  open(iunit,file=ifile,status='old',form='formatted', &
     & position='rewind')

   do k = 1, md%nf

    rdstring = ''
    i = 0
    do j = 1, 1000
     read(iunit,*, iostat=i) rdstring
     if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 999) exit
    enddo

    if (rdstring .ne. header) then
     close(iunit)
     stop 'constraints:init_constr: constraint input matrices not found'
    endif

    read(iunit,*) ca%cset(k)%mol_arch
    read(iunit,*) ca%cset(k)%nc

! read number of secondary particles
    read(iunit,*) ca%cset(k)%nsec
    ca%cset(k)%nbas = md%nsf(k) - ca%cset(k)%nsec

    call create_constraints(ca%cset(k), md, k)  ! allocate constraint arrays for each fraction

    select case(ca%cset(k)%mol_arch)
    case (0)                            ! no constraints
     ca%cset(k)%is_basic_fr(:) = .true. ! all particles are basic particles
     allocate(ca%cset(k)%ndf(md%nsf(k)), stat=error) ! Stephen alteration 11/12/13
     if(error .ne. 0) stop 'constraints:init_constr: allocation error'
     ca%cset(k)%ndf(1:md%nsf(k)) = 3.d0  ! Degrees of freedom for atom with no constraints
     cycle
    case (1)                            ! arbitrary - read from file
     call init_constr_arbitrary         ! internal subroutine to aid program readability
    case (2)                            ! linear fjc molecules
     call init_constr_fjc(ca%cset(k), md%nsf(k))
    case (3)                            ! linear alkane molecules
     call init_constr_linalk(ca%cset(k), md%nsf(k))
    case default
     stop 'constraints:init_constr: Invalid molecular architecture'
    end select

! Assign the number of degrees of freedom for each atom in a molecule of fraction i 
! Modified for molecules with secondary particles - PD 20/8/15
    allocate(ca%cset(k)%ndf(md%nsf(k)), stat=error)
    if(error .ne. 0) stop 'constraints:init_constr: allocation error'
    i1 = 0
    do i = 1, md%nsf(k)                ! loop over atoms in a molecule of this fraction
     if(ca%cset(k)%is_basic_fr(i)) then
      i1 = i1 + 1
      ca%cset(k)%ndf(i) = 3.d0  ! basic atoms start with 3 degrees of freedom
      do j = 1, ca%cset(k)%nc           ! loop over constraints
! Subtract away 1/2 degree of freedom for every constraint on particle
       ca%cset(k)%ndf(i) = ca%cset(k)%ndf(i) - 0.5d0*abs(ca%cset(k)%nij(i1,j))
      enddo
     else
      ca%cset(k)%ndf(i) = 0.d0  ! secondary atoms have no degrees of freedom
     endif
    enddo

    if(ca%cset(k)%nc .eq. 0) cycle

! Initialize the sparse form of the M matrix
! The M matrix is currently stored in nij
! Put evaluation of count function outside allocate intrinsic procedure to satisfy intel compiler
    i2 = count(ca%cset(k)%nij .ne. 0.d0)
    allocate(ca%cset(k)%ksp(i2,3), stat=error)
    if(error .ne. 0) stop 'constraints:init_constr: allocation error'
! Assign values to the sparse storage version of M
    i1 = 0
    i2 = 0
    do i = 1, md%nsf(k)
     if(.not. ca%cset(k)%is_basic_fr(i)) cycle
     i2 = i2 + 1
     do j = 1, ca%cset(k)%nc
      if(ca%cset(k)%nij(i,j) .eq. 0.d0) cycle
      i1 = i1 + 1
      ca%cset(k)%ksp(i1,1) = i2
      ca%cset(k)%ksp(i1,2) = j
!      ca%cset(k)%ksp(i1,3) = ca%cset(k)%nij(i2,j)*spd%mass(spd%spid(i))
      ca%cset(k)%ksp(i1,3) = ca%cset(k)%nij(i2,j)
     enddo
    enddo

! The N matrix is now the M matrix divided by the mass
    j = 0
    do i = 1, md%nsf(k)
     if(.not. ca%cset(k)%is_basic_fr(i)) cycle
     j = j + 1
     ca%cset(k)%nij(j,:) = ca%cset(k)%nij(j,:)/spd%mass(md%fr(k)%spid(i))
    enddo

! do initializations if there are secondary particles for this fraction
    if(ca%cset(k)%nsec .ne. 0) call init_secondary(ca%cset(k), md, spd, k)

! Construct L matrix from difference of rows of N matrix
    do j = 1, ca%cset(k)%nc
     do i = 1, ca%cset(k)%nc
      ca%cset(k)%ka(j,i) = ca%cset(k)%nij(ca%cset(k)%nclst(j,2),i) &
                       & - ca%cset(k)%nij(ca%cset(k)%nclst(j,1),i)
     enddo
    enddo

! Construct sparse L matrix index vector
! Allocate the vector
    allocate(ca%cset(k)%ind(count_sp(ca%cset(k)%ka)), stat=error)
    if(error .ne. 0) stop 'constraints:init_constr: allocation error'
! Assign values to the index vector
    call set_spind(ca%cset(k)%ka,ca%cset(k)%ind)

    ca%cset(k)%xprev = 0.d0

   enddo

   close(iunit)

  call init_basic_mask(ca, md)

  ca%cset(:)%its = 0
  ca%cset(:)%rmsberr = 0.d0
  ca%cset(:)%rmsverr = 0.d0

  contains
!***********************************************************************
  subroutine init_constr_arbitrary
! Internal subroutine  - only used by init_constr.
! Has access to all variables in init_constr.
!***********************************************************************
  implicit none

! read data for secondary particles
  if(ca%cset(k)%nsec > 0) ca%any_secondary = .TRUE.
  read(iunit,*) (ca%cset(k)%is_basic_fr(j), j=1, md%nsf(k))
  if((ca%cset(k)%nsec + count(ca%cset(k)%is_basic_fr)) .ne. md%nsf(k)) &
  & stop 'constraints:init_constr_arbitrary:number of primary plus secondary particles does not match number of particles'

  if(ca%cset(k)%nc .ne. 0) then    

! read constraint list and bond lengths
  do i = 1, ca%cset(k)%nc
   read(iunit,*) ca%cset(k)%nclst(i,1), ca%cset(k)%nclst(i,2), ca%cset(k)%blength(i)
  enddo

! read constraint matrix for basic particles
  do i = 1, ca%cset(k)%nbas
   read(iunit,*) (ca%cset(k)%nij(i,j), j=1, ca%cset(k)%nc)
  enddo
  endif

! allocate matrices for secondary constraints
  if(ca%cset(k)%nsec .ne. 0) then
   allocate(ca%cset(k)%cai(ca%cset(k)%nsec,ca%cset(k)%nbas), stat = error)
   if(error /= 0) stop 'constraints:init_constr: allocation error for ca%cset(k)%cai'
   allocate(ca%cset(k)%aabi(ca%cset(k)%nsec,ca%cset(k)%nsec), stat = error) ! allocate A inv matrix
   if(error /= 0) stop 'constraints:init_constr: allocation error for ca%cset(k)%aabi'
   allocate(ca%cset(k)%nblst(ca%cset(k)%nbas), stat = error) ! allocate index vector for basic particles
   if(error /= 0) stop 'constraints:init_constr: allocation error for ca%cset(k)%nblst'
   allocate(ca%cset(k)%nslst(ca%cset(k)%nsec), stat = error) ! allocate index vector for secondary particles
   if(error /= 0) stop 'constraints:init_constr: allocation error for ca%cset(k)%nslst'

! read C matrix for secondary particles
   do i = 1, ca%cset(k)%nsec
    read(iunit,*) (ca%cset(k)%cai(i,j), j=1, ca%cset(k)%nbas)
   enddo

  endif

! create and initialize array for starting configuration of molecule
  allocate(ca%cset(k)%mol_template(md%nsf(k),size(md%rcm,2)), stat = error)
  if(error /= 0) stop 'constraints:init_constr_arbitrary: allocation error'
  do i = 1, md%nsf(k)
   read(iunit,*) (ca%cset(k)%mol_template(i,j), j=1, size(md%rcm,2))
  enddo

  end subroutine init_constr_arbitrary
!***********************************************************************
  end subroutine init_constr
!***********************************************************************
  function matinv(matrix)
!***********************************************************************
! compute the inverse of a square matrix
  implicit none
  double precision, dimension(:,:) :: matrix
  double precision, dimension(size(matrix,1),size(matrix,1)) :: matinv
  
  select case(size(matrix,1))
  case(1)
   matinv(1,1) = 1.d0/matrix(1,1)
  case(2)
   stop 'matinv: not implemented'
  case(3)
   stop 'matinv: not implemented'
  case default
   stop 'matinv: not implemented'
  end select
  end function matinv
!***********************************************************************
  subroutine init_secondary(c,md,spd,k)
!***********************************************************************
  implicit none
  type (constr_data_type) :: c
  type(mol_set), intent(in) :: md
  type(species_set) :: spd
  double precision, dimension(c%nbas,c%nc) :: ntemp
  integer :: isec, jsec, i, j, i1, i2, ibas, kbas, i3, k
  double precision :: s1

! Construct index vectors for basic and secondary particles
  ibas = 0
  isec = 0
  do i = 1, md%nsf(k)
   if(c%is_basic_fr(i)) then
    ibas = ibas + 1
    c%nblst(ibas) = i
   else
    isec = isec + 1
    c%nslst(isec) = i
   endif
  enddo

! Calculate the A matrix and its inverse
  isec = 0
  do i = 1, md%nsf(k)
   if(c%is_basic_fr(i)) cycle
   isec = isec + 1
   jsec = 0
   do j = 1, md%nsf(k)
     if(c%is_basic_fr(j)) cycle
     jsec = jsec + 1
     c%aabi(isec,jsec) = 0.d0
     if(i .eq. j) c%aabi(isec,jsec) = 1.d0/spd%mass(md%fr(k)%spid(j))  !mass of secondary particles is required here
     do i1 = 1, md%nsf(k)
      if(.not. c%is_basic_fr(i1)) cycle
      c%aabi(isec,jsec) = c%aabi(isec,jsec) + &
      & c%cai(isec,i1)*c%cai(jsec,i1)/spd%mass(md%fr(k)%spid(i1))
     enddo
    enddo
    c%aabi = matinv(c%aabi) ! replace aabi with its inverse
  enddo

! Calculate the N matrix
  ntemp = c%nij
  ibas = 0
  do i = 1, md%nsf(k)
   if(.not. c%is_basic_fr(i)) cycle
   ibas = ibas + 1
   do j = 1, c%nc
    isec = 0
    do i1 = 1, md%nsf(k)
     if(c%is_basic_fr(i1)) cycle
     isec = isec + 1
     jsec = 0
     do i2 = 1, md%nsf(k)
      if(c%is_basic_fr(i2)) cycle
      jsec = jsec + 1
      kbas = 0
      s1 = 0.d0
      do i3 = 1, md%nsf(k)
       if(.not. c%is_basic_fr(i3)) cycle
       kbas = kbas + 1
       s1 = s1 + c%cai(jsec,kbas)*ntemp(kbas,j)
      enddo
      c%nij(ibas,j) = ntemp(ibas,j) - &
      & c%cai(isec,ibas)*c%aabi(isec,jsec)*s1/spd%mass(md%fr(k)%spid(i))
     enddo
    enddo
   enddo
  enddo

  end subroutine init_secondary
!***********************************************************************
  subroutine init_constr_fjc(cd, ns)
!***********************************************************************
  implicit none
 
  type (constr_data_type) :: cd
  integer :: i, ns

! Simple constraint matrix initialization for linear fjc molecules

  cd%nij = 0.d0
  cd%nclst = 0
  cd%blength = 1.d0

  do i = 1, ns-1
   cd%nclst(i,1) = i
   cd%nclst(i,2) = i+1
  enddo

  do i = 1, ns-1
   cd%nij(i,i) = -1.d0
   cd%nij(i+1,i) = 1.d0
  enddo

  cd%is_basic_fr(:) = .true. ! all particles are basic particles

  end subroutine init_constr_fjc
!***********************************************************************
  subroutine init_constr_linalk(cd, ns)
! Constraint matrix initialization for linear alkanes
!***********************************************************************
  implicit none
 
  type (constr_data_type) :: cd
  double precision :: bangle, d1, d2, pi
  integer :: i, j, ns

  pi = 4.d0*atan(1.d0)
  bangle = 2.d0*acos(0.5d0/cos(pi/6.d0))
  d1 = 0.3900d0        ! take care here: reduced units are assumed
  d2 = 2.d0*d1*sin(bangle/2.d0)

! Simple constraint matrix initialization for linear alkane molecules.
! The first nsf-1 constraints are bond length constraints, and
! the next nsf-2 constraints are bond angle constraints.

  do i = 1, ns-1     ! bond length constraints
   cd%nclst(i,1) = i
   cd%nclst(i,2) = i+1
   cd%blength(i) = d1
  enddo

  do i = 1, ns-2  ! bond angle constraints
   j = i + ns - 1
   cd%nclst(j,1) = i
   cd%nclst(j,2) = i+2
   cd%blength(j) = d2
  enddo

  cd%nij = 0.d0

  do i = 1, ns-1  ! bond length constraints
   cd%nij(i,i) = -1.d0
   cd%nij(i+1,i) = 1.d0
  enddo

  do i = 1, ns-2  ! bond angle constraints
   j = i + ns - 1
   cd%nij(i,j) = -1.d0
   cd%nij(i+2,j) = 1.d0
  enddo

  cd%is_basic_fr(:) = .true. ! all particles are basic particles

  end subroutine init_constr_linalk
!***********************************************************************
  subroutine init_imask_constr(ca, md)
! Bonded (constrained) sites do not interact via a potential.
! Set the interaction mask to false for pairs of sites that are joined
! by a constraint.
!***********************************************************************
  implicit none

  type (constraint_set) :: ca
  type(mol_set), intent(inout) :: md
  integer :: j, k, ii, ij

! set mask to false for every pair of atoms that are constrained
! if there are secondary particles, assume that all particles are constrained
  do j = 1, md%nf
   if(ca%cset(j)%nsec .ne. 0) then
    md%frac_mask(j)%fmask = .false.
   else
    do k = 1, ca%cset(j)%nc
     ii = ca%cset(j)%nclst(k,1)
     ij = ca%cset(j)%nclst(k,2)
     md%frac_mask(j)%fmask(ii,ij) = .false.
     md%frac_mask(j)%fmask(ij,ii) = .false.
    enddo
   endif
  enddo

  end subroutine init_imask_constr
!***********************************************************************
  subroutine init_basic_mask(ca, md)
! Initialize the mask to identify basic particles for solution of
! the equations of motion.
!***********************************************************************
  implicit none

  type (constraint_set) :: ca
  type(mol_set), intent(in) :: md
  integer :: i, j, k, is, error

  allocate(ca%is_basic(md%np), stat = error)
  if(error .ne. 0) stop 'constraints:init_basic_mask: allocation error'

! save the mask to identify basic particles
  is = 0
  do i = 1, md%nf
   do j = 1, md%nmf(i)
    do k = 1, md%nsf(i)
     is = is + 1
     if(ca%cset(i)%is_basic_fr(k)) ca%is_basic(is) = .true.
    enddo
   enddo
  enddo

  end subroutine init_basic_mask
!***************************************************************************** 
  subroutine linalk_molecule_maker(rt, ns, blen, ba, da)
!*****************************************************************************
  implicit none
 
  double precision, dimension(:,:) :: rt
  double precision, dimension(3) :: a, b, c, cprime
  double precision, dimension(3,3) :: epr, t
  double precision :: blen, ba, da ! bond length, bond angle, dihedral angle
  double precision :: pi, pi2, theta, epr1l, epr2l, bl, phi
  integer :: ns  ! number of sites
  integer :: i, j, k

  pi2 = 2.0d0*atan(1.0d0)
  pi = 2.0d0*pi2
  theta = ba*pi/180.0d0
  phi = da*pi/180.0d0

! define the first three sites
  rt(1,1) = -blen*cos(pi-theta)
  rt(1,2) = blen*sin(pi-theta)
  rt(1,3) = 0.0d0
  rt(2,1) = 0.0d0
  rt(2,2) = 0.0d0
  rt(2,3) = 0.0d0
  rt(3,1) = blen
  rt(3,2) = 0.0d0
  rt(3,3) = 0.0d0

! define first two bond vectors
! required to define a plane for the next bond
  a = -rt(1,:)
  b = rt(3,:)

! now do the other sites
  do j = 4, ns

! unit vectors for the primed coordinate system
! z given by the direction of b
! y given by b cross a
! x given by y cross z
   epr(2,1) = b(2)*a(3)-b(3)*a(2)
   epr(2,2) = b(3)*a(1)-b(1)*a(3)
   epr(2,3) = b(1)*a(2)-b(2)*a(1)
   epr2l = sqrt(epr(2,1)**2+epr(2,2)**2+epr(2,3)**2)
   bl = sqrt(b(1)**2+b(2)**2+b(3)**2)

   do i = 1, 3
    epr(2,i) = epr(2,i)/epr2l
    epr(3,i) = b(i)/bl
   enddo

   epr(1,1) = epr(2,2)*epr(3,3)-epr(2,3)*epr(3,2)
   epr(1,2) = epr(2,3)*epr(3,1)-epr(2,1)*epr(3,3)
   epr(1,3) = epr(2,1)*epr(3,2)-epr(2,2)*epr(3,1)
   epr1l = sqrt(epr(1,1)**2+epr(1,2)**2+epr(1,3)**2)

   do i = 1, 3
    epr(1,i) = epr(1,i)/epr1l
   enddo

! transformation matrix
   do i = 1, 3
    do k = 1, 3
     t(i,k) = epr(k,i)
    enddo
   enddo

   cprime(1) = blen*cos(theta-pi2)*cos(phi)
   cprime(2) = blen*cos(theta-pi2)*sin(phi)
   cprime(3) = blen*sin(theta-pi2)

! transform to unprimed coordinate system
   do i = 1, 3
    c(i) = 0.d0
    do k = 1, 3
     c(i) = c(i) + t(i,k)*cprime(k)
    enddo
   enddo

! new point is given by the sum of vectors
   rt(j,:) = rt(j-1,:) + c(:)

! update the two most recent bonds
   a = b
   b = c

  enddo

  end subroutine linalk_molecule_maker
!***********************************************************************
  subroutine rd_constraints(ca,header,ifile,iunit)
!***********************************************************************
! Reads constraint variables from input file.
! 
!***********************************************************************
  implicit none
 
  type(constraint_set) :: ca
  
  integer :: i, j, iunit
  
  character(len=*) :: header
  character(len=32) :: rdstring
  character(len=32) :: ifile      !input file name

  rdstring = ''
  i = 0
  open(iunit,file=ifile,status='old',form='formatted', &
     & position='rewind')
  do j = 1, 1000
   read(iunit,*, iostat=i) rdstring
   if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 999) exit
  enddo

  if (rdstring .eq. header) then
   read(iunit,*) ca%csolvertype
   read(iunit,*) ca%cfb, ca%dfb
   read(iunit,*) ca%stol
   close(iunit)
  else
   close(iunit)
   stop 'constraints:rd_constraints: constraint input data not found'
  endif

! initialize flag for first run
  ca%cflag = .false.

! check for invalid input values

  end subroutine rd_constraints
!***********************************************************************
  subroutine update_secondary(atd, md, spd, ca, dr)
!***********************************************************************
! Update positions and momenta of secondary particles.
! 
!***********************************************************************
  implicit none
 
  type(atom_type) :: atd
  type(constraint_set) :: ca
  type(mol_set) :: md
  type(species_set) :: spd
  double precision, dimension(:,:), optional :: dr
  double precision :: massic, rold
  integer :: ia, ib, ic, ifr, is, fs, idim, ndim, nmf, im
  
  ndim = size(atd%r,2)

  fs = 1
  nmf = 0
  do ifr = 1, md%nf
   fs = fs + nmf
   nmf = md%nmf(ifr)
   if(ca%cset(ifr)%nsec .eq. 0) cycle
   do im = 1, md%nmf(ifr)
    do is = 1, ca%cset(ifr)%nsec
     ia = fs - 1 + (im-1)*md%nsf(ifr) + ca%cset(ifr)%nslst(is)
     do idim = 1, ndim
      rold = atd%r(ia,idim)
      atd%r(ia,idim) = 0.d0
      atd%p(ia,idim) = 0.d0
      do ib = 1, ca%cset(ifr)%nbas
       ic = fs - 1 + (im-1)*md%nsf(ifr) + ca%cset(ifr)%nblst(ib)
       massic = spd%mass(spd%spid(ic))
       atd%r(ia,idim) = atd%r(ia,idim) + ca%cset(ifr)%cai(is,ib)*atd%r(ic,idim)
       atd%p(ia,idim) = atd%p(ia,idim) + ca%cset(ifr)%cai(is,ib)*atd%p(ic,idim)/massic
      enddo
      atd%p(ia,idim) = spd%mass(spd%spid(ia))*atd%p(ia,idim)
      if(present(dr)) dr(ia,idim) = atd%r(ia,idim) - rold
     enddo
    enddo
   enddo
  enddo

  end subroutine update_secondary
!***********************************************************************
  end module constraints
