  module dihedral_force
!***********************************************************************
! Calculates the dihedral angles, force and energy.
! Curently only for rigid bonds and bond angles.
! Based on alk20.f. For flexible bonds, see butane8.f (Toxvaerd
! potential)
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************
  use atoms
  use molecules
  use species
  use neighbour_list
  use constraints

  implicit none

  private

! Public procedures
  public :: add_dihedral_forces, destroy_dihed, init_imask_dihed, &
           & rd_dihed_input, wr_dihed

! Public data types
  public :: dihed_data_type, dihed_set

  type :: dihed_data_type
   integer :: nda ! num. dihed angles
   integer, dimension(:,:), allocatable :: ndalst ! dihedral interaction list
  end type dihed_data_type

  type :: dihed_set
   integer :: dfarch ! = 0 for no dihedral forces
                     ! = 1 assume all linear alkanes, ndalst is computed
                     ! = 2 for arbitrary architecture, ndalst is in a file
   type(dihed_data_type), dimension(:), allocatable :: dset
   double precision, dimension(:), allocatable :: dc ! dihedral potential constants
   double precision :: dai ! initial dihedral angle when making molecules
 end type dihed_set

  contains
!***********************************************************************
  subroutine add_dihedral_forces(atd,da,md,ca,ud,pta)
! Provides a cleaner public interface to the old, ugly
! private constraint force procedure.
!***********************************************************************
  implicit none

  type(atom_type) :: atd
  type (dihed_set) :: da
  type(mol_set) :: md
  type (constraint_set) :: ca

  double precision :: ud, dsq, sintsq
  double precision, dimension(size(atd%r,2),size(atd%r,2)) :: pta
  double precision, dimension(size(atd%r,1),size(atd%r,2)) :: fdi
  integer :: i, isf, jsf, np, ndim

  ndim = size(atd%r,2)

  ud = 0.d0   ! initialize total dihedral energy

! Calculate dihedral forces for each fraction of molecules in sequence.

  jsf = 0

  do i = 1, md%nf

   np = md%nsf(i)*md%nmf(i)
   isf = jsf + 1
   jsf = isf - 1 + np

   if(da%dset(i)%nda .eq. 0) cycle

   fdi = 0.d0

! temporarily calculate dsq and sinthetasq (sintsq) for linear alkanes
! every time this procedure is called - later, incorporate flexible angles etc.

   dsq = ca%cset(i)%blength(1)**2
   sintsq = (ca%cset(i)%blength(ca%cset(i)%nc)**2/dsq)*(1.d0 - &
            & ca%cset(i)%blength(ca%cset(i)%nc)**2/(4.d0*dsq))

   call dihedral_force_energy(atd%r(isf:jsf,:), md%nsf(i), md%nmf(i), &
    & da%dset(i)%ndalst,da%dc,dsq,sintsq,fdi(isf:jsf,:), atd%eint(isf:jsf), ud, pta)

   atd%f(isf:jsf,:) = atd%f(isf:jsf,:) + fdi(isf:jsf,:)

  enddo

  end subroutine add_dihedral_forces
!***********************************************************************
  subroutine dihedral_force_energy(r,ns,nmf,ndalst,ak,dsq,sintsq,fi,ei,ud,pta)
!***********************************************************************
  implicit none

  double precision, dimension(:,:), intent(inout) :: r, fi
  double precision, dimension(:), intent(in) :: ak
  double precision, dimension(:), intent(inout) :: ei
  double precision, dimension(:,:), intent(inout) :: pta
  double precision, dimension(3) :: xm, ym, zm, bs
  double precision, dimension(4,3) :: fd
  double precision :: d12, d23, d13, fg, x, ud, u0, dphi, dsq, sintsq
  integer, dimension(:,:), intent(in) :: ndalst

  integer, intent(in) :: ns, nmf
  integer :: i, im, j, k, ism, i1, i2, i3, i4, i5, ii, ndim

  ndim = size(r,2)

  ism = -ns ! initialize site offset so that it starts at zero

! Calculate dihedral forces

  do im = 1, nmf

   ism = ism + ns   ! calculate site offset for molecule im, start at zero

   do k = 1, size(ndalst,1)

    i1 = ndalst(k,1) + ism
    i2 = ndalst(k,2) + ism
    i3 = ndalst(k,3) + ism
    i4 = ndalst(k,4) + ism

    xm(1) = r(i2,1) - r(i1,1)
    ym(1) = r(i2,2) - r(i1,2)
    zm(1) = r(i2,3) - r(i1,3)
    bs(1) = xm(1)*xm(1) + ym(1)*ym(1) + zm(1)*zm(1)
    xm(2) = r(i3,1) - r(i2,1)
    ym(2) = r(i3,2) - r(i2,2)
    zm(2) = r(i3,3) - r(i2,3)
    bs(2) = xm(2)*xm(2) + ym(2)*ym(2) + zm(2)*zm(2)
    xm(3) = r(i4,1) - r(i3,1)
    ym(3) = r(i4,2) - r(i3,2)
    zm(3) = r(i4,3) - r(i3,3)
    bs(3) = xm(3)*xm(3) + ym(3)*ym(3) + zm(3)*zm(3)

    d12 = xm(1)*xm(2) + ym(1)*ym(2) + zm(1)*zm(2)
    d23 = xm(2)*xm(3) + ym(2)*ym(3) + zm(2)*zm(3)
    d13 = xm(1)*xm(3) + ym(1)*ym(3) + zm(1)*zm(3)

    fg = (bs(1)*bs(2) - d12*d12)*(bs(2)*bs(3) - d23*d23)
    x = -(d12*d23 - d13*bs(2))/sqrt(fg)
    if(abs(x).gt.1.d0) x = sign(1.d0,x)

!  ryckaert-bellemans potential energy function
    u0 = ak(1) + x*(ak(2) + x*(ak(3) + x*(ak(4) + x*(ak(5) + ak(6)*x))))
    dphi = ak(2) + x*(2.d0*ak(3) + x*(3.d0*ak(4) + x*(4.d0*ak(5) + 5.d0*ak(6)*x)))
    dphi = dphi/(dsq*sintsq)

!  dihedral forces
    fd(1,1) = dphi*xm(3)
    fd(1,2) = dphi*ym(3)
    fd(1,3) = dphi*zm(3)
    fd(2,1) = -fd(1,1)
    fd(2,2) = -fd(1,2)
    fd(2,3) = -fd(1,3)
    fd(3,1) = dphi*xm(1)
    fd(3,2) = dphi*ym(1)
    fd(3,3) = dphi*zm(1)
    fd(4,1) = -fd(3,1)
    fd(4,2) = -fd(3,2)
    fd(4,3) = -fd(3,3)

    do ii = 1, 4
     i5 = ndalst(k,ii) + ism
     fi(i5,1) = fi(i5,1) + fd(ii,1)
     fi(i5,2) = fi(i5,2) + fd(ii,2)
     fi(i5,3) = fi(i5,3) + fd(ii,3)
     ei(i5) = u0*0.25d0
     do j = 1, ndim
      do i = 1, ndim
       pta(j,i) = pta(j,i) + r(i5,j)*fd(ii,i)
      enddo
     enddo
    enddo

    ud = ud + u0

   enddo
  enddo

  end subroutine dihedral_force_energy
!***********************************************************************
  subroutine init_imask_dihed(da, md)
! Dihedral interacting sites do not interact via a LJ potential.
! Set the interaction mask to false for pairs of sites that interact
! via a dihedral potential.
! Assume that sites connected by a bond or a bond angle constraint are
! already masked out. This procedure only masks out LJ interactions between
! sites 1 and 4 for each dihedral interaction.
!***********************************************************************
  implicit none

  type (dihed_set) :: da
  type(mol_set), intent(inout) :: md
  integer :: j, k, ii, ij

! new for fraction masking
  do j = 1, md%nf
   do k = 1, da%dset(j)%nda
    ii = da%dset(j)%ndalst(k,1)
    ij = da%dset(j)%ndalst(k,4)
    md%frac_mask(j)%fmask(ii,ij) = .false.
    md%frac_mask(j)%fmask(ij,ii) = .false.
   enddo
  enddo

  end subroutine init_imask_dihed
!***********************************************************************
  subroutine create_dihedral_set(da, nf)
!***********************************************************************
  implicit none

  type (dihed_set), intent(inout) :: da
  integer, intent(in) :: nf
  integer :: error

  allocate(da%dset(nf),stat = error)
  if(error .ne. 0) stop 'allocation error in create_dihedral_set'

  allocate(da%dc(6),stat = error)
  if(error .ne. 0) stop 'allocation error in create_dihedral_set'

  end subroutine create_dihedral_set
!***********************************************************************
  subroutine create_dihed(da, md)
!***********************************************************************
  implicit none

  type (dihed_set), intent(inout) :: da
  type(mol_set), intent(in) :: md
  integer :: i, error, nda

   do i = 1, md%nf
    nda = da%dset(i)%nda
    if(nda .eq. 0) return
    allocate(da%dset(i)%ndalst(nda,4),stat = error)
    if(error .ne. 0) stop 'allocation error in create_dihed'
   enddo

  end subroutine create_dihed
!***********************************************************************
  subroutine destroy_dihed(da)
!***********************************************************************
  implicit none

  type (dihed_set), intent(inout) :: da
  integer :: i

  do i = 1, size(da%dset)
   if(allocated(da%dset(i)%ndalst)) deallocate(da%dset(i)%ndalst)
  enddo

  if(allocated(da%dset)) deallocate(da%dset)
  if(allocated(da%dc)) deallocate(da%dc)

  end subroutine destroy_dihed
!***********************************************************************
  subroutine rd_dihed_input(da,md,header,ifile,iunit)
!***********************************************************************
! Reads dihedral potential variables from input file.
! 
!***********************************************************************
  implicit none

  type (dihed_set), intent(inout) :: da
  type(mol_set), intent(in) :: md
  
  integer :: i, j, k, iunit
  
  character(len=*) :: header
  character(len=32) :: rdstring
  character(len=32) :: ifile        ! input file names

  rdstring = ''
  i = 0
  open(iunit,file=ifile,status='old',form='formatted', &
     & position='rewind')
  do j = 1, 1000
   read(iunit,*, iostat=i) rdstring
   if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 999) exit
  enddo

  if (rdstring .eq. header) then
   read(iunit,*) da%dfarch           ! read the molecular architecture
   if(da%dfarch .eq. 0)then
    close(iunit)
    return        ! no dihedral forces - return
   endif
   call create_dihedral_set(da, md%nf) 
   read(iunit,*) (da%dset(i)%nda, i=1, md%nf) ! number of dihed. angles
   read(iunit,*) (da%dc(i), i=1, size(da%dc)) ! dihedral force constants
   read(iunit,*) da%dai ! initial dihedral angle when making molecules

   call create_dihed(da, md)

   do k = 1, md%nf
    select case(da%dfarch)
    case(1)            ! all linear alkanes
     call init_dihed_linalk(da, md)
    case(2)            ! arbitrary molecular architecture
     do j = 1, da%dset(i)%nda
      read(iunit,*) (da%dset(i)%ndalst(j,i), i=1, 4)
     enddo
    case default
     stop 'Invalid molecular architecture in dihedral forces'
    end select
   enddo

   close(iunit)
  else
   close(iunit)
   stop 'Dihedral angle input data not found - stop'
  endif

  end subroutine rd_dihed_input
!***********************************************************************
  subroutine init_dihed_linalk(da, md)
!***********************************************************************
  implicit none
 
  type (dihed_set), intent(inout) :: da
  type(mol_set), intent(in) :: md
  integer :: i, j

! Simple dihedral matrix initialization for linear alkane molecules

  do i = 1, md%nf
   if(da%dset(i)%nda .ne. md%nsf(i)-3) stop 'error in number of dihedral angles'
   da%dset(i)%ndalst = 0
   do j = 1, (md%nsf(i)-3)
    da%dset(i)%ndalst(j,1) = j
    da%dset(i)%ndalst(j,2) = j+1
    da%dset(i)%ndalst(j,3) = j+2
    da%dset(i)%ndalst(j,4) = j+3
   enddo
  enddo

  end subroutine init_dihed_linalk
!***********************************************************************
  subroutine wr_dihed(da,ofile,ounit)
! writes input data to the output file
!***********************************************************************
  type (dihed_set), intent(in) :: da
  integer :: ounit
  character(len=32) :: ofile      !output file name

  open(ounit,file=ofile,status='old',form='formatted',position='append')
  write (ounit,*) 'Initial dihedral force values'
  write (ounit,*)
  write (ounit,*) 'dihedral force type       = ', da%dfarch
  write (ounit,*)
  close(ounit)
  
  end subroutine wr_dihed
!***********************************************************************
  end module dihedral_force
