 module ewald
!***********************************************************************
!23456789012345678901234567890123456789012345678901234567890123456789012
!
! module for calculating long range forces via ewald summations
!
! This module must be used in conjunction with the pair force module
! which calculates the real space part of the electrostatic interaction.
! References:
! [1] "The Liquid State: applications of molecular simulations"
! D.M. Heyes (Wiley, 1998).
! [2] D. R. Wheeler, N. G. Fuller and R. L. Rowley
! Molecular Physics, vol 92, pp 55-62 (1997).
!***********************************************************************
 use species
 use molecules
 use atoms
 use v_sub_library
 use simulation_box

 implicit none

 private

! public procedures
 public :: create_ewald_kvec, destroy_ewald_kvec, ewald_fourier, &
 & ewald_self_bonded, init_ewald, init_ewald_kvec, &
 & init_ewald_self_bond, rd_ewald, rd_dsf

 public :: ewald_data_type
 
! uiself : self energy correction for each charged particle
! kappa : constant in exponential
! kmax : maximum lattice vector
! nk : total number of lattice vectors
! kvec : lattice vector
! ksq : wavevector magnitude squared
! iq : indices of all charged particles

 type :: ewald_data_type
  double precision, dimension(:,:), allocatable :: kvec
  double precision, dimension(:), allocatable :: ksq, uiself, q
  double precision :: kappa, uself, ufour, usurf, usb, kmax, kapsq, c1
  double precision :: pfour, psb
  integer, dimension(:), allocatable :: iq, ic
  integer, dimension(:,:), allocatable :: ib
  integer :: nk, nc, nb
  logical :: print_diagnostics
 end type ewald_data_type
 
 double precision :: pi

 contains
!***********************************************************************
  subroutine create_ewald_kvec(ew, m, d)
!***********************************************************************
  implicit none
  type (ewald_data_type) :: ew 
  integer :: error, d, m

  call destroy_ewald_kvec(ew)
  allocate(ew%kvec(m,d), ew%ksq(m), stat = error)
  if(error .ne. 0) stop 'ewald.create_ewald_kvec: allocation error'

  end subroutine create_ewald_kvec
!***********************************************************************
  subroutine destroy_ewald_kvec(ew)
!***********************************************************************
  implicit none
  type (ewald_data_type) :: ew 

  if(allocated(ew%kvec)) deallocate(ew%kvec)
  if(allocated(ew%ksq)) deallocate(ew%ksq)

  end subroutine destroy_ewald_kvec
!***********************************************************************
  subroutine ewald_fourier(ewd, spd, atd, md, box, pta, ptm, jqm)
! calculates the fourier space term of the ewald summation
! Note that symmetry of energy and force with respect to inversion of k
! vector (k => -k) is used to reduce the number of wavevectors used
! by half.
! Note also that our ui is one half of Heyes's expression, so that
! we have utotal = sum(ui) instead of (1/2)sum(ui) as in Heyes.
! Note: use saved allocatable arrays for temporaries so that memory is
! only allocated once.
!***********************************************************************
   implicit none
   type(ewald_data_type) :: ewd
   type(species_set) :: spd
   type(atom_type) :: atd
   type(mol_set) :: md
   type (box_data_type) :: box
   double precision, allocatable, save, dimension(:,:) :: qckr, qskr, tmp1
   double precision, allocatable, save, dimension(:) :: sumqckr, sumqskr, ck, uk
   double precision, allocatable, save, dimension(:,:) :: sumqvckr, sumqvskr
   double precision, allocatable, save, dimension(:,:) :: t1, t2, ak
   double precision, allocatable, save, dimension(:) :: uifour, t3
   double precision, dimension(:,:) :: pta, ptm
   double precision, dimension(size(atd%r,2),size(atd%r,2)) :: ptafour
   double precision, dimension(size(atd%r,2),size(atd%r,2)) :: delta
   double precision, dimension(size(atd%r,2)) :: jqm, jqafour
   double precision :: tmp, volume, c1, bij, tmpr, tmpi
!   double complex :: sqeikr
   integer :: i, j, k, ndim, i1, error

   ndim = size(atd%r,2)

   if(.not. allocated(qckr)) then
    allocate(qckr(ewd%nc,ewd%nk),qskr(ewd%nc,ewd%nk),tmp1(ewd%nc,ewd%nk),sumqckr(ewd%nk), &
    & sumqskr(ewd%nk), sumqvckr(ewd%nk,ndim), sumqvskr(ewd%nk,ndim), ck(ewd%nk), &
    & uk(ewd%nk), t1(ewd%nc,size(atd%r,2)), t2(ewd%nc,size(atd%r,2)), t3(ewd%nc), &
    & ak(ewd%nk,ndim), uifour(ewd%nc), stat=error)
    if(error .ne. 0) stop 'allocation error in ewald_fourier'
   elseif(size(qckr,2) .lt. ewd%nk) then
    deallocate(qckr,qskr,sumqckr,sumqskr,sumqvckr,sumqvskr,ck,uk)
    allocate(qckr(ewd%nc,ewd%nk),qskr(ewd%nc,ewd%nk),tmp1(ewd%nc,ewd%nk),sumqckr(ewd%nk), &
    & sumqskr(ewd%nk),sumqvckr(ewd%nk,ndim),sumqvskr(ewd%nk,ndim),ck(ewd%nk),uk(ewd%nk), &
    & stat=error)
    if(error .ne. 0) stop 'allocation error in ewald_fourier'
   endif

   ptafour = 0.d0
   jqafour = 0.d0
   volume = box%volume
   c1 = 4.d0*pi/volume

   delta = 0.d0
   do i = 1, size(delta,1)
    delta(i,i) = 1.d0
   enddo

   t1 = atd%r(ewd%iq,:)

   do i = 1, ndim
    t2(:,i) = atd%p(ewd%iq,i)/spd%mass(spd%spid(ewd%iq))
   enddo

! calculate the array of k_n.r_i values

   tmp1 = 0.d0
   do k = 1, ndim
    do j = 1, ewd%nk
     do i = 1, ewd%nc
      tmp1(i,j) = tmp1(i,j) + t1(i,k)*ewd%kvec(j,k)
     enddo
    enddo
   enddo

   do i = 1, ewd%nk
    qckr(:,i) = ewd%q*cos(tmp1(:,i))
    qskr(:,i) = ewd%q*sin(tmp1(:,i))
   enddo

   do j = 1, ewd%nk
    sumqckr(j) = sum(qckr(:,j))
    sumqskr(j) = sum(qskr(:,j))
   enddo

   do i = 1, ndim
    do j = 1, ewd%nk
     sumqvckr(j,i) = sum(t2(:,i)*qckr(:,j))
     sumqvskr(j,i) = sum(t2(:,i)*qskr(:,j))
    enddo
   enddo

   ck = c1*exp(-ewd%ksq/(4.d0*ewd%kapsq))/ewd%ksq
   
!   if(ewd%print_diagnostics) then
!   print *, ewd%nk
!   uk = 0.d0
!    do j = 1, ewd%nk
!    sqeikr = (0.d0, 0.d0)
!     do i = 1, ewd%nc
!      sqeikr = sqeikr + ewd%q(i)*exp(cmplx(0.d0, tmp1(i,j)))
!     enddo
!     uk(j) = ck(j)*(abs(sqeikr))**2
!     print *, j, uk(j)  
!    enddo
!   endif

   uifour = 0.d0
   t1 = 0.d0
   t3 = 0.d0
   uk = 0.d0

   do j = 1, ewd%nk
    do i = 1, ewd%nc
     tmpr = qckr(i,j)*sumqckr(j) + qskr(i,j)*sumqskr(j)
     tmp = ck(j)*tmpr
     uifour(i) = uifour(i) + tmp
     uk(j) = uk(j) + tmp
    enddo
   enddo

!   if(ewd%print_diagnostics) then
!    do j = 1, ewd%nk
!     print *, j, uk(j)
!    enddo
!   endif

   do k = 1, ndim
    do j = 1, ewd%nk
     do i = 1, ewd%nc
      tmpi = qckr(i,j)*sumqskr(j) - qskr(i,j)*sumqckr(j)
      t1(i,k) = t1(i,k) - 2.d0*ewd%kvec(j,k)*ck(j)*tmpi
      t3(i) = t3(i) - 2.d0*ewd%kvec(j,k)*ck(j)*(qckr(i,j)*sumqvskr(j,k) - qskr(i,j)*sumqvckr(j,k))
     enddo
    enddo
   enddo

   do k = 1, ndim
    do j = 1, ewd%nk
     ak(j,k) = sumqckr(j)*sumqvckr(j,k) + sumqskr(j)*sumqvskr(j,k)
    enddo
   enddo

! Fourier space contribution to the atomic pressure tensor
! Note: make use of the fact that this is a symmetric tensor
   do i = 1, ndim
    do j = i, ndim
     do k = 1, ewd%nk
      bij = delta(i,j) - 2.d0*ewd%kvec(k,i)*ewd%kvec(k,j)*(1.d0/ewd%ksq(k) + 1.d0/(4.d0*ewd%kapsq))
      ptafour(i,j) = ptafour(i,j) + bij*uk(k)
      jqafour(i) = jqafour(i) + bij*ck(k)*ak(k,j)
     enddo
    enddo
   enddo
! Use symmetry to get the other elements
   do i = 2, ndim
    do j = 1, i-1
     ptafour(i,j) = ptafour(j,i)
    enddo
   enddo

   ewd%ufour = sum(uifour)

   ewd%pfour = tr_div_dim(ptafour)
   pta = pta + ptafour
   ptm = ptm + ptafour
   jqm = jqm + jqafour

! Accumulate internal energy and forces into the atom data
   do i = 1, ewd%nc
    atd%eint(ewd%iq(i)) = atd%eint(ewd%iq(i)) + uifour(i) + ewd%uiself(i)
   enddo

   do k = 1, ndim
    do i = 1, ewd%nc
     atd%f(ewd%iq(i),k) = atd%f(ewd%iq(i),k) + t1(i,k)
    enddo
   enddo

! Add atomic to molecular conversion terms
   do k = 1, ndim
    do i1 = 1, ndim
     do i = 1, ewd%nc
      ptm(k,i1) = ptm(k,i1) - md%ricm(ewd%iq(i),i1)*t1(i,k)
     enddo
    enddo
   enddo

   do k = 1, ndim
    do i = 1, ewd%nc
     jqm(k) = jqm(k) - 0.5d0*md%ricm(ewd%iq(i),k)*t3(i)
    enddo
   enddo

  end subroutine ewald_fourier
!***********************************************************************
  subroutine ewald_self_bonded(ewd, atd, spd, pta)
! Calculates the correction term of the ewald summation to compensate for
! the inclusion of electrostatic interactions between bonded pairs
! of charged atoms in the fourier space part.
! Reference: "The Liquid State: applications of molecular simulations"
! D.M. Heyes (Wiley, 1998).
!***********************************************************************
  implicit none
  type(ewald_data_type) :: ewd
  type(species_set) :: spd
  type(atom_type) :: atd
  double precision, allocatable, save, dimension(:,:) :: rij, fijsb
  double precision, allocatable, save, dimension(:) :: sij, rijsq, uisb
  double precision, dimension(:,:) :: pta
  double precision, dimension(size(atd%r,2),size(atd%r,2)) :: ptaewsb
  double precision :: qiqj
  integer :: i, j, k, ndim, i1, j1, error
   
  ndim = size(atd%r,2)
  ptaewsb = 0.d0

  if(.not. allocated(rij)) then
   allocate(rij(ewd%nb,ndim),fijsb(ewd%nb,ndim),sij(ewd%nb), &
   & rijsq(ewd%nb), uisb(ewd%nb), stat=error)
   if(error .ne. 0) stop 'allocation error in ewald_self_bonded'
  endif

  rijsq = 0.d0
! calculate pair separations
  do j = 1, ndim
   do i = 1, ewd%nb
    rij(i,j) = atd%r(ewd%ib(i,2),j) - atd%r(ewd%ib(i,1),j)
    rijsq(i) = rijsq(i) + rij(i,j)*rij(i,j)
   enddo
  enddo
  sij = sqrt(rijsq)

! calculate the pair potential energies and forces
  do i = 1, ewd%nb
   i1 = spd%spid(ewd%ib(i,1))
   j1 = spd%spid(ewd%ib(i,2))
   qiqj = spd%charge(i1)*spd%charge(j1)
!   print *, ewd%ib(i,1), ewd%ib(i,2)
   uisb(i) = -qiqj*erf(ewd%kappa*sij(i))/sij(i)
!   print *, qiqj, ewd%kappa, sij(i), erf(ewd%kappa*sij(i)), uisb(i)
   do j = 1, ndim
    fijsb(i,j) = rij(i,j)*(-uisb(i) - qiqj*ewd%c1*exp(-ewd%kapsq*rijsq(i)))/rijsq(i)
   enddo
  enddo

! total bond correction internal energy
  ewd%usb = sum(uisb)

! Accumulate bond correction energy and force corrections into the atom data
  do i = 1, ewd%nb
   i1 = ewd%ib(i,1)
   j1 = ewd%ib(i,2)
   atd%eint(i1) = atd%eint(i1) + 0.5*uisb(i)
   atd%eint(j1) = atd%eint(j1) + 0.5*uisb(i)
   do j = 1, ndim
    atd%f(i1,j) = atd%f(i1,j) + fijsb(i,j)
    atd%f(j1,j) = atd%f(j1,j) - fijsb(i,j)
    do k = 1, ndim
     ptaewsb(k,j) = ptaewsb(k,j) - rij(i,k)*fijsb(i,j)
    enddo
   enddo
  enddo

  pta = pta + ptaewsb
!  print *, 'ptaewsb = '
!  print *, ptaewsb
  ewd%psb = tr_div_dim(ptaewsb)
!  print *, 'psb = ', ewd%psb/box%volume

  end subroutine ewald_self_bonded
!***********************************************************************
 subroutine init_ewald(spd, ew)
! creates arrays of charged atom indices and atom charges
! Also calculates the electrostatic self-interaction potential energy.
!***********************************************************************
  implicit none
  type(species_set) :: spd
  type(ewald_data_type) :: ew
  integer :: i, j, k, nc, ierr
  double precision :: sqtpii, pisqr
  
! Note: pi is a module variable, available to all module procedures
  pi = 4.d0*atan(1.d0)
  sqtpii = 1.d0/sqrt(pi)
  pisqr = pi*pi
  ew%kapsq = ew%kappa*ew%kappa
  ew%c1 = 2.d0*ew%kappa/sqrt(pi)

! count the number of charges in the system
  nc = 0
  do i = 1, spd%nsp
   if(spd%has_charge(i)) nc = nc + spd%nasp(i)
  enddo
  ew%nc = nc

! allocate the array to hold the indices
! of charged particles
  allocate(ew%iq(nc), ew%q(nc), ew%uiself(nc), ew%ic(spd%np), stat=ierr)
  if(ierr .ne. 0) stop 'ewald.init_ewald: allocation error'
  ew%ic = 0
  ew%uiself = 0.d0

! initialize the charged particle indexing array and calculate the
! electrostatic self-potential energy correction
  j = 0
  do i = 1, spd%np
   k = spd%spid(i)
   if(spd%has_charge(k)) then
    j = j + 1
    ew%iq(j) = i
    ew%ic(i) = j
    ew%q(j) = spd%charge(k)
    ew%uiself(j) = -ew%kappa*sqtpii*ew%q(j)*ew%q(j)
   endif
  enddo

  ew%uself = sum(ew%uiself) 

  end subroutine init_ewald
!***********************************************************************
  subroutine init_ewald_kvec(boxveci, ewd) 
! calculates the reciprocal vectors 
!***********************************************************************
  implicit none
  type(ewald_data_type) :: ewd
  double precision, dimension(:,:) :: boxveci
  double precision, dimension(3) :: kvec
  double precision, dimension(:,:), allocatable :: ktmp
  double precision, dimension(:), allocatable :: ksqtmp
  double precision :: tmp1, tmp2, twopi, pisqr, ksq, ksqmax
  integer :: ndim, nkmax, nmax, nk, n1, n2, n3, i, error
  
  twopi = 2.d0*pi
  pisqr = pi*pi
  ndim = size(boxveci,1)
  ksqmax = ewd%kmax*ewd%kmax
  
! find the reciprocal lattice vector with the smallest squared magnitude
  tmp1 = dot_product(boxveci(1,:),boxveci(1,:))
  do i = 2, ndim
   tmp2 = dot_product(boxveci(i,:),boxveci(i,:))
   if(tmp2 .lt. tmp1) tmp1 = tmp2
  enddo
  tmp1 = twopi*sqrt(tmp1)

! calculate nmax from the smallest reciprocal lattice vector
! beware: integer result intentionally used to truncate the value
  nmax = ewd%kmax/tmp1
  nkmax = ((2.0*nmax+1.0)**3 - 1.0)/2.0
  
  allocate(ktmp(nkmax,3), ksqtmp(nkmax), stat=error)
  if(error .ne. 0) stop 'ewald.init_ewald_kvec: allocation error'

! find all wavevectors with smaller squared magnitude than ksqmax
! only half of lattice vectors generated, since the results are
! invariant on k -> -k
  ktmp = 0.d0
  ksqtmp = 0.d0

  tmp1 = 0.d0
  tmp2 = 0.d0
  nk = 0
 
  n1 = 0
  n2 = 0
  do n3 = 1, nmax
   call calc_kvec
  enddo

  n1 = 0
  do n2 = 1, nmax
   do n3 = -nmax, nmax
    call calc_kvec
   enddo
  enddo

  do n1 = 1, nmax
   do n2 = -nmax, nmax
    do n3 = -nmax, nmax
     call calc_kvec
    enddo
   enddo
  enddo

  if(.not. allocated(ewd%kvec) .or. (allocated(ewd%kvec) .and. size(ewd%kvec,1) .lt. nk)) then
   call create_ewald_kvec(ewd, nk, ndim)
  endif
  ewd%kvec = ktmp(1:nk,:)
  ewd%ksq = ksqtmp(1:nk)
  ewd%nk = nk
  if(nk .eq. 0) stop 'ewald.init_ewald_kvec: nk=0, kmax is too small'
  
  if(ewd%print_diagnostics) then
   print *, 'init_ewald_kvec diagnostics'
   print *, 'nk = ', nk
   print *, 'kmax = ', sqrt(maxval(ewd%ksq))
   print *, 'C(kmax) = ', exp(-maxval(ewd%ksq)/(4.d0*ewd%kapsq))/maxval(ewd%ksq)
!   do i = 1, nk
!    print *, i, ewd%kvec(i, 1), ewd%kvec(i, 2), ewd%kvec(i, 3), ewd%ksq(i)
!   enddo
  endif
  
  deallocate(ktmp, ksqtmp)

  return
!***********************************************************************
  contains
!***********************************************************************
  subroutine calc_kvec
! This is only correct for an orthogonal box
!***********************************************************************
  kvec(1) = twopi*(boxveci(1,1)*n1 + boxveci(1,2)*n2 + boxveci(1,3)*n3)
  kvec(2) = twopi*(boxveci(2,1)*n1 + boxveci(2,2)*n2 + boxveci(2,3)*n3)
  kvec(3) = twopi*(boxveci(3,1)*n1 + boxveci(3,2)*n2 + boxveci(3,3)*n3)
  ksq = kvec(1)*kvec(1) + kvec(2)*kvec(2) + kvec(3)*kvec(3)
  if(ksq .le. ksqmax) then
   nk = nk + 1
   ktmp(nk,:) = kvec
   ksqtmp(nk) = ksq
  endif
  end subroutine calc_kvec
!***********************************************************************
  end subroutine init_ewald_kvec
!***********************************************************************
 subroutine init_ewald_self_bond(spd, md, ewd)
! Creates arrays of pairs of charged atom indices for constrained,
! bonded atoms to allow calculations of the self-energy correction.
!***********************************************************************
  implicit none
  type(species_set) :: spd
  type(mol_set) :: md
  type(ewald_data_type) :: ewd
  integer :: i, i1, j, k, nb, ierr, im
  
! count the total number of constrained bonds between
! charged particles in the system
  nb = 0
  do i = 1, md%nf
   do j = 1, md%nsf(i)
    do k = j+1, md%nsf(i)
     if(spd%has_charge(md%fr(i)%spid(j)) .and. spd%has_charge(md%fr(i)%spid(k)) .and. .not. md%frac_mask(i)%fmask(j,k)) then
      nb = nb + md%nmf(i)
     endif
    enddo
   enddo
  enddo
  ewd%nb = nb

! allocate the array to hold the indices
! of charged particles
  allocate(ewd%ib(nb,2), stat=ierr)
  if(ierr .ne. 0) stop 'ewald.init_ewald_self_bond: allocation error'

  nb = 0
  im = 0
  do i = 1, md%nf
   do i1 = 1, md%nmf(i)
    im = im + 1
    do j = 1, md%nsf(i)
     do k = j+1, md%nsf(i)
      if(spd%has_charge(md%fr(i)%spid(j)) .and. spd%has_charge(md%fr(i)%spid(k)) .and. .not. md%frac_mask(i)%fmask(j,k)) then
       nb = nb + 1
       ewd%ib(nb,1) = md%fs(im) + j - 1
       ewd%ib(nb,2) = md%fs(im) + k - 1
!       print *, ewd%ib(nb,1), ewd%ib(nb,2)
      endif
     enddo
    enddo
   enddo
  enddo

  end subroutine init_ewald_self_bond
!***********************************************************************
  subroutine rd_ewald(ew, header, ifile, iunit)
! reads ewald variables from inout file
!***********************************************************************
   implicit none
   type (ewald_data_type) :: ew
   integer :: iunit, i, j
   character(len = *) :: header
   character(len = 32) :: rdstring, ifile

   rdstring  = ''
   i = 0
   open(iunit, file = ifile, status = 'old', form = 'formatted', &
   & position = 'rewind')
   do j = 1, 2000
    read(iunit,*,iostat=i) rdstring
    if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 1999) exit
   enddo

   if(rdstring .eq. header) then
    read(iunit, *) ew%kappa
    read(iunit, *) ew%kmax
    read(iunit, *) ew%print_diagnostics
    close(iunit)
   else
    close(iunit)
    stop 'ewald.rd_ewald: input data not found'
   endif 

  end subroutine rd_ewald
!***********************************************************************
  subroutine rd_dsf(ew, header, ifile, iunit)
! reads damped and shifted force kappa variable from input file
!***********************************************************************
   implicit none
   type (ewald_data_type) :: ew
   integer :: iunit, i, j
   character(len = *) :: header
   character(len = 32) :: rdstring, ifile

   rdstring  = ''
   i = 0
   open(iunit, file = ifile, status = 'old', form = 'formatted', &
   & position = 'rewind')
   do j = 1, 2000
    read(iunit,*,iostat=i) rdstring
    if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 1999) exit
   enddo

   if(rdstring .eq. header) then
    read(iunit, *) ew%kappa
    close(iunit)
   else
    close(iunit)
    stop 'ewald.rd_dsf: input data not found'
   endif 

  end subroutine rd_dsf
!***********************************************************************
 end module ewald
