  module gear5
!***********************************************************************
! Four value (fifth order) Gear predictor corrector algorithm
!23456789012345678901234567890123456789012345678901234567890123456789012
 implicit none
 
 private

! Public procedures
 public :: predictor, corrector, gear5_pbc_correction, &
         & init_gear5_scalar, rd_gear5_input, &
         & wr_gear5, rd_gear5_state, store_gear5_state, &
         & create_gear5_array, destroy_gear5_array

! Public data types
 public :: gear5_data_type, gear5_array_type, gear5_scalar_type

 double precision, parameter :: &
      f01 = 251.0d0/720.0d0,  &
      f21 = 11.0d0/12.0d0,  &
      f31 = 1.0d0/3.0d0,  &
      f41 = 1.0d0/24.0d0

 interface predictor
  module procedure predictor_array
  module procedure predictor_scalar
 end interface

 interface corrector
  module procedure corrector_array
  module procedure corrector_scalar
 end interface

 interface rd_gear5_state
  module procedure rd_gear5_array_state
  module procedure rd_gear5_scalar_state
 end interface

 interface store_gear5_state
  module procedure store_gear5_array_state
  module procedure store_gear5_scalar_state
 end interface

 type gear5_data_type
  double precision :: tstep
 end type gear5_data_type

 type gear5_array_type
  double precision, dimension(:,:), allocatable :: d1,d2,d3,d4 ! derivatives
 end type gear5_array_type

 type gear5_scalar_type
  double precision :: d1,d2,d3,d4 ! derivatives
 end type gear5_scalar_type

 contains
!***********************************************************************
 subroutine create_gear5_array(g5a, n, d)
!***********************************************************************
 type (gear5_array_type), intent(out) :: g5a
 integer :: error, n, d

! allocate arrays for gear5 derivative arrays
  allocate(g5a%d1(n,d), g5a%d2(n,d),g5a%d3(n,d),g5a%d4(n,d), stat=error)
  if(error .ne. 0) stop 'allocation error in create_gear5_array'
  g5a%d1 = 0.d0
  g5a%d2 = 0.d0
  g5a%d3 = 0.d0
  g5a%d4 = 0.d0

  end subroutine create_gear5_array
!***********************************************************************
  subroutine destroy_gear5_array(g5a)
!***********************************************************************
  type (gear5_array_type) :: g5a

! deallocate arrays for gear5 derivative arrays
   if(allocated(g5a%d1)) deallocate(g5a%d1)
   if(allocated(g5a%d2)) deallocate(g5a%d2)
   if(allocated(g5a%d3)) deallocate(g5a%d3)
   if(allocated(g5a%d4)) deallocate(g5a%d4)

  end subroutine destroy_gear5_array
!***********************************************************************
  subroutine gear5_pbc_correction(g5a, tstep, rsh, sd, mask)
!***********************************************************************
  use sllod
  type (gear5_array_type) :: g5a
  type (sllod_data_type), intent(in) :: sd
  double precision, dimension(:,:) :: rsh
  double precision :: tstep
  logical, dimension(:) :: mask

  select case(sd%flow_type)
  case(1)           ! shear
  where(mask) &
   g5a%d1(:,1) = g5a%d1(:,1) - sd%strain_rate*rsh(:,2)*tstep
  case(2)             ! elongation
  where(mask) &
   g5a%d1(:,1) = g5a%d1(:,1) - sd%strain_rate*rsh(:,1)*tstep
  where(mask) &
   g5a%d1(:,2) = g5a%d1(:,2) + sd%strain_rate*rsh(:,2)*tstep
  case(3)             ! bulk
  where(mask) &
   g5a%d1(:,1) = g5a%d1(:,1) - sd%strain_rate*rsh(:,1)*tstep
  where(mask) &
   g5a%d1(:,2) = g5a%d1(:,2) - sd%strain_rate*rsh(:,2)*tstep
  where(mask) &
   g5a%d1(:,3) = g5a%d1(:,3) - sd%strain_rate*rsh(:,3)*tstep
  case default          ! do nothing
  end select

  end subroutine gear5_pbc_correction
!***********************************************************************
  subroutine init_gear5_scalar(g5s)
!***********************************************************************
  type (gear5_scalar_type), intent(inout) :: g5s

  g5s%d1 = 0.d0
  g5s%d2 = 0.d0
  g5s%d3 = 0.d0
  g5s%d4 = 0.d0

  end subroutine init_gear5_scalar
!***********************************************************************
  subroutine rd_gear5_input(g5,header,ifile,iunit)
!***********************************************************************
! Reads gear5 variables from input file.
! 
!***********************************************************************
  implicit none
  
  type (gear5_data_type) :: g5
  
  integer :: i, j, iunit
  
  character(len=*) :: header
  character(len=32) :: rdstring
  character(len=32) :: ifile      !input file name

  rdstring = ''
  i = 0
  open(iunit,file=ifile,status='old',form='formatted', &
     & position='rewind')
  do j = 1, 1000
   read(iunit,*, iostat=i) rdstring
   if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 999) exit
  enddo

  if (rdstring .eq. header) then
   read(iunit,*) g5%tstep
   close(iunit)
  else
   close(iunit)
   stop 'gear5 input data not found - stop'
  endif

  end subroutine rd_gear5_input
!***********************************************************************
  subroutine rd_gear5_array_state(a,iunit)
!***********************************************************************
  type (gear5_array_type), intent(inout) :: a
  integer, intent(in) :: iunit

  read(iunit) a%d1, a%d2, a%d3, a%d4
  
  end subroutine rd_gear5_array_state
!***********************************************************************
  subroutine store_gear5_array_state(a,iunit)
!***********************************************************************
  type (gear5_array_type), intent(in) :: a
  integer, intent(in) :: iunit

  write(iunit) a%d1, a%d2, a%d3, a%d4
  
  end subroutine store_gear5_array_state
!***********************************************************************
  subroutine rd_gear5_scalar_state(a,iunit)
!***********************************************************************
  type (gear5_scalar_type), intent(inout) :: a
  integer, intent(in) :: iunit

  read(iunit) a%d1, a%d2, a%d3, a%d4
  
  end subroutine rd_gear5_scalar_state
!***********************************************************************
  subroutine store_gear5_scalar_state(a,iunit)
!***********************************************************************
  type (gear5_scalar_type), intent(in) :: a
  integer, intent(in) :: iunit

  write(iunit) a%d1, a%d2, a%d3, a%d4
  
  end subroutine store_gear5_scalar_state
!***********************************************************************
  subroutine wr_gear5(g5,ofile,ounit)
! writes input data to the output file
!***********************************************************************
  type (gear5_data_type), intent(in) :: g5
  integer :: ounit
  character(len=32) :: ofile      !output file name

  open(ounit,file=ofile,status='old',form='formatted',position='append')
  write (ounit,*) 'Initial gear5 values'
  write (ounit,*)
  write (ounit,*) 'Input timestep            = ', g5%tstep
  write (ounit,*)
  close(ounit)
  
  end subroutine wr_gear5
!***********************************************************************
  subroutine predictor_array_2(r0,r1,r2,r3,r4,mask,delta)
!***********************************************************************
  double precision, dimension (:,:) :: r0,r1,r2,r3,r4
  double precision, dimension (:,:), optional :: delta
  double precision, dimension (size(r0,1),3) :: delr
  double precision, dimension (size(r0,1)) :: m1
  logical, dimension(:) :: mask
  integer :: i,j

  do j = 1, size(r0,1)
   if(mask(j)) then
    m1 = 1.d0
   else
    m1 = 0.d0
   endif
  enddo

  do i = 1, 3
  !dir$ vector always
   do j = 1, size(r0,1)
    delr(j,i) = (r1(j,i) + r2(j,i) + r3(j,i) + r4(j,i))*m1(j)
    r0(j,i) = r0(j,i) + delr(j,i)
    r1(j,i) = r1(j,i) + (2.d0*r2(j,i) + 3.d0*r3(j,i) + 4.d0*r4(j,i))*m1(j)
    r2(j,i) = r2(j,i) + (3.d0*r3(j,i) + 6.d0*r4(j,i))*m1(j)
    r3(j,i) = r3(j,i) + 4.d0*r4(j,i)*m1(j)
   enddo
  enddo

  if(present(delta))  delta = delr

  end subroutine predictor_array_2
!***********************************************************************
  subroutine predictor_array(r0,r1,r2,r3,r4,mask,delta)
!***********************************************************************
  double precision, dimension (:,:) :: r0,r1,r2,r3,r4
  double precision, dimension (:,:), optional :: delta
  double precision :: tmp
  logical, dimension(:) :: mask
  integer :: i,j

  do i = 1, size(r0,2)
   do j = 1, size(r0,1)
    if(.not. mask(j)) cycle
    tmp = r1(j,i) + r2(j,i) + r3(j,i) + r4(j,i)
    r0(j,i) = r0(j,i) + tmp
    r1(j,i) = r1(j,i) + 2.d0*r2(j,i) + 3.d0*r3(j,i) + 4.d0*r4(j,i)
    r2(j,i) = r2(j,i) + 3.d0*r3(j,i) + 6.d0*r4(j,i)
    r3(j,i) = r3(j,i) + 4.d0*r4(j,i)
    if(present(delta)) delta(j,i) = tmp
   enddo
  enddo
  
  end subroutine predictor_array
!***********************************************************************
!  subroutine predictor_array_1(r0,r1,r2,r3,r4,mask,delta)
!***********************************************************************
!  double precision, dimension (:,:) :: r0,r1,r2,r3,r4
!  double precision, dimension (:,:), optional :: delta
!  double precision, dimension (size(r0,1),size(r0,2)) :: tmp
!  logical, dimension(:) :: mask
!  integer :: i

!  do i = 1, size(r0,2)
!  where(mask)
!  tmp(:,i) = r1(:,i) + r2(:,i) + r3(:,i) + r4(:,i)
!  r0(:,i) = r0(:,i) + tmp(:,i)
!  r1(:,i) = r1(:,i) + 2.0*r2(:,i) + 3.0*r3(:,i) + 4.0*r4(:,i)
!  r2(:,i) = r2(:,i) + 3.0*r3(:,i) + 6.0*r4(:,i)
!  r3(:,i) = r3(:,i) + 4.0*r4(:,i)
!  endwhere
!  if(present(delta))  delta(:,i) = tmp(:,i)
!  if(present(delta)) where(mask) delta(:,i) = tmp(:,i)
!  enddo
  
!  end subroutine predictor_array_1
!***********************************************************************
  subroutine predictor_scalar(r0,r1,r2,r3,r4,delta)
!***********************************************************************
  double precision :: r0,r1,r2,r3,r4
  double precision, optional :: delta
  double precision :: tmp

  tmp = r1 + r2 + r3 + r4
  r0 = r0 + tmp
  r1 = r1 + 2.d0*r2 + 3.d0*r3 + 4.d0*r4
  r2 = r2 + 3.d0*r3 + 6.d0*r4
  r3 = r3 + 4.d0*r4

  if(present(delta)) delta = tmp

  end subroutine predictor_scalar
!***********************************************************************
  subroutine corrector_array(r0,r1,r2,r3,r4,eqm,deltat,mask,delta)
!***********************************************************************
  double precision, dimension (:,:) :: r0,r1,r2,r3,r4,eqm
  double precision, dimension (:,:), optional :: delta
  double precision :: rcor,tmp
  double precision :: deltat
  logical, dimension(:) :: mask
  integer :: i,j

  do i = 1, size(r0,2)
   do j = 1, size(r0,1)
    if(.not. mask(j)) cycle
    rcor = r1(j,i) - eqm(j,i)*deltat
    tmp = - rcor*f01
    r0(j,i) = r0(j,i) + tmp
    r1(j,i) = r1(j,i) - rcor
    r2(j,i) = r2(j,i) - rcor*f21
    r3(j,i) = r3(j,i) - rcor*f31
    r4(j,i) = r4(j,i) - rcor*f41
    if(present(delta))  delta(j,i) = tmp
   enddo
  enddo
  
  end subroutine corrector_array
!***********************************************************************
!  subroutine corrector_array(r0,r1,r2,r3,r4,eqm,deltat,mask,delta)
!***********************************************************************
!  double precision, dimension (:,:) :: r0,r1,r2,r3,r4,eqm
!  double precision, dimension (:,:), optional :: delta
!  double precision, dimension (size(r0,1),size(r0,2)) :: rcor,tmp
!  double precision :: deltat
!  logical, dimension(:) :: mask
!  integer :: i

!  do i = 1, size(r0,2)
!  where(mask)
!  rcor(:,i) = r1(:,i) - eqm(:,i)*deltat
!  tmp(:,i) = - rcor(:,i)*f01
!  r0(:,i) = r0(:,i) + tmp(:,i)
!  r1(:,i) = r1(:,i) - rcor(:,i)
!  r2(:,i) = r2(:,i) - rcor(:,i)*f21
!  r3(:,i) = r3(:,i) - rcor(:,i)*f31
!  r4(:,i) = r4(:,i) - rcor(:,i)*f41
!  endwhere
!  if(present(delta)) where(mask) delta(:,i) = tmp(:,i)
!  enddo
!  enddo
  
!  end subroutine corrector_array
!***********************************************************************
  subroutine corrector_scalar(r0,r1,r2,r3,r4,eqm,deltat,delta)
!***********************************************************************
  double precision :: r0,r1,r2,r3,r4,eqm
  double precision, optional :: delta
  double precision :: rcor,tmp
  double precision :: deltat

  rcor = r1 - eqm*deltat
  tmp = - rcor*f01
  r0 = r0 + tmp
  r1 = r1 - rcor
  r2 = r2 - rcor*f21
  r3 = r3 - rcor*f31
  r4 = r4 - rcor*f41

  if(present(delta)) delta = tmp

  end subroutine corrector_scalar
!***********************************************************************
  end module gear5
