  module integrator
!***********************************************************************
! Ordinary differential solvers for molecular dynamics
! destype = 1 Fifth order Gear predictor corrector
!         = 2 Fourth order Runge-Kutta
!
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************
 use gear5

 implicit none
 
 private

! Public procedures
 public :: integrator_input, create_integrator, destroy_integrator, &
           read_integrator, store_integrator

! Public data types
 public :: integrator_data_type


 type integrator_data_type
  double precision :: tstep
  integer :: destype
  type(gear5_array_type), pointer :: rder, pder
  type(gear5_scalar_type), pointer :: zder, eder, vder
 end type integrator_data_type

! Note:
! tstep=timestep
! destype=differential equation solver type (1=gear5, 2=rk4)
! rder=position gear5 predictor arrays
! pder=momentum gear5 predictor arrays
! zder=NH kinetic thermostat zeta gear5 predictor array OR
! eder=NH type barostat deformation rate predictor array
! vder=volume gear5 predictor array
 contains
!***********************************************************************
 subroutine create_integrator(ded, np, ndim)
!***********************************************************************
 type (integrator_data_type), intent(inout) :: ded
 integer :: np, ndim, error
 
 select case (ded%destype)
  case(1) ! gear5
   allocate(ded%rder, ded%pder, stat=error)
   if(error .ne. 0) stop 'allocation error in create_integrator'
   call create_gear5_array(ded%rder, np, ndim)
   call create_gear5_array(ded%pder, np, ndim)
   allocate(ded%zder, ded%eder, ded%vder, stat=error)
   if(error .ne. 0) stop 'allocation error in create_integrator'
  case(2) ! rk4
  end select

  end subroutine create_integrator
!***********************************************************************
  subroutine destroy_integrator(ded)
!***********************************************************************
  type (integrator_data_type) :: ded

! deallocate integrator objects
   if(associated(ded%rder)) then
    call destroy_gear5_array(ded%rder)
    call destroy_gear5_array(ded%pder)
    deallocate(ded%rder)
    deallocate(ded%pder)
    if(associated(ded%zder)) deallocate(ded%zder)
    if(associated(ded%eder)) deallocate(ded%eder)
    if(associated(ded%vder)) deallocate(ded%vder)
   endif

  end subroutine destroy_integrator
!***********************************************************************
  subroutine integrator_input(ded,header,ifile,iunit)
!***********************************************************************
! Reads integrator variables from input file.
! 
!***********************************************************************
  implicit none
  
  type (integrator_data_type) :: ded
  
  integer :: i, j, iunit
  
  character(len=*) :: header
  character(len=32) :: rdstring
  character(len=32) :: ifile      !input file name

  rdstring = ''
  i = 0
  open(iunit,file=ifile,status='old',form='formatted', &
     & position='rewind')
  do j = 1, 1000
   read(iunit,*, iostat=i) rdstring
   if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 999) exit
  enddo

  if (rdstring .eq. header) then
   read(iunit,*) ded%tstep
   read(iunit,*) ded%destype
   close(iunit)
  else
   close(iunit)
   stop 'integrator:integrator_input: integrator input data not found'
  endif

  end subroutine integrator_input
!***********************************************************************
  subroutine read_integrator(ded, iunit)
!***********************************************************************
  type (integrator_data_type) :: ded
  integer, intent(in) :: iunit

  if(ded%destype .eq. 1) then
   call rd_gear5_state(ded%rder,iunit)
   call rd_gear5_state(ded%pder,iunit)
   call rd_gear5_state(ded%zder,iunit)
   call rd_gear5_state(ded%eder,iunit)
   call rd_gear5_state(ded%vder,iunit)
  endif

  end subroutine read_integrator
!***********************************************************************
  subroutine store_integrator(ded, iunit)
!***********************************************************************
  type (integrator_data_type) :: ded
  integer, intent(in) :: iunit

  if(ded%destype .eq. 1) then
   call store_gear5_state(ded%rder,iunit)
   call store_gear5_state(ded%pder,iunit)
   call store_gear5_state(ded%zder,iunit)
   call store_gear5_state(ded%eder,iunit)
   call store_gear5_state(ded%vder,iunit)
  endif

  end subroutine store_integrator
!***********************************************************************
  end module integrator
