!***********************************************************************
   module lattice
!***********************************************************************
!  Module for lattice data and operations.
!  Includes lattice building.
!
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************

  implicit none

  private

  public :: destroy_lattice, &
          & init_lattice, make_lattice, rdinput_lattice,  &
          & wr_lattice

  public :: lattice_data_type

  type lattice_data_type
   integer :: ltype  ! lattice type
   integer :: npc  ! number of lattice points in unit cell
   integer :: nct  ! total number of unit cells in lattice
   integer, dimension(:), allocatable :: nuc  ! number of unit cells in each direction
   double precision :: vcell ! volume per unit cell/a^3
   double precision, dimension(:), allocatable :: lv ! lattice vector lengths
   double precision, dimension(:), allocatable :: lsc ! lattice vector scaling factors
   double precision, dimension(:), allocatable :: offset ! offset of origin for this lattice
   logical :: do_mix ! mix order of molecular positions in this lattice
  end type lattice_data_type

! Values of constants
! ltype:
! 0=cubic FCC, 1=rect. HCP                             : 3d types
! 2=square, 3=centred square                           : 2d types
! 4=HCP layers periodic in x,z but complete in y       : 3d hcp layers
  contains
!***********************************************************************
 subroutine destroy_lattice(lat)
!***********************************************************************
 implicit none
 type (lattice_data_type) :: lat
 integer :: error=0

! deallocate lattice data type arrays
  if(allocated(lat%nuc)) deallocate(lat%nuc, stat=error)
  if(error .ne. 0) stop 'deallocation error in destroy_lattice - stop'
  if(allocated(lat%lv)) deallocate(lat%lv, stat=error)
  if(error .ne. 0) stop 'deallocation error in destroy_lattice - stop'
  if(allocated(lat%lsc)) deallocate(lat%lsc, stat=error)
  if(error .ne. 0) stop 'deallocation error in destroy_lattice - stop'
  if(allocated(lat%offset)) deallocate(lat%offset, stat=error)
  if(error .ne. 0) stop 'deallocation error in destroy_lattice - stop'

  end subroutine destroy_lattice
!***********************************************************************
  subroutine init_lattice(lat)
! Initializes the number of unit cells and the number of lattice points
! per unit cell for a lattice of type determined by the value of ltype.
!***********************************************************************
  implicit none
  
  type(lattice_data_type) :: lat

  integer :: error, ndim

  ndim = size(lat%nuc)

  lat%nct = lat%nuc(1)*lat%nuc(2)
  if(ndim .eq. 3) lat%nct = lat%nct*lat%nuc(3)

  select case(lat%ltype)
  case (0)   ! cubic FCC
   allocate(lat%lv(3), lat%lsc(3),lat%offset(3), stat=error)
   if(error .ne. 0) stop 'allocation error in init_lattice - stop'
   lat%npc = 4
   lat%lv = (/ 1.0, 1.0, 1.0 /)
   lat%vcell = 1.0

  case (1)   ! rectangular unit cell hcp
   allocate(lat%lv(3), lat%lsc(3),lat%offset(3), stat=error)
   if(error .ne. 0) stop 'allocation error in init_lattice - stop'
   lat%npc = 4
   lat%lv = (/ 1.0, 2.0*sqrt(2.0/3.0), sqrt(3.0) /)
   lat%vcell = lat%lv(1)*lat%lv(2)*lat%lv(3)

  case (2)   ! square 2-d
   allocate(lat%lv(2), lat%lsc(2),lat%offset(2), stat=error)
   if(error .ne. 0) stop 'allocation error in init_lattice - stop'
   lat%npc = 1
   lat%lv = (/ 1.0, 1.0 /)
   lat%vcell = 1.0

  case (3)    ! centred square 2-d
   allocate(lat%lv(2), lat%lsc(2),lat%offset(2), stat=error)
   if(error .ne. 0) stop 'allocation error in init_lattice - stop'
   lat%npc = 2
   lat%lv = (/ 1.0, 1.0 /)
   lat%vcell = 1.0

  case (4)    ! hcp layers - rectangular unit cell
   allocate(lat%lv(3), lat%lsc(3),lat%offset(3), stat=error)
   if(error .ne. 0) stop 'allocation error in init_lattice - stop'
   lat%npc = 2
   lat%lv = (/ 1.0, 2.0*sqrt(2.0/3.0), sqrt(3.0) /) ! lv(2) is the hcp cell vector
   lat%vcell = 0.5*lat%lv(1)*lat%lv(2)*lat%lv(3) ! Note - volume per layer

  case default
   stop 'init_lattice: invalid lattice type input'
  end select

  end subroutine init_lattice
!***********************************************************************
  subroutine make_lattice(ictype, nc, nct, ilp, lsc, offset, ratom, do_mix)
! Makes a lattice of type determined by the value of ltype.
!***********************************************************************
  implicit none

  double precision, dimension(:,:) :: ratom
  integer, dimension(:) :: nc
  double precision, dimension(size(nc)) :: cvec
  double precision, dimension(:,:), allocatable :: lattice, unitcell
  double precision, dimension(:) :: lsc
  double precision, dimension(:) :: offset
  integer :: i, j, k, m, nct, ndim, ictype, na
  integer :: ilp, error
  logical :: do_mix

  ndim = size(nc)
  na = size(ratom,1)

  select case(ictype)
  case (0)   ! cubic FCC
   if(size(ratom,1) .gt. ilp*nct) &
    & stop 'make_lattice: Number of molecules must not be greater than 4*ncx*ncy*ncz for this lattice'
   allocate(lattice(ilp*nct, ndim), stat=error)
   if(error .ne. 0) stop 'allocation error in make_lattice'
   allocate(unitcell(ilp, ndim))
   if(error .ne. 0) stop 'allocation error in make_lattice'

   unitcell(1,:) = (/0.0, 0.0, 0.0/)
   unitcell(2,:) = (/0.5, 0.0, 0.5/)
   unitcell(3,:) = (/0.0, 0.5, 0.5/)
   unitcell(4,:) = (/0.5, 0.5, 0.0/)

   m = 0
   do k = 0, nc(3)-1
    do j = 0, nc(2)-1
     do i = 0, nc(1)-1
      cvec(1) = i
      cvec(2) = j
      cvec(3) = k
      lattice(ilp*m+1,:) = unitcell(1,:) + cvec
      lattice(ilp*m+2,:) = unitcell(2,:) + cvec
      lattice(ilp*m+3,:) = unitcell(3,:) + cvec
      lattice(ilp*m+4,:) = unitcell(4,:) + cvec
      m = m+1
     enddo
    enddo
   enddo

  case (1)   ! rectangular unit cell hcp
   if(size(ratom,1) .gt. ilp*nct) &
   & stop 'make_lattice: Number of molecules must not be greater than 4*ncx*ncy*ncz for this lattice'
   allocate(lattice(ilp*nct, ndim), stat=error)
   if(error .ne. 0) stop 'allocation error in make_lattice'
   allocate(unitcell(ilp, ndim), stat=error)
   if(error .ne. 0) stop 'allocation error in make_lattice'

   unitcell(1,:) = (/ 0.0, 0.0, 0.0 /)
   unitcell(2,:) = (/ 0.5, 0.0, 0.5*sqrt(3.0) /)
   unitcell(3,:) = (/ 0.5, sqrt(2.0/3.0), sqrt(3.0)/6.0 /)
   unitcell(4,:) = (/ 1.0, sqrt(2.0/3.0), 2.0*sqrt(3.0)/3.0 /)

   m = 0
   do k = 0, nc(3)-1
    do j = 0, nc(2)-1
     do i = 0, nc(1)-1
      cvec(1) = i
      cvec(2) = j*2.0*sqrt(2.0/3.0)
      cvec(3) = k*sqrt(3.0)
      lattice(ilp*m+1,:) = unitcell(1,:) + cvec
      lattice(ilp*m+2,:) = unitcell(2,:) + cvec
      lattice(ilp*m+3,:) = unitcell(3,:) + cvec
      lattice(ilp*m+4,:) = unitcell(4,:) + cvec
      m = m+1
     enddo
    enddo
   enddo

  case (2)   ! square 2-d
   if(size(ratom,1) .gt. ilp*nct) &
   & stop 'make_lattice: Number of molecules must not be greater than (integer)^3 for this lattice'
   allocate(lattice(ilp*nct, ndim), stat=error)
   if(error .ne. 0) stop 'allocation error in make_lattice'
   allocate(unitcell(ilp, ndim), stat=error)
   if(error .ne. 0) stop 'allocation error in make_lattice'

   unitcell(1,:) = (/0.0, 0.0/)

   m = 0
   do j = 0, nc(2)-1
    do i = 0, nc(1)-1
     cvec(1) = i
     cvec(2) = j
     do k = 1, ilp
      lattice(ilp*m+1,:) = unitcell(k,:) + cvec
     enddo
     m = m+1
    enddo
   enddo

  case (3)    ! centred square 2-d
   if(size(ratom,1) .gt. ilp*nct) &
   & stop 'make_lattice: Number of molecules must not be greater than 2*(integer)^3 for this lattice'
   allocate(lattice(ilp*nct, ndim), stat=error)
   if(error .ne. 0) stop 'allocation error in make_lattice'
   allocate(unitcell(ilp, ndim), stat=error)
   if(error .ne. 0) stop 'allocation error in make_lattice'

   unitcell(1,:) = (/0.0, 0.0/)
   unitcell(2,:) = (/0.5, 0.5/)

   m = 0
   do j = 0, nc(2)-1
    do i = 0, nc(1)-1
     cvec(1) = i
     cvec(2) = j
     do k = 1, ilp
      lattice(ilp*m+1,:) = unitcell(k,:) + cvec
     enddo
     m = m+1
    enddo
   enddo

  case (4)    ! hcp layers - rectangular unit cell
   if(size(ratom,1) .gt. ilp*nct) &
   & stop 'make_lattice: Number of molecules must not be greater than 2*ncx*ncy*ncz for this lattice'
   allocate(lattice(ilp*nct, ndim), stat=error)
   if(error .ne. 0) stop 'allocation error in make_lattice'
   allocate(unitcell(ilp, ndim), stat=error)
   if(error .ne. 0) stop 'allocation error in make_lattice'

   unitcell(1,:) = (/ 0.0, 0.0, 0.0 /)
   unitcell(2,:) = (/ 0.5, 0.0, 0.5*sqrt(3.0) /)

   m = 0
   do j = 0, nc(2)-1        !loop over layers
    do k = 0, nc(3)-1
     do i = 0, nc(1)-1
      cvec(1) = i + mod(j,2)*0.5
      cvec(2) = j*sqrt(2.0/3.0)
      cvec(3) = k*sqrt(3.0) + mod(j,2)*0.5/sqrt(3.0)
      lattice(ilp*m+1,:) = unitcell(1,:) + cvec
      lattice(ilp*m+2,:) = unitcell(2,:) + cvec
      m = m+1
     enddo
    enddo
   enddo

  case default
   stop 'make_lattice: invalid lattice type input'
  end select

  if(size(lsc) .eq. 1) then
   ratom = lsc(1)*lattice
  else
   do i = 1, size(lsc)
    ratom(:,i) = lsc(i)*lattice(:,i) + offset(i)
   enddo
  endif

  if(do_mix) call mix

  deallocate(lattice)
  
  contains
!***********************************************************************
! internal subroutine to mix molecule positions in lattice  
  subroutine mix
  double precision :: temp
  do i = 1, na-1
   call random_number(temp)
   k = int(real(na-i)*temp)
   k = k + i
    do j = 1, ndim
     temp = ratom(i,j)
     ratom(i,j) = ratom(k,j)
     ratom(k,j) = temp
    enddo
  enddo
  end subroutine mix
!***********************************************************************
  end subroutine make_lattice
!***********************************************************************
  subroutine rdinput_lattice(lat,iunit)
!***********************************************************************
! Reads lattice variables from input file.
! 
!***********************************************************************
  implicit none
  
  type(lattice_data_type) :: lat
  
  integer :: j, iunit, error
  logical :: l3d
  
  read(iunit,*) lat%ltype
  if(lat%ltype .eq. 0 .or. lat%ltype .eq. 1 .or. lat%ltype .eq. 4) then
   l3d = .TRUE.
  elseif(lat%ltype .eq. 2 .or. lat%ltype .eq. 3) then
   l3d = .FALSE.
  else
   stop 'rdinput_lattice:invalid lattice type in lattice input data'
  endif
  if(l3d) then
   allocate(lat%nuc(3), stat=error)
   if(error .ne. 0) stop 'allocation error in rd_lattice_input - stop'
  else
   allocate(lat%nuc(2), stat=error)
   if(error .ne. 0) stop 'allocation error in rd_lattice_input - stop'
  endif
  read(iunit,*) (lat%nuc(j), j = 1, size(lat%nuc))
  read(iunit,*) lat%do_mix

  end subroutine rdinput_lattice
!***********************************************************************
  subroutine wr_lattice(lat,ounit)
! writes lattice data to the output file
!***********************************************************************
  implicit none
  type(lattice_data_type) :: lat
  integer :: ounit

  write (ounit,*) 'Initial lattice type    = ', lat%ltype
  
  end subroutine wr_lattice
!***********************************************************************
  end module lattice
