  module molecules
!***********************************************************************
! Module providing functionality for molecular properties.
! Note dimensions of r(ns,nm,3) and p(ns,nm,3).
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************
  use v_sub_library
  use atoms
  use species

  implicit none
 
  private

! Public procedures
  public :: calc_eint_mol, calc_fm, calc_rcm, calc_rcmij, calc_pcm, &
          & create_molecules, destroy_rcmij, init_molecules, &
          & rd_fractions_input, rd_molecules_input, &
          & rescale_mol_momenta, spread_cm, wr_molecules, &
          & xyz_to_r, zero_momenta_m

! Public data types
  public :: mol_pointer, species_pointer, fraction_mask, mol_set

  type mol_pointer
   double precision, dimension(:,:), pointer :: r, p ! (ns,dim)
  end type mol_pointer

  type species_pointer
   integer, dimension(:), allocatable :: spid ! (nsf)
  end type species_pointer

! fmask is true for atoms that can interact via pair forces
  type fraction_mask
   logical, dimension(:,:), allocatable :: fmask !(nsf,nsf)
  end type fraction_mask

  type mol_set
   integer :: nm               ! number of molecules
   integer :: np               ! number of particles
   integer :: nf               ! number of fractions
   integer, dimension(:), allocatable :: nmf  ! number of molecules in this
                                              ! fraction (nf)
   integer, dimension(:), allocatable :: nsf  ! number of sites per molecule
                                              ! in this fraction (nf)
   integer, dimension(:), allocatable :: fs   ! index of first atom of this
                                              ! molecule in atoms (nm)
   integer, dimension(:), allocatable :: ls   ! index of last atom of this
                                              ! molecule in atoms (nm)
   integer, dimension(:), allocatable :: mf   ! index of first molecule of this
                                              ! fraction (nf)
   integer, dimension(:), allocatable :: ml   ! index of last molecule of this
                                              ! fraction (nf)
   integer, dimension(:), allocatable :: fid  ! fraction that this molecule
                                              ! belongs to (nm)
   integer, dimension(:), allocatable :: mid  ! molecule that this site
                                              ! belongs to (np)
   integer, dimension(:,:), allocatable :: intyp  ! interaction type (nsp,nsp) - new Dave Lee
   type(species_pointer), dimension(:), allocatable :: fr  ! species id no. (nf)
   type(fraction_mask), dimension(:), allocatable :: frac_mask ! new for masking (nf)
   double precision, dimension(:), allocatable :: mass  ! molecular mass (nm)
   double precision, dimension(:), allocatable :: massinv ! inverse mass (nm)
   double precision, dimension(:), allocatable :: eint  ! internal energy (nm)
!  molecular centre of mass position and momentum and force on molecule
   double precision, dimension(:,:), allocatable :: rcm, pcm, fm ! (nm,dim)
!  vector from cm to site i of this molecule = ri - rcm
   double precision, dimension(:,:), allocatable :: ricm     ! (np,dim)
!  vector between molecular cm of interacting sites i and j
   double precision, dimension(:,:), allocatable :: rcmij    ! (ni,dim)
   type(mol_pointer), dimension(:), allocatable :: mol  ! array of r,p pointers (nm)
  end type mol_set

  contains
!***********************************************************************
  subroutine calc_eint_mol(md, atd)
!***********************************************************************
  implicit none
  type(mol_set) :: md
  type(atom_type) :: atd
  integer :: i

  do i = 1, md%nm
   md%eint(i) = sum(atd%eint(md%fs(i):md%ls(i)))
  enddo

  end subroutine calc_eint_mol
!***********************************************************************
  subroutine calc_fm(md, f)
!***********************************************************************
  implicit none
  type(mol_set) :: md
  double precision, dimension(:,:), intent(in) :: f
  integer :: i, j, ndim

  ndim = size(f,2)

  md%fm = 0.d0

  do i = 1, md%np
   do j = 1, ndim
    md%fm(md%mid(i),j) = md%fm(md%mid(i),j) + f(i,j)
   enddo
  enddo

  end subroutine calc_fm
!***********************************************************************
  subroutine calc_rcm(md, spd)
!***********************************************************************
  implicit none
  type(mol_set) :: md
  type(species_set) :: spd
  double precision :: tmp
  integer :: i, j, k, ndim

  ndim = size(md%mol(1)%r,2)
  md%rcm = 0.d0

  do j = 1, ndim
   do i = 1, md%nm
    tmp = 0.d0
    do k = 1, md%nsf(md%fid(i))
     tmp = tmp + md%mol(i)%r(k,j) &
                 & *spd%mass(md%fr(md%fid(i))%spid(k))
    enddo
    md%rcm(i,j) = tmp
   enddo
  enddo

  do j = 1, ndim
   do i = 1, md%nm
    md%rcm(i,j) = md%rcm(i,j)*md%massinv(i)
   enddo
  enddo

  do j = 1, ndim
   do i = 1, md%nm
    md%ricm(md%fs(i):md%ls(i),j) = md%mol(i)%r(:,j) - md%rcm(i,j)
   enddo
  enddo

  end subroutine calc_rcm
!***********************************************************************
  subroutine calc_rcm_2(md, spd)
!***********************************************************************
  implicit none
  type(mol_set) :: md
  type(species_set), optional :: spd
  integer :: i, j, ndim

  ndim = size(md%mol(1)%r,2)

  if(present(spd)) then
   do i = 1, md%nm
    do j = 1, ndim
     md%rcm(i,j) = sum(md%mol(i)%r(:,j) &
                 & *spd%mass(md%fr(md%fid(i))%spid))*md%massinv(i)
     md%ricm(md%fs(i):md%ls(i),j) = md%mol(i)%r(:,j) - md%rcm(i,j)
    enddo
   enddo
  else
   do i = 1, md%nm
    do j = 1, ndim
     md%rcm(i,j) = sum(md%mol(i)%r(:,j))*md%massinv(i)
     md%ricm(md%fs(i):md%ls(i),j) = md%mol(i)%r(:,j) - md%rcm(i,j)
    enddo
   enddo
  endif

  end subroutine calc_rcm_2
!***********************************************************************
  subroutine calc_rcm_1(md, spd)
!***********************************************************************
  implicit none
  type(mol_set) :: md
  type(species_set), optional :: spd
  integer :: i, j, ndim

  ndim = size(md%mol(1)%r,2)

! Possibly faster version - needs fixing
!  md%rcm = 0.0
!  do id = 1, ndim
!   do i = 1, md%nm
!    k = md%fid(i)
!    do j = md%fs(i), md%ls(i)
!     md%rcm(i,id) = md%rcm(i,id) + md%mol(i)%r(j,id)*spd%mass(md%fr(k)%spid)
!    enddo
!    md%rcm(i,id) = md%rcm(i,id)*md%massinv(i)
!   enddo
!  enddo

  if(present(spd)) then
   do i = 1, md%nm
    do j = 1, ndim
     md%rcm(i,j) = sum(md%mol(i)%r(:,j) &
                 & *spd%mass(md%fr(md%fid(i))%spid))*md%massinv(i)
    enddo
   enddo
  else
   do i = 1, md%nm
    md%rcm(i,:) = sum(md%mol(i)%r,1)*md%massinv(i)
   enddo
  endif

! position vectors relative to cm
  do j = 1, ndim
   do i = 1, md%nm
    md%ricm(md%fs(i):md%ls(i),j) = md%mol(i)%r(:,j) - md%rcm(i,j)
   enddo
  enddo

  end subroutine calc_rcm_1
!***********************************************************************
  subroutine calc_rcmij(md, rij, ilist, jlist)
! Calculation of molecular centre of mass separation vector for each
! pair of interacting atomic sites.
! Uses site separations rij and site positions relative to the cm
! ricm(i) and ricm(j) to calculate molecular centre of mass separations
! because rij is already minimum imaged.
!***********************************************************************
  implicit none
  type(mol_set) :: md
  double precision, dimension(:,:) :: rij
  integer, dimension(:), intent(in) :: ilist, jlist
  integer :: error

  if(allocated(md%rcmij)) deallocate(md%rcmij)
  allocate(md%rcmij(size(rij,1),size(rij,2)), stat=error)
  if(error .ne. 0) stop 'allocation error in calc_rcmij'
  md%rcmij = rij + md%ricm(ilist,:) - md%ricm(jlist,:)

  end subroutine calc_rcmij
!***********************************************************************
  subroutine calc_pcm(md)
!***********************************************************************
  implicit none
  type(mol_set) :: md
  integer :: i, j, k, ndim

  ndim = size(md%pcm,2)

  md%pcm = 0.d0
  do k = 1, ndim
   do i = 1, md%nm
    do j = 1, md%nsf(md%fid(i))
     md%pcm(i,k) = md%pcm(i,k) + md%mol(i)%p(j,k)
    enddo
   enddo
  enddo

  end subroutine calc_pcm
!***********************************************************************
  subroutine create_molecules(md, dim, spd)
!***********************************************************************
  type(mol_set) :: md
  type(species_set), optional :: spd
  integer :: ierr, dim, int_typ, i, j

  allocate(md%mid(md%np),md%nmf(md%nf),md%nsf(md%nf),md%mass(md%nm), &
         & md%massinv(md%nm),md%eint(md%nm), stat = ierr)
  if(ierr .ne. 0) stop 'molecules:create_molecules: allocation error'

  allocate(md%fs(md%nm),md%ls(md%nm),md%fid(md%nm),md%mol(md%nm), &
         & md%mf(md%nf),md%ml(md%nf), stat = ierr)
  if(ierr .ne. 0) stop 'molecules:create_molecules: allocation error'

  allocate(md%rcm(md%nm,dim), md%pcm(md%nm,dim), md%fm(md%np,dim), &
         & md%ricm(md%np,dim), stat=ierr)
  if(ierr .ne. 0) stop 'molecules:create_molecules: allocation error'

  if(present(spd)) then
   allocate(md%fr(md%nf), stat = ierr)
   if(ierr .ne. 0) stop 'molecules:create_molecules: allocation error'
! for new fraction masking
   allocate(md%frac_mask(md%nf), stat = ierr)
   if(ierr .ne. 0) stop 'molecules:create_molecules: allocation error'
   allocate(md%intyp(spd%nsp,spd%nsp), stat=ierr)
   if(ierr .ne. 0) stop 'molecules:create_molecules: allocation error'

   int_typ = 1
   do i = 1, spd%nsp 
    do j = i, spd%nsp
     md%intyp(i,j) = int_typ
     md%intyp(j,i) = int_typ
     int_typ = int_typ + 1 
    enddo
   enddo

  endif

  end subroutine create_molecules
!***********************************************************************
  subroutine destroy_molecules(md)
!***********************************************************************
  implicit none
  type(mol_set) :: md
  integer :: i, error=0

  if(allocated(md%fr)) then
   do i = 1, md%nf
    deallocate(md%fr(i)%spid, stat=error)
   enddo
   deallocate(md%fr, stat=error)
  endif
  if(error .ne. 0) stop 'deallocation error in destroy_molecules'

! new for fraction masking technique
  if(allocated(md%frac_mask)) then
   do i = 1, md%nf
    deallocate(md%frac_mask(i)%fmask, stat=error)
   enddo
   deallocate(md%frac_mask, stat=error)
  endif
  if(allocated(md%intyp)) deallocate(md%intyp)
  if(error .ne. 0) stop 'molecules:destroy_molecules: deallocation error'

  deallocate(md%mid,md%nmf,md%nsf,md%mass,md%massinv,md%eint,stat = error)
  deallocate(md%fs,md%ls,md%fid,md%mol,md%mf,md%ml,stat = error)
  deallocate(md%rcm, md%pcm, md%fm, md%ricm, md%rcmij, stat=error)
  deallocate(md%intyp, stat=error)
  if(error .ne. 0) stop 'molecules:destroy_molecules: deallocation error'

  do i = 1, md%nm
   nullify(md%mol(i)%r, md%mol(i)%p)
  enddo

  end subroutine destroy_molecules
!***********************************************************************
  subroutine destroy_rcmij(md)
!***********************************************************************
  implicit none
  type(mol_set) :: md

  if(allocated(md%rcmij)) deallocate(md%rcmij)

  end subroutine destroy_rcmij
!***********************************************************************
  subroutine init_molecules(atd, md, spd)
!***********************************************************************
  implicit none
  type(atom_type), target :: atd
  type(mol_set) :: md
  type(species_set), optional :: spd
  integer, dimension(:), allocatable :: isp
  integer :: i, j, isite, imol, error

! create index arrays for first site of molecule and fraction identity
  isite = 1
  imol = 0
  do i = 1, md%nf
   md%mf(i) = imol + 1
   md%ml(i) = imol + md%nmf(i)
   do j = 1, md%nmf(i)
    imol = imol + 1
    md%fs(imol) = isite
    md%fid(imol) = i
    isite = isite + md%nsf(i)
    md%ls(imol) = isite - 1
! initialize inverse molecular masses
    md%mass(imol) = sum(spd%mass(md%fr(i)%spid(:)))
    md%massinv(imol) = 1./md%mass(imol)
    md%mid(md%fs(imol):md%ls(imol)) = imol
   enddo
  enddo

! initialize number of degrees of freedom

! create molecular pointer arrays to atomic positions and momenta
  do i = 1, md%nm
   md%mol(i)%r => atd%r(md%fs(i):md%ls(i),:)
   md%mol(i)%p => atd%p(md%fs(i):md%ls(i),:)
  enddo

! initialize species indices
  if(present(spd)) then
  do i = 1, md%nm
   spd%spid(md%fs(i):md%ls(i)) = md%fr(md%fid(i))%spid
  enddo
  allocate(isp(spd%nsp), stat=error)
  if(error .ne. 0) stop 'allocation error in init_molecules'

  isp = 0
  do i = 1, spd%np
   isp(spd%spid(i)) = isp(spd%spid(i)) + 1
   spd%sp(spd%spid(i))%aid(isp(spd%spid(i))) = i
  enddo
  deallocate(isp)
  endif

  end subroutine init_molecules
!***********************************************************************
  subroutine rescale_mol_momenta(md,tempm,tr)
!***********************************************************************
  implicit none
  type (mol_set) :: md
  double precision :: tempm ! current molecular cm temperature
  double precision :: tr    ! desired molecular cm temperature
  double precision :: heat
  integer :: i
  
  heat = sqrt(tr/tempm)
  do i = 1, md%nm
   md%mol(i)%p = md%mol(i)%p*heat
  enddo

  end subroutine rescale_mol_momenta
!***********************************************************************
  subroutine rd_fractions_input(md,header,ifile,iunit,spd)
!***********************************************************************
  implicit none
  type (mol_set) :: md
  type (species_set), optional :: spd

  integer :: i, j, error, iunit, np
  
  character(len=*) :: header
  character(len=32) :: rdstring
  character(len=32) :: ifile      !input file name

  rdstring = ''
  i = 0
  open(iunit,file=ifile,status='old',form='formatted', &
     & position='rewind')
  do j = 1, 1000
   read(iunit,*, iostat=i) rdstring
   if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 999) exit
  enddo

  if (rdstring .eq. header) then
   np = 0
   do i = 1, md%nf
    read(iunit,*) md%nsf(i), md%nmf(i)
    if(present(spd)) then
     allocate(md%fr(i)%spid(md%nsf(i)),stat = error)
     if(error .ne. 0) stop 'molecules:rd_fractions_input: allocation error'
     read(iunit,*) (md%fr(i)%spid(j), j = 1, md%nsf(i))
! new for fraction masking technique
     allocate(md%frac_mask(i)%fmask(md%nsf(i),md%nsf(i)), stat=error)
     if(error .ne. 0) stop 'molecules:rd_fractions_input: allocation error'
     md%frac_mask(i)%fmask = .true.
    endif
    np = np + md%nsf(i)*md%nmf(i)
   enddo
   close(iunit)
   if(np .ne. md%np) then
    stop 'molecules:rd_fractions_input: Incorrect number of sites in fraction input file'
   endif
  else
   close(iunit)
   stop 'molecules:rd_fractions_input: Molecules input data not found'
  endif

  end subroutine rd_fractions_input
!***********************************************************************
  subroutine rd_molecules_input(md,header,ifile,iunit)
!***********************************************************************
  implicit none
  type (mol_set) :: md
 
  integer :: i, j, iunit
  
  character(len=*) :: header
  character(len=32) :: rdstring
  character(len=32) :: ifile      !input file name

  rdstring = ''
  i = 0
  open(iunit,file=ifile,status='old',form='formatted', &
     & position='rewind')
  do j = 1, 1000
   read(iunit,*, iostat=i) rdstring
   if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 999) exit
  enddo

  if (rdstring .eq. header) then
   read(iunit,*) md%nm, md%np, md%nf
   close(iunit)
  else
   close(iunit)
   stop 'Molecules input data not found - stop'
  endif

  end subroutine rd_molecules_input
!***********************************************************************
  subroutine spread_cm(r,ns,nm,x,y,z)
!***********************************************************************
  implicit none
  double precision, dimension (:), intent(out) :: x,y,z
  double precision, dimension (:,:), intent(in) :: r
  integer, intent(in) :: ns,nm
  integer :: i,i1,i2

  do i = 1, nm
   i1 = (i-1)*ns+1
   i2 = i*ns
   x(i1:i2) = r(i,1)
   y(i1:i2) = r(i,2)
   z(i1:i2) = r(i,3)
  enddo

  end subroutine spread_cm
!***********************************************************************
  subroutine wr_molecules(md,ofile,ounit)
! writes input data to the output file
!***********************************************************************
  implicit none
  type (mol_set), intent(in) :: md
  integer :: ounit
  character(len=32) :: ofile      !output file name

  open(ounit,file=ofile,status='old',form='formatted',position='append')
  write (ounit,*) 'Parameters for molecule set'
  write (ounit,*)
  write (ounit,*) 'Number of molecules       = ', md%nm
  write (ounit,*) 'Number of fractions       = ', md%nf
  write (ounit,*) 'Number of particles       = ', md%np
  write (ounit,*)
  close(ounit)
  
  end subroutine wr_molecules
!***********************************************************************
  subroutine xyz_to_r(x,y,z,ns,nm,r)
!***********************************************************************
  implicit none
  double precision, dimension (:), intent(in) :: x,y,z
  double precision, dimension (:,:,:), intent(out) :: r
  integer, intent(in) :: ns,nm
  integer :: i,i1,i2

  do i = 1, nm
   i1 = (i-1)*ns+1
   i2 = i*ns
   r(:,i,1) = x(i1:i2)
   r(:,i,2) = y(i1:i2)
   r(:,i,3) = z(i1:i2)
  enddo

  end subroutine xyz_to_r
!***********************************************************************
  subroutine zero_momenta_m(p,md,spd,dim)
!***********************************************************************
  implicit none
  type (mol_set), intent(in) :: md
  type (species_set), intent(in):: spd
  double precision, dimension(:,:), intent(inout) :: p
  integer :: i, n
  integer, optional :: dim
  
  if(present(dim)) then
! set total momentum to zero in chosen direction
   do i = 1, md%np
    p(i,dim) = p(i,dim) - &
    sum(p(:,dim))*spd%mass(spd%spid(i))*md%massinv(md%mid(i))/md%nm
   enddo
  else
! set total momentum to zero in each direction
   do n = 1, 3
    do i = 1, md%np
     p(i,n) = p(i,n) - &
     sum(p(:,n))*spd%mass(spd%spid(i))*md%massinv(md%mid(i))/md%nm
    enddo
   enddo
  endif

  end subroutine zero_momenta_m
!***********************************************************************
  end module molecules
