  module neighbour_list
!***********************************************************************
! Creates a neighbour list for 2-body forces.
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************
 use pdutil
 use simulation_box
 use sllod
 use molecules
 use species

 implicit none
 
 private

! Public procedures
 public :: compress_nlist, destroy_nlist, find_pair_separations, &
         & form_nlist, increment_nlist_drtot, init_nlist, &
         & allowed_interaction, &
         & rd_nlist_input, reform_nlist_now, sum_pair_forces, &
         & sum_pair_energies, wr_nlist

! Public data types
 public :: nlist_data_type

 type :: nlist_data_type
  double precision, dimension(:,:), allocatable :: rij, drtot
  double precision, dimension(:), allocatable :: rijsq, old_rijsq
  double precision :: shell, search_rsq, dsqmax
  double precision :: a_cll_sz_x, a_cll_sz_y  ! new for cell method
  double precision, dimension(:,:), allocatable :: fi  ! force vectors
  integer, dimension(:,:), allocatable :: nlist, ilist, ijindex
  integer, dimension(:), allocatable :: nabi, nabj
  integer, dimension(:), allocatable :: aid   ! atom ids for this interaction
  logical, dimension(:), allocatable :: nmask   ! mask array for interacting pairs
  integer :: nlist_type        ! 1 = all-particles, 2 = cell
  integer :: nlist_reform_type ! 1 = particle displacements, 2 = pair separations
  integer :: np, nab, inab
  integer, dimension(:,:), allocatable :: nclist  ! new for cell method
  integer :: nnc, ncxs, ncys, nczs, nct       ! new for cell method
  integer :: intyp, sp1, sp2                  ! interaction type
 end type nlist_data_type

 contains
!***********************************************************************
  subroutine all_pairs_method(nl, md)
!***********************************************************************
  type (nlist_data_type), intent(inout) :: nl
  type (mol_set), intent(in) :: md
  integer, dimension(nl%np*(nl%np-1)/2) :: iitmp, jitmp
  integer :: i, j, np, ni, k, ip, error

  np = nl%np
  ni = (np*(np-1)/2)

  ip = 0
  do k = 1, ni
   i = k-((k-1)/np)*np
   j = k+1-((k+(k-1)/np)/np)*np+(k-1)/np
   if(allowed_interaction(nl, md, i, j)) then
    ip = ip + 1
    iitmp(ip) = i
    jitmp(ip) = j
   endif
  enddo

  allocate(nl%ijindex(ip,2), stat=error)
  if(error .ne. 0) stop 'neighbour_list:all_pairs_method: allocation error'

  nl%ijindex(:,1) = iitmp(1:ip)
  nl%ijindex(:,2) = jitmp(1:ip)

  end subroutine all_pairs_method
!***********************************************************************
  function allowed_interaction(nl, md, i, j)
!***********************************************************************
  type (nlist_data_type), intent(inout) :: nl
  type (mol_set), intent(in) :: md
  integer :: i, j
  integer :: mol_i, mol_j, frac_i, frac_j, site_i, site_j, spec_i, spec_j  
  logical :: right_species, same_molecule, interacting
  logical :: allowed_interaction

  mol_i = md%mid(nl%aid(i))
  mol_j = md%mid(nl%aid(j))
  frac_i = md%fid(mol_i)
  frac_j = md%fid(mol_j)
  site_i = nl%aid(i) - md%fs(mol_i) + 1
  site_j = nl%aid(j) - md%fs(mol_j) + 1
  spec_i = md%fr(frac_i)%spid(site_i)
  spec_j = md%fr(frac_j)%spid(site_j)
  
  right_species = nl%intyp .eq. md%intyp(spec_i,spec_j)
  same_molecule = mol_i .eq. mol_j
  if(same_molecule) then
   interacting = md%frac_mask(frac_i)%fmask(site_i,site_j)
  else
   interacting = .false.
  endif

  allowed_interaction = right_species .and. &
  & (.not. same_molecule .or. (same_molecule .and. interacting))

  end function allowed_interaction
!***********************************************************************
  subroutine cell_method(nl, box, ratom, md)
! forms neighbour list using the cell_method
! reference:
! Matin, Daivis and Todd, Computer Physics Communications 2002.
!
! This cell code assumes a 3D system and that the lh box corner is
! at the origin (box%origin = 0).
!***********************************************************************
  type (nlist_data_type), intent(inout) :: nl
  type (box_data_type), intent(in) :: box
  type (mol_set), intent(in) :: md
  double precision, dimension(:,:) :: ratom
  double precision, dimension(3,3) :: cellvec

  double precision :: lx, ly, lz, cll_vol

  integer, allocatable, dimension(:), save :: head
  integer, allocatable, dimension(:,:), save :: ijtmp
  integer, dimension(size(ratom,1)) :: pinc

  integer :: new_ncxs, new_ncys, new_nczs, num_int
  integer :: i, k, cll, clx, cly, clz, ip, error
  
! data for fraction masking
  integer :: mol_i, mol_j, frac_i, frac_j, site_i, site_j, species_i, species_j, &
          &  cell_j, particle_i, particle_j, m, n
  logical :: dointer

! need box lengths to determine number of cells in each direction

  lx = (box%boxvec(1,1)**2 + box%boxvec(1,2)**2)**0.5
  ly = (box%boxvec(1,1)*box%boxvec(2,2) - box%boxvec(1,2)*box%boxvec(2,1))/lx
  lz = box%boxvec(3,3)

  new_ncxs = int(lx/nl%a_cll_sz_x)
  new_ncys = int(ly/nl%a_cll_sz_y)
  new_nczs = int(lz/nl%a_cll_sz_y)   !approx cell size in z direction is the same as y
  if((new_ncxs.ne.nl%ncxs).or.(new_ncys.ne.nl%ncys).or.(new_nczs.ne.nl%nczs)) then
   nl%ncxs = new_ncxs
   nl%ncys = new_ncys
   nl%nczs = new_nczs
   nl%nct = nl%ncxs*nl%ncys*nl%nczs
   call form_template(nl)
  endif

  cll_vol = box%volume/float(nl%nct)

! modified to only deallocate and reallocate if nct changes
  if(.not. allocated(head)) then
   allocate(head(nl%nct), stat = error)
   if(error .ne. 0) stop 'neighbour_list:cell_method: allocation error'
  elseif (size(head) .ne. nl%nct) then
   deallocate(head)
   allocate(head(nl%nct), stat = error)
   if(error .ne. 0) stop 'neighbour_list:cell_method: allocation error'
  endif

  pinc = 0
  head = 0

! create cell vectors
  cellvec(1,:) = box%boxvec(1,:)/nl%ncxs
  cellvec(2,:) = box%boxvec(2,:)/nl%ncys
  cellvec(3,:) = box%boxvec(3,:)/nl%nczs

! sort particles into cells

  do i = 1, nl%np
   clx = nint((ratom(i,1)*cellvec(2,2)*cellvec(3,3) &
       & - ratom(i,2)*cellvec(2,1)*cellvec(3,3))/cll_vol - 0.5)
   cly = nint((-ratom(i,1)*cellvec(1,2)*cellvec(3,3) &
       & + ratom(i,2)*cellvec(1,1)*cellvec(3,3))/cll_vol - 0.5)
   clz = nint(ratom(i,3)/cellvec(3,3) - 0.5)
   clx = modulo(clx,nl%ncxs)
   cly = modulo(cly,nl%ncys)
   clz = modulo(clz,nl%nczs)
   cll = clz*(nl%ncxs*nl%ncys) + cly*nl%ncxs + clx + 1
   pinc(i) = head(cll)
   head(cll) = i
  enddo


! Estimate the total number of interactions for
! initial array dimensions.
! Set the initial dimension so that it will increase if resize_ijtmp
! is called - the constant 5 is needed.
  num_int = nl%np * nl%np * int(real(nl%nnc) / real(nl%nct)) + 5
  if(allocated(ijtmp)) deallocate(ijtmp)
  allocate(ijtmp(num_int,2), stat=error)
  if(error .ne. 0) stop 'neighbour_list:cell_method: allocation error'

! Same cell interactions

  ijtmp = 0

  ip = 0
  do k = 1, nl%nct
   particle_i = head(k)
   do while (particle_i .ne. 0)
    mol_i = md%mid(nl%aid(particle_i))
    frac_i = md%fid(mol_i)
    site_i = nl%aid(particle_i) - md%fs(mol_i) + 1
    species_i = md%fr(frac_i)%spid(site_i)
    particle_j = pinc(particle_i)
    do while (particle_j .ne. 0)
     dointer = .false.
     mol_j = md%mid(nl%aid(particle_j))
     frac_j = md%fid(mol_j)
     site_j = nl%aid(particle_j) - md%fs(mol_j) + 1
     species_j = md%fr(frac_j)%spid(site_j)
     if(nl%intyp .eq. md%intyp(species_i,species_j)) then
      if(mol_i .eq. mol_j) then
       dointer = md%frac_mask(frac_i)%fmask(site_i,site_j)
      else
       dointer = .true.
      endif
     endif
     if(dointer) then
      ip = ip + 1
       if(ip .gt. num_int) call resize_ijtmp
       ijtmp(ip,1) = particle_i
       ijtmp(ip,2) = particle_j
     endif
     particle_j = pinc(particle_j)
    enddo
    particle_i = pinc(particle_i)
   enddo
  enddo

! different cell interactions

  do m = 1, nl%nct
   particle_i = head(m)
   do while (particle_i .ne. 0)
    mol_i = md%mid(nl%aid(particle_i))
    frac_i = md%fid(mol_i)
    site_i = nl%aid(particle_i) - md%fs(mol_i) + 1
    species_i = md%fr(frac_i)%spid(site_i)
    do n = 1, nl%nnc
     cell_j = nl%nclist(m,n)
     particle_j = head(cell_j)
     do while (particle_j .ne. 0)
      dointer = .false.
      mol_j = md%mid(nl%aid(particle_j))
      frac_j = md%fid(mol_j)
      site_j = nl%aid(particle_j) - md%fs(mol_j) + 1
      species_j = md%fr(frac_j)%spid(site_j)
      if(nl%intyp .eq. md%intyp(species_i,species_j)) then
       if(mol_i .eq. mol_j) then
        dointer = md%frac_mask(frac_i)%fmask(site_i,site_j)
       else
        dointer = .true.
       endif
      endif
      if(dointer) then
       ip = ip + 1
       if(ip .gt. num_int) call resize_ijtmp
       ijtmp(ip,1) = particle_i
       ijtmp(ip,2) = particle_j
      endif
      particle_j = pinc(particle_j)
     enddo
    enddo
    particle_i = pinc(particle_i)
   enddo
  enddo

! must alter below code to allow deallocation of appropriate arrays

  if(allocated(nl%ijindex)) deallocate(nl%ijindex)
  allocate(nl%ijindex(ip,2), stat=error)
  if(error .ne. 0) stop 'allocation error in cell_method'

  nl%ijindex(:,1) = ijtmp(1:ip,1)
  nl%ijindex(:,2) = ijtmp(1:ip,2)

  contains

!***********************************************************************
  subroutine resize_ijtmp
!***********************************************************************
  integer, allocatable, dimension(:,:) :: ijswap
  integer :: num_int_new, error

  num_int_new = int(1.2*num_int)
  allocate(ijswap(num_int_new,2), stat=error)
  if(error .ne. 0) stop 'allocation error in resize_ijtmp'
  ijswap(1:num_int,1) = ijtmp(:,1)
  ijswap(1:num_int,2) = ijtmp(:,2)
  call move_alloc(ijswap,ijtmp)
  num_int = num_int_new

  end subroutine resize_ijtmp
!***********************************************************************
  end subroutine cell_method
!***********************************************************************
  subroutine compress_nlist(nl)
!***********************************************************************
  type (nlist_data_type), intent(inout) :: nl
  integer :: i, j

! replace pack with scalar code
  j = 0
  do i = 1, nl%nab
   if(nl%nmask(i)) then
    j = j + 1
    nl%ilist(j,1) = nl%nlist(i,1)
    nl%ilist(j,2) = nl%nlist(i,2)
   endif
  enddo

  end subroutine compress_nlist
!***********************************************************************
  subroutine destroy_nlist(nl)
!***********************************************************************
  type (nlist_data_type), intent(inout) :: nl

! destroy neighbour list
  if(allocated(nl%drtot)) deallocate(nl%drtot)
  if(allocated(nl%old_rijsq)) deallocate(nl%old_rijsq)
  if(allocated(nl%nlist)) deallocate(nl%nlist)
  if(allocated(nl%ilist)) deallocate(nl%ilist)
  if(allocated(nl%rijsq)) deallocate(nl%rijsq)
  if(allocated(nl%nmask)) deallocate(nl%nmask)
  if(allocated(nl%rij)) deallocate(nl%rij)
  if(allocated(nl%nabi)) deallocate(nl%nabi)
  if(allocated(nl%nabj)) deallocate(nl%nabj)
  if(allocated(nl%ijindex)) deallocate(nl%ijindex)
  if(allocated(nl%nclist)) deallocate(nl%nclist)
  if(allocated(nl%aid)) deallocate(nl%aid)
  if(allocated(nl%fi)) deallocate(nl%fi)

  end subroutine destroy_nlist
!***********************************************************************
  subroutine find_pair_separations(pairs, box, ratom, cutsq, nmask, rij, rijsq, ni)
!***********************************************************************
  type (box_data_type), intent(inout) :: box
  double precision, dimension(:,:) :: ratom
  double precision, dimension(:,:) :: rij
  double precision, dimension(:), optional :: rijsq
  double precision :: cutsq
  integer, dimension(:,:) :: pairs
  logical, dimension(:), optional :: nmask
  integer, optional :: ni
  integer :: i, j, npair, ndim

! find particle separations
!  rij = ratom(pairs(:,2),:) - ratom(pairs(:,1),:)

  npair = size(rij,1)
  ndim = 3

  do j = 1, ndim
  !dir$ vector always
   do i = 1, npair
    rij(i,j) = ratom(pairs(i,2),j) - ratom(pairs(i,1),j)
   enddo
  enddo

  call minimum_image(box, rij)

! find interacting pairs
!  if(present(rijsq)) rijsq = sum(rij*rij,2)

  if(present(rijsq)) then
   if(ndim .eq. 2) then
  !dir$ vector always
    do i = 1, npair
     rijsq(i) = rij(i,1)*rij(i,1) +rij(i,2)*rij(i,2)
    enddo
   else
  !dir$ vector always
    do i = 1, npair
     rijsq(i) = rij(i,1)*rij(i,1) + rij(i,2)*rij(i,2) + rij(i,3)*rij(i,3)
    enddo
   endif
  endif

  if(present(nmask)) nmask = rijsq .le. cutsq
  if(present(ni)) ni = count(nmask)
! print *, 'ni = ', ni
  end subroutine find_pair_separations
!***********************************************************************
  subroutine form_nlist(nl, box, ratom, md)
!***********************************************************************
  type (nlist_data_type), intent(inout) :: nl
  type (box_data_type), intent(inout) :: box
  type (mol_set), intent(inout) :: md
  double precision, dimension(:,:) :: ratom
  double precision, dimension(:,:), allocatable :: rij
  integer :: i, j, error, np, ndim, npair
  logical, dimension(:), allocatable :: nmask

! restrict to 3d systems in this version
  ndim = 3
! form ijindex array

  select case(nl%nlist_type)
! all-particle neighbour list formation
   case(1)
    if(.not. allocated(nl%ijindex)) then
     np = nl%np
     call all_pairs_method(nl, md)
    endif
   case(2)
! cell code neighbour list formation
   if(ndim .eq. 3) call cell_method(nl, box, ratom, md)
   case default
    stop 'neighbour_list:form_nlist: Invalid neighbour list type'
  end select

  npair = size(nl%ijindex,1)

! find particle separations
  if(allocated(rij)) deallocate(rij)
  allocate(rij(npair,ndim), stat=error)
  if(error .ne. 0) stop 'neighbour_list:form_nlist allocation error for rij'

  do j = 1, ndim
   do i = 1, npair
    rij(i,j) = ratom(nl%ijindex(i,2),j) - ratom(nl%ijindex(i,1),j)
   enddo
  enddo

! find minimum image separations
  call minimum_image(box, rij)

! eliminate pairs outside the tolerance shell and find total number of
! pairs in list
  if(allocated(nmask)) deallocate(nmask)
  allocate(nmask(npair), stat=error)
  if(error .ne. 0) stop 'neighbour_list:form_nlist allocation error for nmask'
  nmask = sum(rij*rij,2) .le. nl%search_rsq
  nl%nab = count(nmask)

!  if(nl%nab .eq. 0) stop 'neighbour_list:form_nlist: No particles in &
!  & neighbour list - increase tolerance'

  if(nl%nab .eq. 0) then
! reset total displacement since last neighbour list reformation
   nl%drtot = 0.d0
   return
  endif

! Allocate arrays for neighbour list and used in pair force
  if(allocated(nl%nlist)) deallocate(nl%nlist)
  if(allocated(nl%ilist)) deallocate(nl%ilist)
  if(allocated(nl%rij)) deallocate(nl%rij)
  if(allocated(nl%rijsq)) deallocate(nl%rijsq)
  if(allocated(nl%nmask)) deallocate(nl%nmask)
  allocate(nl%nlist(nl%nab,2), nl%ilist(nl%nab,2), nl%rij(nl%nab,ndim), &
  & nl%rijsq(nl%nab), nl%nmask(nl%nab), stat=error)
  if(error .ne. 0) stop 'allocation error in form_nlist'

! create neighbour list
  nl%nlist(:,1) = pack(nl%ijindex(:,1), mask=nmask)
  nl%nlist(:,2) = pack(nl%ijindex(:,2), mask=nmask)

  if(nl%nlist_reform_type .eq. 2) then
   if(allocated(nl%drtot)) deallocate(nl%drtot)
   if(allocated(nl%old_rijsq)) deallocate(nl%old_rijsq)
   allocate(nl%drtot(nl%nab,ndim), stat=error)
   if(error .ne. 0) stop 'allocation error in form_nlist for nl%drtot'
   allocate(nl%old_rijsq(nl%nab), stat=error)
   if(error .ne. 0) stop 'allocation error in form_nlist for nl%old_rijsq'
   j = 1
   do i = 1, npair
    if(nmask(i)) then
     nl%drtot(j,:) = rij(i,:)
     j = j + 1
    endif
   enddo
   nl%old_rijsq = sum(nl%drtot**2,2)
  else
! reset total displacement since last neighbour list reformation
   nl%drtot = 0.d0
  endif

! deallocate temporary variables
  if(allocated(nmask)) deallocate(nmask)
  if(allocated(rij)) deallocate(rij)
  
  end subroutine form_nlist
!***********************************************************************
  subroutine form_template(nl)
! determines neighbouring cell information
!***********************************************************************
  type (nlist_data_type), intent(inout) :: nl

  integer, dimension(nl%nnc) :: j_cll, j_clx, j_cly, j_clz
  integer, dimension(nl%nnc,3) :: shifts
  integer :: clx, cly, clz, cll, error

  if((nl%ncxs.lt.3).or.(nl%ncys.lt.3).or.(nl%nczs.lt.3.)) then
   write(*,*) 'nnc = ', nl%nnc
   write(*,*) 'ncxs = ', nl%ncxs
   write(*,*) 'ncys = ', nl%ncys
   write(*,*) 'nczs = ', nl%nczs
   write(*,*) ' require ncps = 3'
   stop 'not enough cells for cell method'
  endif

  shifts(1,:) = (/ 0, 0, -1 /)
  shifts(2,:) = (/ 1, 0, -1 /)
  shifts(3,:) = (/ 1, 0, 0 /)
  shifts(4,:) = (/ 1, 0, 1 /)
  shifts(5,:) = (/ -1, 1, -1 /)
  shifts(6,:) = (/ 0, 1, -1 /)
  shifts(7,:) = (/ 1, 1, -1 /)
  shifts(8,:) = (/ -1, 1, 0 /)
  shifts(9,:) = (/ 0, 1, 0 /)
  shifts(10,:) = (/ 1, 1, 0 /)
  shifts(11,:) = (/ -1, 1, 1 /)
  shifts(12,:) = (/ 0, 1, 1 /)
  shifts(13,:) = (/ 1, 1, 1 /)

  if(allocated(nl%nclist)) deallocate(nl%nclist)
  allocate(nl%nclist(nl%nct,nl%nnc), stat=error)
  if(error .ne. 0) stop 'allocation error in neighbour_list form_template'

  do clx = 0, nl%ncxs-1
   do cly = 0, nl%ncys-1
    do clz = 0, nl%nczs-1
     cll = clz*nl%ncxs*nl%ncys + cly*nl%ncxs + clx + 1
     j_clx = clx + shifts(:,1)
     j_cly = cly + shifts(:,2)
     j_clz = clz + shifts(:,3)
     j_clx = modulo(j_clx,nl%ncxs)
     j_cly = modulo(j_cly,nl%ncys)
     j_clz = modulo(j_clz,nl%nczs)
     j_cll = j_clz*(nl%ncxs*nl%ncys) + j_cly*nl%ncxs + j_clx + 1
     nl%nclist(cll,:) = j_cll
    enddo
   enddo
  enddo

  end subroutine form_template
!***********************************************************************
  subroutine increment_nlist_drtot(nl,dr)
!***********************************************************************
  type (nlist_data_type), intent(inout) :: nl
  double precision, dimension(:,:) :: dr
  integer i

  if(nl%nlist_reform_type .eq. 2) then
   if(nl%nab .eq. 0) stop 'nlist_reform_type must equal 1 if nab = 0'
   do i = 1, nl%nab
    nl%drtot(i,:) = nl%drtot(i,:) + dr(nl%nlist(i,1),:) + dr(nl%nlist(i,2),:)
   enddo
  else
   nl%drtot = nl%drtot + dr
  endif

  end subroutine increment_nlist_drtot
!***********************************************************************
 subroutine init_nlist(nl, sd, rcut, ndim, spd)
!***********************************************************************
 type (nlist_data_type), intent(inout) :: nl
 type (sllod_data_type), intent(in) :: sd
 type (species_set), intent(in) :: spd
 double precision :: rcut, factor
 integer :: ndim, error

! do some initializations
  nl%search_rsq = (nl%shell+rcut)**2

! PD version
  nl%np = spd%nasp(nl%sp1)
  if(nl%sp1 .ne. nl%sp2) nl%np = nl%np + spd%nasp(nl%sp2)

  allocate(nl%aid(nl%np), nl%fi(nl%np,ndim), stat=error)
  if(error .ne. 0) stop 'neighbour_list:init_nlist: allocation error'
  nl%aid(1:spd%nasp(nl%sp1)) = spd%sp(nl%sp1)%aid
  if(nl%sp1 .ne. nl%sp2) nl%aid(spd%nasp(nl%sp1)+1:nl%np) = spd%sp(nl%sp2)%aid

  select case(nl%nlist_reform_type)
   case(1) ! based on individual particle displacements
    allocate(nl%drtot(nl%np,ndim), stat = error)
    if(error .ne. 0) stop 'neighbour_list: init_nlist: allocation error for drtot'
    nl%drtot = 0.d0
    nl%dsqmax = (nl%shell*0.5d0)**2
   case(2) ! based on pair separations
    nl%dsqmax = 2.d0*nl%shell*(rcut + 0.5d0*nl%shell)
   case default
    stop 'invalid nl%nlist_reform_type'
   end select

  allocate(nl%nabi(nl%np), nl%nabj(nl%np), stat=error)
  if(error .ne. 0) stop 'neighbour_list: init_nlist: allocation error for nl%nabi'

! cell_method initializations
  if(nl%nlist_type == 2) then

   select case(ndim)
    case(2)
     nl%ncxs = 0
     nl%ncys = 0
     nl%nnc = 4
    case(3)
     nl%ncxs = 0
     nl%ncys = 0
     nl%nczs = 0
     nl%nnc = 13
    case default
   end select

   select case(sd%flow_type)
    case(shear)
     factor = sqrt(5.0)/2.0
    case(pef)
     factor = sqrt(2.0)
    case default
     factor = 1.0
   end select

   nl%a_cll_sz_x = (nl%shell+rcut)*factor
   nl%a_cll_sz_y = nl%shell+rcut

  endif

  end subroutine init_nlist
!***********************************************************************
  subroutine rd_nlist_input(nl,header,ifile,iunit)
!***********************************************************************
! Reads neighbour list variables from input file.
! 
!***********************************************************************
  type (nlist_data_type), intent(inout) :: nl
  
  integer :: i, j, iunit
  
  character(len=*) :: header
  character(len=32) :: rdstring
  character(len=32) :: ifile         ! input file name

  rdstring = ''
  i = 0
  open(iunit,file=ifile,status='old',form='formatted', &
     & position='rewind')
  do j = 1, 1000
   read(iunit,*, iostat=i) rdstring
   if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 999) exit
  enddo

  if (rdstring .eq. header) then
   read(iunit,*) nl%shell
   read(iunit,*) nl%nlist_type
   read(iunit,*) nl%nlist_reform_type
   close(iunit)
  else
   close(iunit)
    stop 'Neighbour list input data not found - stop'
  endif

  end subroutine rd_nlist_input
!***********************************************************************
  function reform_nlist_now(nl)
!***********************************************************************
  implicit none
  type (nlist_data_type), intent(inout) :: nl
  logical :: reform_nlist_now
  
  if(nl%nlist_reform_type .eq. 2) then
   ! check based on changes in pair separations
   reform_nlist_now = (maxval(abs(sum(nl%drtot**2,2) - nl%old_rijsq)) .ge. nl%dsqmax)
  else
   ! check based on single particle displacements
   reform_nlist_now = (maxval(sum(nl%drtot**2,2)) .ge. nl%dsqmax)
  endif

  end function reform_nlist_now
!***********************************************************************
  subroutine sum_pair_energies(nl, uij, ei)
!***********************************************************************
! Accumulates pair energies into the total site (atom) energy.
! 
!***********************************************************************
  type (nlist_data_type), intent(in) :: nl
  double precision, dimension(:) :: uij, ei
  double precision, dimension(nl%np) :: eitmp
  integer :: i, j, k

  eitmp = 0.d0

  do i = 1, nl%inab
   j = nl%ilist(i,1)
   k = nl%ilist(i,2)
   eitmp(j) = eitmp(j) + 0.5d0*uij(i)
   eitmp(k) = eitmp(k) + 0.5d0*uij(i)
  enddo

  do i = 1, nl%np
   ei(nl%aid(i)) = ei(nl%aid(i)) + eitmp(i)
  enddo

  end subroutine sum_pair_energies
!***********************************************************************
  subroutine sum_pair_forces(nl, fij)
!***********************************************************************
! Accumulates pair forces into the total interaction site (atom) force.
! 
!***********************************************************************
  type (nlist_data_type), intent(inout) :: nl
  double precision, dimension(:,:) :: fij
  integer :: i, j, k

  do i = 1, nl%inab
   j = nl%ilist(i,1)
   nl%fi(j,1) = nl%fi(j,1) + fij(i,1)
   nl%fi(j,2) = nl%fi(j,2) + fij(i,2)
   nl%fi(j,3) = nl%fi(j,3) + fij(i,3)
  enddo
  do i = 1, nl%inab
   k = nl%ilist(i,2)
   nl%fi(k,1) = nl%fi(k,1) - fij(i,1)
   nl%fi(k,2) = nl%fi(k,2) - fij(i,2)
   nl%fi(k,3) = nl%fi(k,3) - fij(i,3)
  enddo

  end subroutine sum_pair_forces
!***********************************************************************
  subroutine wr_nlist(nl,ofile,ounit)
! writes input data to the output file
!***********************************************************************
  type (nlist_data_type), intent(in) :: nl
  integer :: ounit
  character(len=32) :: ofile      !output file name

  open(ounit,file=ofile,status='old',form='formatted',position='append')
  write (ounit,*) 'Neighbour list data'
  write (ounit,*)
  write (ounit,*) 'Neighbour list type       = ', nl%nlist_type, ' (1=brute, 2=cell)'
  write (ounit,*) 'N list shell thickness    = ', nl%shell
  write (ounit,*)
  close(ounit)
  
  end subroutine wr_nlist
!***********************************************************************
  end module neighbour_list
