!***********************************************************************
   program nemd
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************
! Modular nonequilibrium molecular dynamics program - molecular NEMD.
! Homogeneous NEMD in 3D with periodic boundary conditions.
!
! This version: P.J. Daivis 2017.
! Do not distribute without permission.
!
! Acknowledge as follows.
! Please cite:
! "Nonequilibrium molecular dynamics: Theory, algorithms and applications"
! by B.D. Todd and P.J. Daivis (Cambridge University Press 2017)
!
!***********************************************************************

  use pdutil
  use atoms
  use constraints
  use dihedral_force
  use bond_stretch
  use bond_bend
  use gear5
  use integrator
  use pair_force
  use molecules
  use neighbour_list
  use properties
  use run_control
  use simulation_box
  use sllod
  use species
  use system
  use thermostat
  use tcf
  use lattice
  use ewald
  use ensemble_av

  implicit none

  type(atom_type) :: atd
  type(constraint_set) :: ca
  type(dihed_set) :: da
  type(bondstr_set) :: bs
  type(bondbend_set) :: bb
  type(integrator_data_type) :: ded
  type(pf_data_type), dimension(:), allocatable :: pf
  type(mol_set) :: md
  type(nlist_data_type), dimension(:), allocatable :: nl
  type(properties_type) :: pd
  type(run_control_type) :: rc
  type(box_data_type) :: box
  type(sllod_data_type) :: sd
  type(species_set) :: spd
  type(tcf_data) :: scfd, scfd2, jqjqcf, vacfd
  type(tcf_data), dimension(:), allocatable :: jmjmcf, jqjmcf, jmjqcf
  type(system_data_type) :: sys
  type(ewald_data_type) :: ewd
  type(eav_data) :: ead

  type(plink), pointer :: root, current

! declare local variables for main program
  double precision, dimension(3) :: tmp
  double precision :: sf
  integer :: i, istep, j, k, m
  character(len=2) :: kbchar
  
!***********************************************************************
! Start of executable code
!***********************************************************************

! initialize input file name
  call init_input_file

! read input files and do initializations
  call init_1

! initialize output file name
  call init_output_files

! initialize positions and momenta
  if(rc%runtype .eq. 0) then
   call start_new_state
  else
   call start_saved_state
  endif

! do more initializations
  call init_2

! main molecular dynamics loop
   do istep = 1, rc%steps_this_run

    call mdstep(rc,box,atd,ded,pf,nl,pd,sd,md,ca,spd,da,bs,bb,sys,ewd)

! accumulate properties
    if(pd%accumulate_properties_now) call accumulate_properties(pd)
! accumulate steady-state tcfs
    call calculate_tcfs
! write state and property files
    if(rc%store_state_now) call save_state_files

   enddo
! end of main md loop

! save state and properties files at end of run
! regardless of the value of kstore
  call save_state_files
! output results
  call write_output

! deallocate dynamically allocated memory (mainly pointer arrays)
   call cleanup

   if(ead%ni .gt. 1) then
    print *, 'Program nemd ended normally on image '//ead%imlabel
   else
    print *, 'Program nemd ended normally'
   endif
   stop
!***********************************************************************
   contains
! Note: internal subroutines get access to all variables of
! main program by host association
!
!***********************************************************************
  subroutine calculate_tcfs
!***********************************************************************
    integer :: i1, i2, i3

! form symmetric traceless pressure tensor elements
    if(scfd%kcf .ne. 0) then
     if(mod(istep,scfd%kcf) .eq. 0) then
      scfd%nv(1) = pd%inst(pd%i(22)) - pd%inst(pd%i(21))
      scfd%nv(2) = (pd%inst(pd%i(22)+1) + pd%inst(pd%i(22)+3))*0.5d0
      scfd%nv(3) = (pd%inst(pd%i(22)+2) + pd%inst(pd%i(22)+6))*0.5d0
      scfd%nv(4) = pd%inst(pd%i(22)+4) -  pd%inst(pd%i(21))
      scfd%nv(5) = (pd%inst(pd%i(22)+5) + pd%inst(pd%i(22)+7))*0.5d0
      call accum_tcf(scfd)
      if(scfd%kcfm .ne. 0 .and. mod(istep,scfd2%kcf) .eq. 0) then
       scfd2%nv = scfd%nv
       call accum_tcf(scfd2)
      endif
     endif
    endif

! heat flux acf
    if(jqjqcf%kcf .ne. 0) then
     if(mod(istep,jqjqcf%kcf) .eq. 0) then
      jqjqcf%nv(1) = pd%inst(pd%i(9))
      jqjqcf%nv(2) = pd%inst(pd%i(9)+1)
      jqjqcf%nv(3) = pd%inst(pd%i(9)+2)
      call accum_tcf(jqjqcf)
     endif
    endif

! Multicomponent solution correlation functions
    if(md%nf .ge. 2) then
! mass-mass flux acf
     if(jmjmcf(1)%kcf .ne. 0) then
      i3=0
      do i1 = 1, md%nf-1
       do i2 = 1, md%nf-1
        i3 = i3+1
        if(mod(istep,jmjmcf(i3)%kcf) .eq. 0) then
         jmjmcf(i3)%nv(1) = pd%inst(pd%i(8)+(i1-1)*3)
         jmjmcf(i3)%nv(2) = pd%inst(pd%i(8)+(i1-1)*3+1)
         jmjmcf(i3)%nv(3) = pd%inst(pd%i(8)+(i1-1)*3+2)
         tmp(1) = pd%inst(pd%i(8)+(i2-1)*3)
         tmp(2) = pd%inst(pd%i(8)+(i2-1)*3+1)
         tmp(3) = pd%inst(pd%i(8)+(i2-1)*3+2)
         call accum_tcf(jmjmcf(i3),tmp)
        endif
       enddo
      enddo
     endif

! heat-mass flux cross correlation function
     if(jqjmcf(1)%kcf .ne. 0) then
      do i1 = 1, md%nf-1
       if(mod(istep,jqjmcf(i1)%kcf) .eq. 0) then
        jqjmcf(i1)%nv(1) = pd%inst(pd%i(9))
        jqjmcf(i1)%nv(2) = pd%inst(pd%i(9)+1)
        jqjmcf(i1)%nv(3) = pd%inst(pd%i(9)+2)
        tmp(1) = pd%inst(pd%i(8)+(i1-1)*3)
        tmp(2) = pd%inst(pd%i(8)+(i1-1)*3+1)
        tmp(3) = pd%inst(pd%i(8)+(i1-1)*3+2)
        call accum_tcf(jqjmcf(i1), tmp)
       endif
      enddo
     endif

! mass-heat flux cross correlation function
     if(jmjqcf(1)%kcf .ne. 0) then
      do i1 = 1, md%nf-1
       if(mod(istep,jmjqcf(i1)%kcf) .eq. 0) then
        jmjqcf(i1)%nv(1) = pd%inst(pd%i(8)+(i1-1)*3)
        jmjqcf(i1)%nv(2) = pd%inst(pd%i(8)+(i1-1)*3+1)
        jmjqcf(i1)%nv(3) = pd%inst(pd%i(8)+(i1-1)*3+2)
        tmp(1) = pd%inst(pd%i(9))
        tmp(2) = pd%inst(pd%i(9)+1)
        tmp(3) = pd%inst(pd%i(9)+2)
        call accum_tcf(jmjqcf(i1),tmp)
       endif
      enddo
     endif
    endif

! velocity auto correlation function
    if(vacfd%kcf .ne. 0) then
    if(mod(istep,vacfd%kcf) .eq. 0) then
    m = 0
    do j = 1, md%nm
     do k = 1, 3
      m = m + 1
      vacfd%nv(m) = md%pcm(j,k)*md%massinv(j)
      if(m .eq. vacfd%nsr) exit
     enddo
     if(m .eq. vacfd%nsr) exit
    enddo
    call accum_tcf(vacfd)
    endif
    endif

  end subroutine calculate_tcfs
!***********************************************************************
  subroutine cleanup
!***********************************************************************
   call destroy_box(box)
   call destroy_atoms(atd)
   call destroy_integrator(ded)
   call destroy_properties(pd)
!   call destroy_molecules(md)
!   call destroy_constraints(cset)
!   call destroy_constraint_set(cset)

  end subroutine cleanup
!***********************************************************************
  subroutine define_properties
! Each property is dynamically added to the list of computed properties,
! depending on the values of options at run-time.
! Every property has an id number and a dimension, so that more than one
! value will be stored for each property, e.g. there are nine elements
! of the pressure tensor.
! Current value of maxid = 61
!***********************************************************************
  pd%nprop = 0
  pd%maxid = 0
! first property link must be root of list
  call add_property(pd,root,'Utotal',1,1,.true.)
  current => root
  call add_property(pd,current,'(Utotal)^2',1,2)
  call add_property(pd,current,'Upot-short',1,3)
  call add_property(pd,current,'Ukin',1,4)
  call add_property(pd,current,'Ukin',spd%nsp,5)
  call add_property(pd,current,'Upot',spd%npsp,6)
  do i = 1, spd%npsp
   if(pf(i)%pftype .eq. 3 .or. pf(i)%pftype .eq. 4) then
    call add_property(pd,current,'U-Ewald-Fourier',1,47)
    call add_property(pd,current,'U-Ewald-self',1,48)
    call add_property(pd,current,'U-Ewald-selfb',1,49)
    call add_property(pd,current,'P-es',1,50)
    call add_property(pd,current,'U-es/3V',1,51)
    exit
   elseif(pf(i)%pftype .eq. 5 .or. pf(i)%pftype .eq. 6) then
    call add_property(pd,current,'P-es',1,50)
    call add_property(pd,current,'U-es/3V',1,51)
    exit
   endif
  enddo
   call add_property(pd,current,'Tka',1,10)
   call add_property(pd,current,'Tkm',1,20)
   call add_property(pd,current,'alpha',1,16)
   call add_property(pd,current,'Volume',1,7)
   call add_property(pd,current,'Jm',(md%nf-1)*3,8)
   call add_property(pd,current,'Jqm',3,9)
   call add_property(pd,current,'P(a)',1,11)
   call add_property(pd,current,'P(a)',3*3,12)
   call add_property(pd,current,'P(m)',1,21)
   call add_property(pd,current,'P(m)',3*3,22)
   call add_property(pd,current,'strain rate',1,17)
   call add_property(pd,current,'strain',1,18)
   call add_property(pd,current,'strain period',1,19)

  call add_property(pd,current,'boxvec',3**2,36)

  if(da%dfarch .ne. 0) then
   call add_property(pd,current,'Udihed',1,41)
  endif

  if(bs%bstype .ne. 0) then
   call add_property(pd,current,'Ubond_stretch',1,37)
   call add_property(pd,current,'bond_len_av',1,38)
  endif

  if(bb%bbtype .ne. 0) then
   call add_property(pd,current,'Ubond_bend',1,57)
   call add_property(pd,current,'bond_ang_av',1,58)
  endif

  if(ca%csolvertype .ne. 0) then
   call add_property(pd,current,'rms-bond-error',md%nf,39)
   call add_property(pd,current,'rms-vel-error',md%nf,52)
   call add_property(pd,current,'its-gssolv',md%nf,40)
  endif

  call add_property(pd,current,'Total mom',3,42)
  call add_property(pd,current,'rcm',3,43)
  call add_property(pd,current,'pcm',3,44)
  call add_property(pd,current,'fm',3,45)

  nullify(current%next)
  call create_property_arrays(pd,root,rc%steps_this_run)
  call delete_property_list(root)

  end subroutine define_properties
!***********************************************************************
  subroutine init_1
!***********************************************************************
  integer :: error
  call rd_run_control(rc,'Run_control_input_data',1)

  call init_ensav(ead,rc%num_ens_mem)

  call rd_atoms_input(atd,'Atoms_input_data',rc%ifile,1)
  call create_atoms(atd, 3)
  call rd_species_input(spd,atd%n,'Species_input_data',rc%ifile,1)

  call integrator_input(ded,'Integrator_input_data',rc%ifile,1)
  call create_integrator(ded, atd%n, 3)

! Initialize arrays of neighbour lists and pair forces
  allocate(nl(spd%npsp), pf(spd%npsp), stat=error)
  if(error .ne. 0) stop 'init_1: allocation error for nl'
  m = 1
  do i = 1, spd%nsp
   do j = i, spd%nsp
    nl(m)%sp1 = i
    nl(m)%sp2 = j
    nl(m)%intyp = m
    m = m + 1
   enddo
  enddo
  
  do i = 1, spd%npsp
   write(kbchar,'(i2.2)') i
   call rd_pf_input(pf(i),'pair_force_parameters_'//kbchar,rc%ifile,1)
   call rd_nlist_input(nl(i),'Neighbour_list_input_data_'//kbchar, &
   & rc%ifile,1)
  enddo

  call rd_properties_input(pd,'Properties_input_data',rc%ifile,1)
  call rd_sllod_input(sd,'Sllod_input_data',rc%ifile,1)
  call rd_molecules_input(md,'Molecules_input_data',rc%ifile,1)
  call create_molecules(md, 3, spd)
  call rd_fractions_input(md,'Fractions_input_data',rc%ifile,1,spd)
  call init_molecules(atd, md, spd)

! intialise ewald data type if required
! done after init_molecules because it uses spd%spid (species id numbers)
  do i = 1, spd%npsp
   if(pf(i)%pftype .eq. 3 .or. pf(i)%pftype .eq. 4) then
    call rd_ewald(ewd,'Ewald_data',rc%ifile,1)
    call init_ewald(spd, ewd)
    exit
   endif
   if(pf(i)%pftype .eq. 5 .or. pf(i)%pftype .eq. 6) then
    call rd_dsf(ewd,'Ewald_data',rc%ifile,1)
    exit
   endif
  enddo

! transfer kappa value from Ewald input to pair forces and initialize constants
  do i = 1, spd%npsp
   if(pf(i)%pftype .eq. 3 .or. pf(i)%pftype .eq. 4) then
    call init_pf_ew(pf(i), nl(i), spd, ewd%kappa)
   elseif(pf(i)%pftype .eq. 5 .or. pf(i)%pftype .eq. 6) then
    call init_pf_dsf(pf(i), nl(i), spd, ewd%kappa)
   endif
  enddo

  call rd_constraints(ca,'Constraint_input_data',rc%ifile,1)
  call create_constraint_set(ca, md%nf)
  call init_constr(ca,md,spd,'Constraint_matrix_data',rc%ifile,1)
  call check_primary_constraints(ca, md, spd)

  call rd_dihed_input(da,md,'Dihedral_force_input_data',rc%ifile,1)
  call rd_bondstr_input(bs, md,'Bond_stretch_input_data',rc%ifile,1)
  call rd_bondbend_input(bb, md,'Bond_bend_input_data',rc%ifile,1)
  call rd_system_input(sys,md,'System_input_data',rc%ifile,1)

! initialize integrator for NH type thermostats and barostat
  sys%td%zeta = 0.d0
  sys%prd%edot = 0.d0

  if(ded%destype .eq. 1) then
   if(sys%td%thermotype .eq. 2 .or. sys%td%thermotype .eq. 3) then
    call init_gear5_scalar(ded%zder)
   endif
   if(sys%prd%barotype .eq. 2) call init_gear5_scalar(ded%eder)
  endif

! read input data and create arrays for correlation functions
  call rdinput_tcf(scfd,rc%ifile,'tcf_input_data_scfd',501)
  if(scfd%kcfm .ne. 0) then
   scfd2%lcf = scfd%lcf
   scfd2%kcf = scfd%kcf*scfd%kcfm
   scfd2%nsr = scfd%nsr
  endif
  call rdinput_tcf(jqjqcf,rc%ifile,'tcf_input_data_jqjqcf',501)
  call rdinput_tcf(vacfd,rc%ifile,'tcf_input_data_vacfd',501)

! Correlation functions for multicomponent systems
  if(md%nf .gt. 1) then
   allocate(jmjmcf((md%nf-1)**2), stat=error)
   if(error .ne. 0) stop 'wmd2: allocation error for jmjmcf'
   allocate(jmjqcf(md%nf-1), stat=error)
   if(error .ne. 0) stop 'wmd2: allocation error for jmjqcf'
   allocate(jqjmcf(md%nf-1), stat=error)
   if(error .ne. 0) stop 'wmd2: allocation error for jqjmcf'
   do i = 1, (md%nf-1)**2
    call rdinput_tcf(jmjmcf(i),rc%ifile,'tcf_input_data_jmjmcf',501)
   enddo
   do i = 1, md%nf-1
    call rdinput_tcf(jqjmcf(i),rc%ifile,'tcf_input_data_jqjmcf',501)
   enddo
   do i = 1, md%nf-1
    call rdinput_tcf(jmjqcf(i),rc%ifile,'tcf_input_data_jmjqcf',501)
   enddo
  endif

  call init_tcf
  
! initialize neighbour list
  do i = 1, spd%npsp
   if(pf(i)%pftype .ne. 0) call init_nlist(nl(i), sd, pf(i)%rcut, 3, spd)
  enddo

  call init_inter_masks
  
  call create_box(box)

  end subroutine init_1
!***********************************************************************
  subroutine init_2
!***********************************************************************

! Rescale box vectors if required
  if(rc%kboxsc .eq. -1) then
   do i = 1, 3
    box%boxvec(:,i) = box%boxvec(:,i)*rc%bf(i)
   enddo
  endif

! Calculate initial box volume and inverse box vectors
  box%volume = volume(box%boxvec)
  box%boxveci = matinv3d(transpose(box%boxvec))

! initialize volume integrator for NH type barostats
  if(ded%destype .eq. 1) then
   if(sys%prd%barotype .ne. 0) then
    sys%prd%volume = box%volume
    call init_gear5_scalar(ded%vder)
   endif
  endif

! Initialise ewald fourier space vectors and factors if required.
! This must be done *after* the box vectors are initialized
  do i = 1, spd%npsp
   if(pf(i)%pftype .eq. 3 .or. pf(i)%pftype .eq. 4) then
    call init_ewald_kvec(box%boxveci, ewd)
    exit
   endif
  enddo

  do i = 1, spd%npsp
   if(pf(i)%pftype .eq. 3 .or. pf(i)%pftype .eq. 4) then
    call init_ewald_self_bond(spd, md, ewd)
    exit
   endif
  enddo

! Zero momenta if required
  do i = 1, 3
   if(rc%kpz(i) .eq. -1) call zero_momenta_m(atd%p, md, spd, i)
  enddo
! Zero y-momentum if doing pef
  if(sd%flow_type .eq. 2) call zero_momenta_m(atd%p, md, spd, 2)

! Rescale momenta to give new input temperature if required
! Fix this for directional thermostats
   if(sys%td%ktsc .eq. -1) then
    call calc_pcm(md)
    call rescale_momenta(md%pcm, &
     & sys%td%tin, sys%td%ndfm, md%massinv,sf)
    do i = 1, md%nm
     do k = 1, md%nsf(md%fid(i))
      do j = 1, 3
       md%mol(i)%p(k,j) = md%mol(i)%p(k,j)*sf
      enddo
     enddo
    enddo
   endif

! Setup box for flow
  call setup_box(sd,box,atd,ded)

! write headers for output file and tseries file
  if(pd%ktseries .ne. 0) call wr_property_headers(pd, pd%tseries_file, 1)

! form initial neighbour list
  do i = 1, spd%npsp
   if(pf(i)%pftype .ne. 0) call form_nlist(nl(i), box, atd%r(nl(i)%aid,:), md)
  enddo

  end subroutine init_2
!***********************************************************************
  subroutine init_input_file
!***********************************************************************
  call get_command_argument(1, rc%ifile)  ! get input file name from command line
  if(len_trim(rc%ifile) .eq. 0) rc%ifile = 'nemd.in' ! default input file

  end subroutine init_input_file
!***********************************************************************
  subroutine init_output_files
!***********************************************************************
  rc%ofile = 'nemd.out'            ! output file
  if(rc%num_ens_mem .ne. 1) then
   pd%tseries_file = 'nemd.out_tseries'//ead%imlabel  ! time series data file
   rc%sfile = 'snap'//ead%imlabel//'.xyz' ! snapshot file
  else
   pd%tseries_file = 'nemd.out_tseries'  ! time series data file
   rc%sfile = 'snap.xyz' ! snapshot file
  endif

! create and close blank output files
  call init_file(rc%ofile,1)
  if(pd%ktseries .ne. 0) call init_file(pd%tseries_file,1)
  if(rc%ksnapshot .ne. 0) call init_file(rc%sfile,1)

  end subroutine init_output_files
!***********************************************************************
  subroutine init_inter_masks
! Initialize intramolecular interaction masks
! The masks are used to prevent atoms from interacting with LJ type
! pair forces
!***********************************************************************

! eliminate interactions between constrained pairs of atoms
  if(ca%csolvertype .ne. 0) call init_imask_constr(ca, md)
! eliminate interactions between atoms 1 and 4 of each dihedral interaction
  if(da%dfarch .ne. 0) call init_imask_dihed(da, md)
! eliminate interactions between atoms with a bond stretch (except FENE)
  if(bs%bstype .eq. 1) call init_imask_bondstr(bs, md)
! eliminate interactions between atoms with a bond bend
  if(bb%bbtype .ne. 0) call init_imask_bondbend(bb, md)

  end subroutine init_inter_masks
!***********************************************************************
  subroutine init_momenta
!***********************************************************************
  call initialize_momenta(md%pcm, sys%td%tin, sys%td%ndfm, md%massinv, ead)
  do i = 1, md%nm
   do k = 1, md%nsf(md%fid(i))
    do j = 1, 3
     md%mol(i)%p(k,j) = md%pcm(i,j)*spd%mass(md%fr(md%fid(i))%spid(k))*md%massinv(i)
    enddo
   enddo
  enddo

  end subroutine init_momenta
!***********************************************************************
  subroutine init_ndf
!***********************************************************************
  sys%td%ndfa = 3*md%np - 3
  sys%td%ndfm = 3*md%nm - 3
  if(sys%td%thermotype .eq. 1) then
   sys%td%ndfa = sys%td%ndfa - 1
   sys%td%ndfm = sys%td%ndfm - 1
  elseif(sys%td%thermotype .eq. 4 .or. sys%td%thermotype .eq. 6) then
   sys%td%ndfa = sys%td%ndfa - 3
   sys%td%ndfm = sys%td%ndfm - 3
  endif
! do the same for directional kinetic thermostats
  sys%td%ndfad = md%np - 1
  sys%td%ndfmd = md%nm - 1
  if(sys%td%thermotype .eq. 4 .or. sys%td%thermotype .eq. 6) then
   sys%td%ndfad = sys%td%ndfad - 1
   sys%td%ndfmd = sys%td%ndfmd - 1
  endif
! also subtract the number of degrees of freedom lost due to primary and
! secondary molecular constraints
  do j = 1, md%nf
   sys%td%ndfa = sys%td%ndfa - md%nmf(j)*ca%cset(j)%nc - &
   & md%nmf(j)*ca%cset(j)%nsec*3
  enddo

  end subroutine init_ndf
!***********************************************************************
  subroutine init_tcf
! Initialize time correlation function data
!***********************************************************************

  if(scfd%kcf .ne. 0) then
   if(rc%runtype .eq. 2) then
    call set_tcf(scfd, accum_type = 5, typ = 2, fname = 'nemd.scf'//ead%imlabel, &
    & tfname = 'nemd.out_scf')
   else
    call set_tcf(scfd, accum_type = 5, typ = 1, fname = 'nemd.scf'//ead%imlabel, &
    & tfname = 'nemd.out_scf')
   endif
   call create_tcf(scfd,rc%efile,501,502)
  endif

  if(scfd2%kcf .ne. 0) then
   if(rc%runtype .eq. 2) then
    call set_tcf(scfd2, accum_type = 5, typ = 2, fname = 'nemd.scf2'//ead%imlabel, &
    & tfname = 'nemd.out_scf2')
   else
    call set_tcf(scfd2, accum_type = 5, typ = 1, fname = 'nemd.scf2'//ead%imlabel, &
    & tfname = 'nemd.out_scf2')
   endif
   call create_tcf(scfd2,rc%efile,501,502)
  endif

  if(jqjqcf%kcf .ne. 0) then
   if(rc%runtype .eq. 2) then
    call set_tcf(jqjqcf, accum_type = 1, typ = 2, fname = 'nemd.jqjqcf'//ead%imlabel, &
    & tfname = 'nemd.out_jqjqcf')
   else
    call set_tcf(jqjqcf, accum_type = 1, typ = 1, fname = 'nemd.jqjqcf'//ead%imlabel, &
    & tfname = 'nemd.out_jqjqcf')
   endif
   call create_tcf(jqjqcf,rc%efile,501,502)
  endif

! Correlation functions for multicomponent systems only
  if(md%nf .ge. 2) then
   if(jmjmcf(1)%kcf .ne. 0) then
    do i = 1, (md%nf-1)**2
     write(kbchar,'(i2.2)') i
     if(rc%runtype .eq. 2) then
      call set_tcf(jmjmcf(i), accum_type = 6, typ = 2, fname = 'nemd.jmjmcf'//kbchar//ead%imlabel, &
      & tfname = 'nemd.out_jmjmcf'//kbchar)
     else
      call set_tcf(jmjmcf(i), accum_type = 6, typ = 1, fname = 'nemd.jmjmcf'//kbchar//ead%imlabel, &
      & tfname = 'nemd.out_jmjmcf'//kbchar)
     endif
     call create_tcf(jmjmcf(i),rc%efile,501,502)
    enddo
   endif

   if(jqjmcf(1)%kcf .ne. 0) then
    do i = 1, md%nf-1
     write(kbchar,'(i2.2)') i
     if(rc%runtype .eq. 2) then
      call set_tcf(jqjmcf(i), accum_type = 6, typ = 2, fname = 'nemd.jqjmcf'//kbchar//ead%imlabel, &
      & tfname = 'nemd.out_jqjmcf'//kbchar)
     else
      call set_tcf(jqjmcf(i), accum_type = 6, typ = 1, fname = 'nemd.jqjmcf'//kbchar//ead%imlabel, &
      & tfname = 'nemd.out_jqjmcf'//kbchar)
     endif
     call create_tcf(jqjmcf(i),rc%efile,501,502)
    enddo
   endif

   if(jmjqcf(1)%kcf .ne. 0) then
    do i = 1, md%nf-1
     write(kbchar,'(i2.2)') i
     if(rc%runtype .eq. 2) then
      call set_tcf(jmjqcf(i), accum_type = 6, typ = 2, fname = 'nemd.jmjqcf'//kbchar//ead%imlabel, &
      & tfname = 'nemd.out_jmjqcf'//kbchar)
     else
      call set_tcf(jmjqcf(i), accum_type = 6, typ = 1, fname = 'nemd.jmjqcf'//kbchar//ead%imlabel, &
      & tfname = 'nemd.out_jmjqcf'//kbchar)
     endif
     call create_tcf(jmjqcf(i),rc%efile,501,502)
    enddo
   endif
  endif
! end of multicomponent system section

  if(vacfd%kcf .ne. 0) then
   if(rc%runtype .eq. 2) then
    call set_tcf(vacfd, accum_type = 1, typ = 2, fname = 'nemd.vacf'//ead%imlabel, &
    & tfname = 'nemd.out_vacf')
   else
    call set_tcf(vacfd, accum_type = 1, typ = 1, fname = 'nemd.vacf'//ead%imlabel, &
    & tfname = 'nemd.out_vacf')
   endif
   call create_tcf(vacfd,rc%efile,501,502)
  endif

  end subroutine init_tcf
!***********************************************************************
  subroutine read_state(rc,box,atd,ded,sd,sys,fnprefix,funit,pd)
!***********************************************************************
  use run_control
  use integrator
  use gear5
  use simulation_box
  use atoms
  use properties
  use sllod
  use pdutil
  use system

  implicit none

  type(run_control_type), optional :: rc
  type(box_data_type) :: box
  type(atom_type) :: atd
  type(integrator_data_type) :: ded
  type(properties_type), optional :: pd
  type (sllod_data_type) :: sd
  type(system_data_type), optional :: sys

  character(len=*) :: fnprefix      !file name prefix
  integer :: funit

  open(unit=funit,file=trim(adjustl(fnprefix))//'.state'//ead%imlabel, &
      & form='unformatted',status='old',position='rewind')
  if(present(rc)) call rd_run_control_state(rc,funit)
  call rd_box_state(box,funit)
  call rd_atom_state(atd,funit)
  call read_integrator(ded,funit)
  call rd_sllod_state(sd,funit)
  if(present(sys)) call rd_system_state(sys,funit)
  close(funit)

  if(present(pd)) then
   open(unit=funit,file=trim(adjustl(fnprefix))//'.properties'//ead%imlabel, &
       & form='unformatted',status='old',position='rewind')
   call rd_properties(pd,funit)
   close(funit)
  endif

  end subroutine read_state
!***********************************************************************
  subroutine save_state_files
!***********************************************************************
  call write_state(rc,box,atd,ded,sd,sys,'nemd',2,pd)
  if(scfd%kcf .ne. 0) call store_tcf(scfd,500)
  if(scfd2%kcf .ne. 0) call store_tcf(scfd2,500)
  if(jqjqcf%kcf .ne. 0) call store_tcf(jqjqcf,500)
  if(md%nf .gt. 1) then
   do i = 1, (md%nf-1)**2
    if(jmjmcf(i)%kcf .ne. 0) call store_tcf(jmjmcf(i),500)
   enddo
   do i = 1, md%nf-1
    if(jqjmcf(i)%kcf .ne. 0) call store_tcf(jqjmcf(i),500)
   enddo
   do i = 1, md%nf-1
    if(jmjqcf(i)%kcf .ne. 0) call store_tcf(jmjqcf(i),500)
   enddo
  endif

  if(vacfd%kcf .ne. 0) call store_tcf(vacfd,500)

  end subroutine save_state_files
!***********************************************************************
  subroutine setup_box(sd,box,atd,ded)
!***********************************************************************
  use integrator
  use simulation_box
  use atoms
  use sllod
  use v_sub_library

  implicit none

  type(box_data_type) :: box
  type(atom_type) :: atd
  type(integrator_data_type) :: ded
  type (sllod_data_type) :: sd

  if(sd%flow_type .ne. 0 .and. box%state .eq. 0) then
   select case(sd%flow_type)
    case(1)
     call init_sllod(sd, box%boxvec)
     box%state = 1
    case(2)
     call init_pef(sd, 3, 2, box%angle)
     call xy_array_rotate(box%boxvec(:,1:2), box%angle)
     call xy_array_rotate(atd%r(:,1:2), box%angle)
     call xy_array_rotate(atd%p(:,1:2), box%angle)
     select case(ded%destype)
     case(1)
      call xy_array_rotate(ded%rder%d1(:,1:2), box%angle)
      call xy_array_rotate(ded%rder%d2(:,1:2), box%angle)
      call xy_array_rotate(ded%rder%d3(:,1:2), box%angle)
      call xy_array_rotate(ded%rder%d4(:,1:2), box%angle)
      call xy_array_rotate(ded%pder%d1(:,1:2), box%angle)
      call xy_array_rotate(ded%pder%d2(:,1:2), box%angle)
      call xy_array_rotate(ded%pder%d3(:,1:2), box%angle)
      call xy_array_rotate(ded%pder%d4(:,1:2), box%angle)
     end select
     box%state = 2
    case(3)
     ! bulk deformation - do nothing
    case default
     stop 'setup_box: Invalid flow type'
   end select
   box%boxvec0 = box%boxvec
   sd%strain = 0.d0
  endif

  if(sd%flow_type .eq. 1 .and. box%state .ne. 1) &
   stop 'setup_box: Box not initialized for shear'
  if(sd%flow_type .eq. 2 .and. box%state .ne. 2) &
   stop 'setup_box: Box not initialized for elong'

  end subroutine setup_box
!***********************************************************************
  subroutine start_new_state
!***********************************************************************
! build crystal or read coordinate file
   if(sys%coordinate_input .eq. 0) then
    call start_from_crystal
   elseif(sys%coordinate_input .eq. 1) then
    call rd_config(atd%r, box, rc%rsc, 'config.xyz', 1)
    call calc_rcm(md, spd)
   endif

   if(rc%ksnapshot .ne. 0) call wr_snapshot(atd%r, box%boxvec, spd%label(spd%spid), rc, 3)

! initialize number of degrees of freedom
   call init_ndf
! initialize momenta
   call init_momenta
   
! set up property arrays
   call define_properties
   call zero_properties(pd)

   call zero_nstep(rc)

  end subroutine start_new_state
!***********************************************************************
  subroutine start_saved_state
! start simulation from a saved state
!***********************************************************************

! initialize number of degrees of freedom
   call init_ndf

! set up array of property names
   call define_properties

   if(rc%runtype .eq. 2) then
    call read_state(rc,box,atd,ded,sd,sys,'nemd',2,pd)
   else
    call read_state(rc,box,atd,ded,sd,sys,'nemd',2)
    call zero_properties(pd)
    call zero_nstep(rc)
   endif

  end subroutine start_saved_state
!***********************************************************************
  subroutine start_from_crystal
!***********************************************************************
  call init_lattice(sys%lat)

! make a lattice with a point for each molecule
! note - build_system also initializes box vectors
  call build_system(sys, box, md%rcm)

! put the molecule on each lattice point
  do i = 1, md%nm
   select case(ca%cset(md%fid(i))%mol_arch)
    case(0)
     md%mol(i)%r = 0.d0
    case(1)
     md%mol(i)%r = ca%cset(md%fid(i))%mol_template
    case(2)
     call fjc_molecule_maker(md%mol(i)%r, md%nsf(md%fid(i)), 1.d0)
    case(3)
     call linalk_molecule_maker(md%mol(i)%r, md%nsf(md%fid(i)), 0.39d0, &
     & 109.47d0, da%dai)
    case default
     stop 'nemd: Invalid value of mol_arch'
   end select
  enddo

  do i = 1, md%nf
   if(allocated(ca%cset(i)%mol_template)) deallocate(ca%cset(i)%mol_template)
  enddo

  do i = 1, md%nm
   do j = 1, md%nsf(md%fid(i))
    md%mol(i)%r(j,:) = md%mol(i)%r(j,:) + md%rcm(i,:)
   enddo
  enddo

  end subroutine start_from_crystal
!***********************************************************************
  subroutine write_output
! Write averaged output to text files.
! Only done at the end of each run.
! Modified to allow for coarray ensemble averaging.
!***********************************************************************

! write property averages to output file
! copy data to temporary coarrays and ensemble average into image 1 blocks
   if(ead%ni .gt. 1) then
    call eav_blocks(ead,pd)
   endif
! write block averages and overall averages to output file
   if(ead%im .eq. 1) then
    call wr_property_headers(pd, rc%ofile, 1)
    call wr_property_blockaverages(pd,ded%tstep,rc%ofile,1)
    call wr_property_headers(pd, rc%ofile, 1)
    call wr_property_averages(pd,ded%tstep,rc%ofile,1)
   endif

! write correlation functions and integrals to output file
   if(scfd%kcf .ne. 0) then
    scfd%cfnorm=5.d0/box%volume
    scfd%cfintfact = dble(ded%tstep/sys%td%tin)
    call write_tcf(scfd,ead,ded%tstep, &
     'Shear stress autocorrelation function',29)
   endif

   if(scfd2%kcf .ne. 0) then
    scfd2%cfnorm=5.d0/box%volume
    scfd2%cfintfact = dble(ded%tstep/sys%td%tin)
    call write_tcf(scfd2,ead,ded%tstep, &
     'Shear stress autocorrelation function',29)
   endif

   if(jqjqcf%kcf .ne. 0) then
    jqjqcf%cfnorm=3.d0/box%volume
    jqjqcf%cfintfact = dble(ded%tstep/sys%td%tin**2)
    call write_tcf(jqjqcf,ead,ded%tstep,'Heat flux autocorrelation function',29)
   endif

! Multicomponent correlation functions
   if(md%nf .gt. 1) then
   do i = 1, (md%nf-1)**2
    if(jmjmcf(i)%kcf .ne. 0) then
     jmjmcf(i)%cfnorm=3.d0/box%volume
     jmjmcf(i)%cfintfact = dble(ded%tstep/sys%td%tin)
     write(kbchar,'(i2.2)') i
     call write_tcf(jmjmcf(i),ead,ded%tstep,'Mass flux correlation function_'//kbchar,29)
    endif
   enddo
   do i = 1, md%nf-1
    if(jqjmcf(i)%kcf .ne. 0) then
     jqjmcf(i)%cfnorm=3.d0/box%volume
     jqjmcf(i)%cfintfact = dble(ded%tstep/sys%td%tin)
     write(kbchar,'(i2.2)') i
     call write_tcf(jqjmcf(i),ead,ded%tstep, &
     & 'Heat-mass flux cross correlation function_'//kbchar,29)
    endif
   enddo
   do i = 1, md%nf-1
    if(jmjqcf(i)%kcf .ne. 0) then
     jmjqcf(i)%cfnorm=3.d0/box%volume
     jmjqcf(i)%cfintfact = dble(ded%tstep/sys%td%tin)
     write(kbchar,'(i2.2)') i
     call write_tcf(jmjqcf(i),ead,ded%tstep, &
     & 'Mass - heat flux cross correlation function_'//kbchar,29)
    endif
   enddo
   endif
! End of multicomponent correlation functions

   if(vacfd%kcf .ne. 0) then
    vacfd%cfnorm=dble(vacfd%nsr)
    vacfd%cfintfact = dble(ded%tstep)
    call write_tcf(vacfd,ead,ded%tstep,'Velocity autocorrelation function',29)
   endif

  end subroutine write_output
!***********************************************************************
  subroutine write_state(rc,box,atd,ded,sd,sys,fnprefix,funit,pd)
!***********************************************************************
  use run_control
  use integrator
  use simulation_box
  use gear5
  use atoms
  use properties
  use sllod
  use system

  implicit none

  type(run_control_type), optional :: rc
  type(box_data_type) :: box
  type(atom_type) :: atd
  type(integrator_data_type) :: ded
  type(properties_type), optional :: pd
  type (sllod_data_type), intent(in) :: sd
  type(system_data_type), optional :: sys

  character(len=*) :: fnprefix      !file name prefix
  integer :: funit

  open(unit=funit,file=trim(adjustl(fnprefix))//'.state'//ead%imlabel, &
      & form='unformatted',status='replace',position='rewind')
  if(present(rc)) call store_run_control_state(rc,funit)
  call store_box_state(box,funit)
  call store_atom_state(atd,funit)
  call store_integrator(ded,funit)
  call store_sllod_state(sd,funit)
  if(present(sys)) call store_system_state(sys,funit)
  close(funit)

  if(present(pd)) then
   open(unit=funit,file=trim(adjustl(fnprefix))//'.properties'//ead%imlabel, &
       & form='unformatted',status='replace',position='rewind')
   call store_properties(pd,funit)
   close(funit)
  endif

  end subroutine write_state
!***********************************************************************
  end program nemd
!***********************************************************************
