  module pair_force
!***********************************************************************
! Calculates the force and energy between pairs of
! particles for a variety of potentials.
! pftype = 0          No force
! pftype = 1          Lennard-Jones
! pftype = 2          SHREP (see Hess & Kroger)
! pftype = 3          truncated electrostatic with Ewald summation
! pftype = 4          truncated electrostatic with Ewald plus LJ
! pftype = 5          damped and shifted force electrostatic (Wolf method)
! pftype = 6          damped and shifted force electrostatic plus LJ (Wolf plus LJ)
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************
 use neighbour_list
 use simulation_box
 use species
 use ewald
! use pgi_erf

 implicit none

 private

! Public procedures
 public :: calculate_pf, init_pf_ew, init_pf_dsf, rd_pf_input

! Public data types
 public :: pf_data_type, LJ_data_type, SHREP_data_type, es_ew_data_type, &
 & es_dsf_data_type

 type :: pf_data_type
  double precision, dimension(:,:), allocatable :: rij, fij
  double precision, dimension(:), allocatable :: uij, rijsq
  double precision :: rcut, rcutsq
  integer :: pftype
  type (LJ_data_type), pointer :: ljd
  type (SHREP_data_type), pointer :: shrepd
  type (es_ew_data_type), pointer :: esd
  type (es_dsf_data_type), pointer :: dsfd
 end type pf_data_type

 type :: LJ_data_type
  double precision :: epsilon, sigma, rcut, core
  double precision :: sigmasq, c1, c2, c3, ushift, ulj
  logical :: do_shift, use_core
 end type LJ_data_type

 type :: SHREP_data_type
  double precision :: epsilon, sigma, rcut
  double precision :: sigmasq, c2, c1
 end type SHREP_data_type

 type :: es_ew_data_type
  double precision :: kappa, rcut, qiqj, kapsq, c1, c2, ues
  double precision :: pes
 end type es_ew_data_type

! Data type for electrostatic damped and shifted force type potential
 type :: es_dsf_data_type
  double precision :: kappa, rcut, qiqj, kapsq, c1, c2, c3, ues
  double precision :: pes
 end type es_dsf_data_type

 contains
!***********************************************************************
  subroutine calculate_pf(nl, ratom, box, pf)
!***********************************************************************
  type (pf_data_type), intent(inout) :: pf
  type (box_data_type), intent(inout) :: box
  type (nlist_data_type), intent(inout) :: nl
  double precision, dimension(:,:) :: ratom
  integer :: i, j, k, ndim
  
  ndim = size(ratom,2)
  
! find particle separations
  call find_pair_separations(nl%nlist, box, ratom, pf%rcutsq, &
  & nl%nmask, nl%rij, nl%rijsq, nl%inab)

  call compress_nlist(nl)

! Allocate arrays for pair forces and energies,
! only if they do not exist or if the neighbour list has changed size.
  if(allocated(pf%rij)) then
   if(nl%nab .ne. size(pf%rij,1)) then
    call create_pf(pf, nl%nab, ndim)
   endif
  else
   call create_pf(pf, nl%nab, ndim)
  endif

  j = 0
  do i = 1, nl%nab
   if(nl%nmask(i)) then
    j = j + 1
    pf%rijsq(j) = nl%rijsq(i)
    do k = 1, ndim
     pf%rij(j,k) = nl%rij(i,k)
    enddo
   endif
  enddo 

  if(nl%inab .ne. 0) then
 
  select case(pf%pftype)

  case(1)  ! Lennard-Jones
   call LJ_force_energy(pf%rij(1:nl%inab,:), pf%rijsq(1:nl%inab), &
   & pf%ljd, pf%fij(1:nl%inab,:), pf%uij(1:nl%inab))
  case(2)  ! SHREP
   call SHREP_force_energy(pf%rij(1:nl%inab,:),pf%rijsq(1:nl%inab), &
   & pf%shrepd,pf%fij(1:nl%inab,:),pf%uij(1:nl%inab))
  case(3)  ! electrostatic by Ewald summation
   call es_ew_force_energy(pf%rij(1:nl%inab,:), pf%rijsq(1:nl%inab), &
   & pf%esd, pf%fij(1:nl%inab,:), pf%uij(1:nl%inab))
  case(4)  ! electrostatic by Ewald summation plus Lennard-Jones
   call es_ew_LJ_force_energy(pf%rij(1:nl%inab,:), pf%rijsq(1:nl%inab), &
   & pf%ljd, pf%esd, pf%fij(1:nl%inab,:), pf%uij(1:nl%inab))
  case(5)  ! electrostatic by damped and shifted force
   call es_dsf_force_energy(pf%rij(1:nl%inab,:), pf%rijsq(1:nl%inab), &
   & pf%dsfd, pf%fij(1:nl%inab,:), pf%uij(1:nl%inab))
  case(6)  ! electrostatic by damped and shifted force plus Lennard-Jones
   call es_dsf_LJ_force_energy(pf%rij(1:nl%inab,:), pf%rijsq(1:nl%inab), &
   & pf%ljd, pf%dsfd, pf%fij(1:nl%inab,:), pf%uij(1:nl%inab))

  end select
  
  endif

  end subroutine calculate_pf
!***********************************************************************
  subroutine init_pf_ew(pf,nl,spd,kappa)
!***********************************************************************
  type (pf_data_type) :: pf
  type(species_set) :: spd
  type(nlist_data_type) :: nl
  double precision :: kappa
  double precision :: pi

  pi = 4.d0*atan(1.d0)

  select case(pf%pftype)

  case(3)  ! electrostatic by Ewald summation
   pf%esd%qiqj = spd%charge(nl%sp1)*spd%charge(nl%sp2)
   pf%esd%kappa = kappa
   pf%esd%kapsq = kappa*kappa
   pf%esd%c1 = 2.d0*kappa/sqrt(pi)
  case(4)  ! electrostatic by Ewald summation plus Lennard-Jones
   pf%esd%qiqj = spd%charge(nl%sp1)*spd%charge(nl%sp2)
   pf%esd%kappa = kappa
   pf%esd%kapsq = kappa*kappa
   pf%esd%c1 = 2.d0*kappa/sqrt(pi)
  case default
   stop 'pair_force.init_pf_ew: invalid value of pftype'
  end select

  end subroutine init_pf_ew
!***********************************************************************
  subroutine init_pf_dsf(pf,nl,spd,kappa)
!***********************************************************************
  type (pf_data_type) :: pf
  type(species_set) :: spd
  type(nlist_data_type) :: nl
  double precision :: kappa, rc, rcsq
  double precision :: pi

  pi = 4.d0*atan(1.d0)

  select case(pf%pftype)

  case(5)  ! electrostatic by damped and shifted force
   pf%dsfd%qiqj = spd%charge(nl%sp1)*spd%charge(nl%sp2)
   rc = pf%dsfd%rcut
   rcsq = rc*rc
   pf%dsfd%kappa = kappa
   pf%dsfd%kapsq = kappa*kappa
   pf%dsfd%c1 = 2.d0*kappa/sqrt(pi)
   pf%dsfd%c2 = erfc(kappa*rc)/rc
   pf%dsfd%c3 = erfc(kappa*rc)/rcsq + pf%dsfd%c1*exp(-pf%dsfd%kapsq*rcsq)/rc
  case(6)  ! electrostatic by damped and shifted force plus Lennard-Jones
   pf%dsfd%qiqj = spd%charge(nl%sp1)*spd%charge(nl%sp2)
   rc = pf%dsfd%rcut
   rcsq = rc*rc
   pf%dsfd%kappa = kappa
   pf%dsfd%kapsq = kappa*kappa
   pf%dsfd%c1 = 2.d0*kappa/sqrt(pi)
   pf%dsfd%c2 = erfc(kappa*rc)/rc
   pf%dsfd%c3 = erfc(kappa*rc)/rcsq + pf%dsfd%c1*exp(-pf%dsfd%kapsq*rcsq)/rc
  case default
   stop 'pair_force.init_pf_dsf: invalid value of pftype'
  end select

  end subroutine init_pf_dsf
!***********************************************************************
  subroutine rd_pf_input(pf,header,ifile,iunit)
!***********************************************************************
! Reads Lennard-Jones potential variables from input file.
! 
!***********************************************************************
  type (pf_data_type), intent(inout) :: pf
  
  integer :: i, j, iunit, error
  
  character(len=*) :: header
  character(len=32) :: rdstring
  character(len=32) :: ifile        ! input file name

  rdstring = ''
  i = 0
  open(iunit,file=ifile,status='old',form='formatted', &
     & position='rewind')
  do j = 1, 1000
   read(iunit,*, iostat=i) rdstring
   if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 999) exit
  enddo

  if(rdstring .eq. header) then
   read(iunit,*) pf%pftype
   select case (pf%pftype)
   case(1) ! LJ potential
    allocate(pf%ljd, stat=error)
    if(error .ne. 0) stop 'allocation error in rd_pf_input for pf%ljd'
    read(iunit,*) pf%ljd%epsilon
    read(iunit,*) pf%ljd%sigma
    read(iunit,*) pf%ljd%rcut
    read(iunit,*) pf%ljd%core
    read(iunit,*) pf%ljd%do_shift
    read(iunit,*) pf%ljd%use_core
   case(2) ! SHREP potential
    allocate(pf%shrepd, stat=error)
    if(error .ne. 0) stop 'allocation error in rd_pf_input for pf%shrepd'
    read(iunit,*) pf%shrepd%epsilon        ! epsilon
    read(iunit,*) pf%shrepd%sigma          ! sigma
    read(iunit,*) pf%shrepd%rcut           ! rcut
   case(3) ! Coulomb potential
    allocate(pf%esd, stat=error)
    if(error .ne. 0) stop 'allocation error in rd_pf_input for pf%esd'
    read(iunit,*) pf%esd%rcut           ! electrostatic force cutoff
   case(4) ! Coulomb plus LJ potential
    allocate(pf%ljd, stat=error)
    if(error .ne. 0) stop 'allocation error in rd_pf_input for pf%ljd'
    allocate(pf%esd, stat=error)
    if(error .ne. 0) stop 'allocation error in rd_pf_input for pf%esd'
    read(iunit,*) pf%ljd%epsilon
    read(iunit,*) pf%ljd%sigma
    read(iunit,*) pf%ljd%rcut
    read(iunit,*) pf%ljd%core
    read(iunit,*) pf%ljd%do_shift
    read(iunit,*) pf%ljd%use_core
    read(iunit,*) pf%esd%rcut           ! electrostatic force cutoff
   case(5) ! Coulomb plus LJ potential
    allocate(pf%dsfd, stat=error)
    if(error .ne. 0) stop 'allocation error in rd_pf_input for pf%dsfd'
    read(iunit,*) pf%dsfd%rcut           ! electrostatic force cutoff
   case(6) ! Coulomb plus LJ potential
    allocate(pf%ljd, stat=error)
    if(error .ne. 0) stop 'allocation error in rd_pf_input for pf%ljd'
    allocate(pf%dsfd, stat=error)
    if(error .ne. 0) stop 'allocation error in rd_pf_input for pf%dsfd'
    read(iunit,*) pf%ljd%epsilon
    read(iunit,*) pf%ljd%sigma
    read(iunit,*) pf%ljd%rcut
    read(iunit,*) pf%ljd%core
    read(iunit,*) pf%ljd%do_shift
    read(iunit,*) pf%ljd%use_core
    read(iunit,*) pf%dsfd%rcut           ! electrostatic force cutoff
   end select
   close(iunit)
  else
   close(iunit)
   stop 'pair_force:rd_pf_input: Pair-force input data not found'
  endif

! Initialize pf constants

  select case(pf%pftype)
  case(0)          ! no pair force - do nothing
  case(1)          ! Lennard-Jones
   if(.not. pf%ljd%use_core) pf%ljd%core = 0.d0
   pf%ljd%sigmasq = pf%ljd%sigma*pf%ljd%sigma
   pf%ljd%c1 = 4.d0*pf%ljd%epsilon
   pf%ljd%c2 = 48.d0*pf%ljd%epsilon
   pf%ljd%c3 = (pf%ljd%sigma/pf%ljd%rcut)**6
   if (pf%ljd%do_shift) then
    pf%ljd%ushift = pf%ljd%c1*pf%ljd%c3*(pf%ljd%c3 - 1.d0)
   else
    pf%ljd%ushift = 0.d0
   endif
   pf%rcut = pf%ljd%rcut ! this is the cutof used by the neighbour list
   pf%rcutsq = pf%ljd%rcut*pf%ljd%rcut
  case(2)          ! shrep
   pf%rcut = pf%shrepd%rcut ! this is the cutoff used by the neighbour list
   pf%rcutsq = pf%shrepd%rcut*pf%shrepd%rcut
   pf%shrepd%c1 = pf%shrepd%epsilon
   pf%shrepd%c2 = 24.d0*pf%shrepd%epsilon/pf%shrepd%sigma
  case(3)          ! electrostatic - Ewald
   pf%rcut = pf%esd%rcut ! this is the cutoff used by the neighbour list
   pf%rcutsq = pf%esd%rcut*pf%esd%rcut
  case(4)          ! electrostatic - Ewald plus LJ
   if(.not. pf%ljd%use_core) pf%ljd%core = 0.d0
   pf%ljd%sigmasq = pf%ljd%sigma*pf%ljd%sigma
   pf%ljd%c1 = 4.d0*pf%ljd%epsilon
   pf%ljd%c2 = 48.d0*pf%ljd%epsilon
   pf%ljd%c3 = (pf%ljd%sigma/pf%ljd%rcut)**6
   if (pf%ljd%do_shift) then
    pf%ljd%ushift = pf%ljd%c1*pf%ljd%c3*(pf%ljd%c3 - 1.d0)
   else
    pf%ljd%ushift = 0.d0
   endif
   pf%rcut = max(pf%esd%rcut, pf%ljd%rcut)  ! this is the cutoff used by the neighbour list
   pf%rcutsq = pf%rcut*pf%rcut
  case(5)          ! electrostatic - damped and shifted force
   pf%rcut = pf%dsfd%rcut ! this is the cutoff used by the neighbour list
   pf%rcutsq = pf%dsfd%rcut*pf%dsfd%rcut
  case(6)          ! electrostatic - electrostatic damped and shifted force plus LJ
   if(.not. pf%ljd%use_core) pf%ljd%core = 0.d0
   pf%ljd%sigmasq = pf%ljd%sigma*pf%ljd%sigma
   pf%ljd%c1 = 4.d0*pf%ljd%epsilon
   pf%ljd%c2 = 48.d0*pf%ljd%epsilon
   pf%ljd%c3 = (pf%ljd%sigma/pf%ljd%rcut)**6
   if (pf%ljd%do_shift) then
    pf%ljd%ushift = pf%ljd%c1*pf%ljd%c3*(pf%ljd%c3 - 1.d0)
   else
    pf%ljd%ushift = 0.d0
   endif
   pf%rcut = max(pf%dsfd%rcut, pf%ljd%rcut)  ! this is the cutoff used by the neighbour list
   pf%rcutsq = pf%rcut*pf%rcut
  case default
   stop 'pair_force:rd_pf_input: Invalid pair force type value'
  end select

  end subroutine rd_pf_input
!***********************************************************************
  subroutine create_pf(pf, n, d)
!***********************************************************************
  type (pf_data_type), intent(inout) :: pf
  integer :: error, n, d

! Allocate arrays for pair forces and energies
  call destroy_pf(pf)
  allocate(pf%rij(n,d), pf%fij(n,d), pf%uij(n), pf%rijsq(n), stat=error)
  if(error .ne. 0) stop 'pair_force.create_pf: allocation error'

  end subroutine create_pf
!***********************************************************************
  subroutine destroy_pf(pf)
!***********************************************************************
  type (pf_data_type), intent(inout) :: pf

! Deallocate arrays for pair interaction separations,
! forces and energies
  if(allocated(pf%rij)) deallocate(pf%rij)
  if(allocated(pf%fij)) deallocate(pf%fij)
  if(allocated(pf%uij)) deallocate(pf%uij)
  if(allocated(pf%rijsq)) deallocate(pf%rijsq)

  end subroutine destroy_pf
!***********************************************************************
 subroutine es_ew_force_energy(rij,rijsq,esd,fij,uij)
! Force and energy calculation for electrostatic interaction by
! Ewald summation.
!***********************************************************************
 double precision, dimension(:,:) :: rij, fij
 double precision, dimension(:) :: rijsq, uij
 double precision, dimension(size(rij,1)) :: t1, t2
 type(es_ew_data_type) :: esd
 integer :: i, ndim
 
 ndim = size(rij,2)

! Calculate pair forces and energies

! Potential energy
  t1 = sqrt(rijsq)

  t2 = esd%qiqj*erfc(esd%kappa*t1)/t1
  uij = t2
  esd%ues = sum(uij)

! Force. Assumes that rij = rj - ri.

  do i = 1, ndim
   fij(:,i) = -(t2 + esd%qiqj*esd%c1*exp(-esd%kapsq*rijsq))*rij(:,i)/rijsq
  enddo

 end subroutine es_ew_force_energy
!***********************************************************************
 subroutine es_dsf_force_energy(rij,rijsq,dsfd,fij,uij)
! Force and energy calculation for electrostatic interaction by
! the damped and shifted force method.
!***********************************************************************
 double precision, dimension(:,:) :: rij, fij
 double precision, dimension(:) :: rijsq, uij
 double precision, dimension(size(rij,1)) :: t2, t3
 type(es_dsf_data_type) :: dsfd
 double precision :: t1, c1
 integer :: i, j, ndim, np
 
 ndim = size(rij,2)
 np = size(rij,1)
 t3 = sqrt(rijsq)
 t2 = erfc(dsfd%kappa*t3)
 c1 = dsfd%qiqj*(dsfd%c2 + dsfd%c3*dsfd%rcut)

!DIR$ vector always
 do i = 1, np
! ES Potential energy
  uij(i) = dsfd%qiqj*(t2(i)/t3(i) + dsfd%c3*t3(i)) - c1
  dsfd%ues = dsfd%ues + uij(i)

! ES Force. Assumes that rij = rj - ri.
! dsfd%pes = 0.d0

  t1 = dsfd%qiqj*(-t2(i)/rijsq(i) - dsfd%c1*exp(-dsfd%kapsq*rijsq(i))/t3(i) + dsfd%c3)/t3(i)

!DIR$ unroll = 3
  do j = 1, 3
   fij(i,j) = t1*rij(i,j)
!  dsfd%pes = dsfd%pes - sum(t2*rij(:,i))
  enddo
 enddo

! dsfd%pes = dsfd%pes/3.d0

 end subroutine es_dsf_force_energy
!***********************************************************************
 subroutine es_ew_LJ_force_energy_1(rij,rijsq,ljd,esd,fij,uij)
! Force and energy calculation for electrostatic interaction by
! Ewald summation.
!***********************************************************************
 double precision, dimension(:,:) :: rij, fij
 double precision, dimension(:) :: rijsq, uij
 double precision, dimension(size(rij,1)) :: t1, t2, t3
 type(LJ_data_type) :: ljd
 type(es_ew_data_type) :: esd
 integer :: i, ndim
 
 ndim = size(rij,2)
 t3 = sqrt(rijsq)

! Calculate pair forces and energies
 if(ljd%use_core) then
  t1 = (sqrt(ljd%sigmasq)/(t3 - ljd%core))**6    !26-10-06 sigmasq replaced by sqrt(sigmasq)
 else
  t1 = (ljd%sigmasq/rijsq)**3
 endif

! LJ Potential energy
 uij = ljd%c1*t1*(t1 - 1.d0) - ljd%ushift
 ljd%ulj = sum(uij)

 t1 = ljd%c2*t1*(0.5d0 - t1)/rijsq  !t1 is now equal to (dphi/dr)/r

! LJ  Force. Assumes that rij = rj - ri.
 do i = 1, ndim
  fij(:,i) = t1*rij(:,i)
 enddo

! ES Potential energy
 t1 = esd%qiqj*erfc(esd%kappa*t3)/t3
 uij = uij + t1
 esd%ues = sum(t1)

! ES Force. Assumes that rij = rj - ri.
 esd%pes = 0.d0

 t2 = esd%qiqj*esd%c1*exp(-esd%kapsq*rijsq)

 do i = 1, ndim
  t3 = (t1 + t2)*rij(:,i)/rijsq
  fij(:,i) = fij(:,i) - t3
  esd%pes = esd%pes + sum(t3*rij(:,i))
 enddo

 esd%pes = esd%pes/3.d0

 end subroutine es_ew_LJ_force_energy_1
!***********************************************************************
 subroutine es_dsf_LJ_force_energy_1(rij,rijsq,ljd,dsfd,fij,uij)
! Force and energy calculation for electrostatic interaction by
! damped shifted force method.
!***********************************************************************
 double precision, dimension(:,:) :: rij, fij
 double precision, dimension(:) :: rijsq, uij
 double precision, dimension(size(rij,1)) :: t1, t2, t3
 type(LJ_data_type) :: ljd
 type(es_dsf_data_type) :: dsfd
 integer :: i, ndim
 
 ndim = size(rij,2)
 t3 = sqrt(rijsq)

! Calculate pair forces and energies
 if(ljd%use_core) then
  t1 = (sqrt(ljd%sigmasq)/(t3 - ljd%core))**6    !26-10-06 sigmasq replaced by sqrt(sigmasq)
 else
  t1 = (ljd%sigmasq/rijsq)**3
 endif

! LJ Potential energy
 uij = ljd%c1*t1*(t1 - 1.d0) - ljd%ushift
 ljd%ulj = sum(uij)

! set t1 equal to (dphi/dr)/r
 t1 = ljd%c2*t1*(0.5d0 - t1)/rijsq

! LJ  Force. Assumes that rij = rj - ri.
 do i = 1, ndim
  fij(:,i) = t1*rij(:,i)
 enddo

! ES Potential energy
! t1 = dsfd%qiqj*(erfc(dsfd%kappa*t3)/t3 - dsfd%c2 + dsfd%c3*(t3 - dsfd%rcut))
 t1 = dsfd%qiqj*(erfc(dsfd%kappa*t3)/t3 - dsfd%c2 + dsfd%c3*(t3 - dsfd%rcut))
 uij = uij + t1
 dsfd%ues = sum(t1)

! ES Force. Assumes that rij = rj - ri.
 dsfd%pes = 0.d0

 t1 = dsfd%qiqj*(-erfc(dsfd%kappa*t3)/rijsq - dsfd%c1*exp(-dsfd%kapsq*rijsq)/t3 + dsfd%c3)/t3

 do i = 1, ndim
  t2 = t1*rij(:,i)
  fij(:,i) = fij(:,i) + t2
  dsfd%pes = dsfd%pes - sum(t2*rij(:,i))
 enddo

 dsfd%pes = dsfd%pes/3.d0

 end subroutine es_dsf_LJ_force_energy_1
!***********************************************************************
 subroutine es_ew_LJ_force_energy(rij,rijsq,ljd,esd,fij,uij)
! Force and energy calculation for electrostatic interaction by
! Ewald summation.
!***********************************************************************
 double precision, dimension(:,:) :: rij, fij
 double precision, dimension(:) :: rijsq, uij
 double precision, dimension(size(rij,1)) :: dij, erij, eij
 double precision :: t1, t2, t3, dcij, sigma, pes
 type(LJ_data_type) :: ljd
 type(es_ew_data_type) :: esd
 integer :: i, j, npair
 
 npair = size(rij,1)
 pes = 0.d0
 esd%ues = 0.d0
 dij = sqrt(rijsq)
 erij = erfc(esd%kappa*dij)/dij
 eij = exp(-esd%kapsq*rijsq)

! Calculate pair forces and energies

  if(ljd%use_core) then
   sigma = sqrt(ljd%sigmasq)
!dir$ vector always
   do i = 1, npair
    dcij = dij(i) - ljd%core
    t1 = (sigma/dcij)**6
    t2 = ljd%c2*t1*(0.5d0 - t1)/dcij  !t2 is equal to (dphi/dr)
! Potential energy
    uij(i) = ljd%c1*t1*(t1 - 1.d0) - ljd%ushift
! Force. Assumes that rij = rj - ri.
!dir$ novector
    do j = 1, 3
     fij(i,j) = t2*rij(i,j)/dij(i)
    enddo
! ES Potential energy
    t1 = esd%qiqj*erij(i)
    uij(i) = uij(i) + t1
    esd%ues = esd%ues + t1
! ES Force. Assumes that rij = rj - ri.
    t2 = esd%qiqj*esd%c1*eij(i)
!dir$ novector
    do j = 1, 3
     t3 = (t1 + t2)*rij(i,j)/rijsq(i)
     fij(i,j) = fij(i,j) - t3
     pes = pes + t3*rij(i,j)
    enddo
   enddo

  else

!dir$ vector always
   do i = 1, npair
    t1 = (ljd%sigmasq/rijsq(i))**3
    t2 = ljd%c2*t1*(0.5d0 - t1)/rijsq(i)  !t2 is equal to (dphi/dr)/r
! Potential energy
    uij(i) = ljd%c1*t1*(t1 - 1.d0) - ljd%ushift
! Force. Assumes that rij = rj - ri.
!dir$ novector
    do j = 1, 3
     fij(i,j) = t2*rij(i,j)
    enddo
! ES Potential energy
    t1 = esd%qiqj*erij(i)
    uij(i) = uij(i) + t1
    esd%ues = esd%ues + t1
! ES Force. Assumes that rij = rj - ri.
    t2 = esd%qiqj*esd%c1*eij(i)
!dir$ novector
    do j = 1, 3
     t3 = (t1 + t2)*rij(i,j)/rijsq(i)
     fij(i,j) = fij(i,j) - t3
     pes = pes + t3*rij(i,j)
    enddo
   enddo
  endif

 esd%pes = pes/3.d0

 end subroutine es_ew_LJ_force_energy
!***********************************************************************
 subroutine es_dsf_LJ_force_energy(rij,rijsq,ljd,dsfd,fij,uij)
! Force and energy calculation for electrostatic interaction by
! damped shifted force method.
!***********************************************************************
 double precision, dimension(:,:) :: rij, fij
 double precision, dimension(:) :: rijsq, uij
 double precision, dimension(size(rij,1)) :: dij, erij, eij
 double precision :: t1, t2, dcij, sigma, pes
 type(LJ_data_type) :: ljd
 type(es_dsf_data_type) :: dsfd
 integer :: i, j, ndim, npair

 ndim = size(rij,2)
 npair = size(rij,1)
 pes = 0.d0
 dsfd%ues = 0.d0
 dij = sqrt(rijsq)
 erij = erfc(dsfd%kappa*dij)
 eij = exp(-dsfd%kapsq*rijsq)/dij

! Calculate pair forces and energies

  if(ljd%use_core) then
   sigma = sqrt(ljd%sigmasq)
!dir$ vector always
   do i = 1, npair
    dcij = sqrt(rijsq(i)) - ljd%core
    t1 = (sigma/dcij)**6
    t2 = ljd%c2*t1*(0.5d0 - t1)/dcij  !t2 is equal to (dphi/dr)
! Potential energy
    uij(i) = ljd%c1*t1*(t1 - 1.d0) - ljd%ushift
! Force. Assumes that rij = rj - ri.
!dir$ novector
    do j = 1, 3
     fij(i,j) = t2*rij(i,j)/sqrt(rijsq(i))
    enddo
! ES Potential energy
    t1 = dsfd%qiqj*(erij(i)/dij(i) &
       - dsfd%c2 + dsfd%c3*(dij(i) - dsfd%rcut))
    uij(i) = uij(i) + t1
    dsfd%ues = dsfd%ues + t1
! ES Force. Assumes that rij = rj - ri.
    t1 = dsfd%qiqj*(-erij(i)/rijsq(i) &
       - dsfd%c1*eij(i) + dsfd%c3)/dij(i)
!dir$ novector
    do j = 1, 3
     t2 = t1*rij(i,j)
     fij(i,j) = fij(i,j) + t2
     pes = pes - t2*rij(i,j)
    enddo
   enddo

  else

!dir$ vector always
   do i = 1, npair
    t1 = (ljd%sigmasq/rijsq(i))**3
    t2 = ljd%c2*t1*(0.5d0 - t1)/rijsq(i)  !t2 is equal to (dphi/dr)/r
! Potential energy
    uij(i) = ljd%c1*t1*(t1 - 1.d0) - ljd%ushift
! Force. Assumes that rij = rj - ri.
!dir$ novector
    do j = 1, 3
     fij(i,j) = t2*rij(i,j)
    enddo
! ES Potential energy
    t1 = dsfd%qiqj*(erij(i)/dij(i) &
       - dsfd%c2 + dsfd%c3*(dij(i) - dsfd%rcut))
    uij(i) = uij(i) + t1
    dsfd%ues = dsfd%ues + t1
! ES Force. Assumes that rij = rj - ri.
    t1 = dsfd%qiqj*(-erij(i)/rijsq(i) &
       - dsfd%c1*eij(i) + dsfd%c3)/dij(i)
!dir$ novector
    do j = 1, 3
     t2 = t1*rij(i,j)
     fij(i,j) = fij(i,j) + t2
     pes = pes - t2*rij(i,j)
    enddo
   enddo
  endif

 dsfd%pes = pes/3.d0

 end subroutine es_dsf_LJ_force_energy
!***********************************************************************
 subroutine LJ_force_energy(rij,rijsq,ljd,fij,uij)
! subroutine LJ_force_energy(rij,rijsq,ljd,fij,uij,sd2uij,sd2uijrij)
!
! Note: configurational temperature calculation disabled - needs
! to be reinstated, but efficiently
!***********************************************************************
 double precision, dimension(:,:) :: rij, fij
 double precision, dimension(:) :: rijsq, uij
 type(LJ_data_type) :: ljd
 double precision :: t1, t2, sigma, dcij
! double precision, dimension(:) :: sd2uijrij
 integer :: i, j, ndim, npair
 
 ndim = size(rij,2)
 npair = size(rij,1)

! Calculate pair forces and energies

  if(ljd%use_core) then
   sigma = sqrt(ljd%sigmasq)
!dir$ vector always
   do i = 1, npair
    dcij = sqrt(rijsq(i)) - ljd%core
    t1 = (sigma/dcij)**6
    t2 = ljd%c2*t1*(0.5d0 - t1)/dcij  !t2 is equal to (dphi/dr)
! Potential energy
    uij(i) = ljd%c1*t1*(t1 - 1.d0) - ljd%ushift
! Force. Assumes that rij = rj - ri.
!dir$ novector
!dir$ unroll=3
    do j = 1, 3
     fij(i,j) = t2*rij(i,j)/sqrt(rijsq(i))
    enddo
   enddo

  else

   !dir$ vector always
   do i = 1, npair
    t1 = (ljd%sigmasq/rijsq(i))**3
    t2 = ljd%c2*t1*(0.5d0 - t1)/rijsq(i)  !t2 is equal to (dphi/dr)/r
! Potential energy
    uij(i) = ljd%c1*t1*(t1 - 1.d0) - ljd%ushift
! Force. Assumes that rij = rj - ri.
!dir$ novector
!dir$ unroll=3
    do j = 1, 3
     fij(i,j) = t2*rij(i,j)
    enddo
   enddo
  endif

! sum d2phi/dr2. Assumes that rij = rj - ri. Required for Tconf.
!  sd2uij = 0.d0

!  sd2uij = 2.d0*sum(ljd%c2*t1*(13.d0*t1 - 3.5d0)/rijsq + 2.0d0*t2)

!  do i = 1, ndim
!   sd2uijrij(i) = 2.d0*sum(rij(:,i)*rij(:,i)*ljd%c2*t1*(13.d0*t1 - 3.5d0) &
!   & /(rijsq*rijsq) + (1.d0 - rij(:,i)*rij(:,i)/rijsq)*t2)
!   sd2uij = sd2uij + sd2uijrij(i)
!  enddo

  end subroutine LJ_force_energy
!***********************************************************************
 subroutine SHREP_force_energy(rij,rijsq,shrepd,fij,uij)
! Constants for SHREP potential
!***********************************************************************
 double precision, dimension(:,:) :: rij,fij
 double precision, dimension(:) :: rijsq, uij
 double precision, dimension(size(rij,1)) :: t1, t2
 type(SHREP_data_type) :: shrepd
 integer :: i,ndim
 
 ndim = size(rij,2)

! Calculate pair forces and energies

  t1 = (9.0 - 8.0*sqrt(rijsq/shrepd%sigmasq))
  t2 = t1*t1

! Potential energy

  uij = shrepd%c1*t1*t2

! Force. Assumes that rij = rj - ri.

  do i = 1, ndim
   fij(:,i) = -shrepd%c2*t2*rij(:,i)/sqrt(rijsq)
  enddo

 end subroutine SHREP_force_energy
!***********************************************************************
!  subroutine wr_pf(pf,ofile,ounit)
! writes input data to the output file
!***********************************************************************
!  type (pf_data_type), intent(in) :: pf
!  integer :: ounit
!  character(len=32) :: ofile      !output file name

!  open(ounit,file=ofile,status='old',form='formatted',position='append')
!  write (ounit,*) 'Initial pf force values'
!  write (ounit,*)
!  write (ounit,*) 'pair force type           = ', pf%pftype
!  write (ounit,*) 'pf epsilon                = ', pf%epsilon
!  write (ounit,*) 'pf sigma                  = ', pf%sigma
!  write (ounit,*) 'pf rcut                   = ', pf%rcut
!  write (ounit,*) 'pf core                   = ', pf%core
!  write (ounit,*) 'Do shifted potential?     = ', pf%do_shift
!  write (ounit,*) 'Use core potential?       = ', pf%use_core
!  write (ounit,*)
!  close(ounit)
  
!  end subroutine wr_pf
!***********************************************************************
  end module pair_force
