  module properties
!***********************************************************************
! Provides routines that operate on properties, totals and averages.
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************
  use ensemble_av

  implicit none

  save
 
  private

! Public procedures
  public :: accumulate_properties, check_properties_flags, &
          & create_property_arrays, &
          & delete_property_list, destroy_properties, &
          & name_property, rd_properties_input, rd_properties, store_properties, &
          & jqkinetic, jqpotential, &
          & sum_psq_on_m, sum_rij_fij, &
          & wr_property_averages, wr_property_blockaverages, &
          & wr_property_headers, wr_property_tseries, zero_properties, &
          & add_property, eav_blocks

! Public data types
  public :: properties_type, plink

  type :: properties_type
   double precision, dimension(:,:), allocatable :: blocks
   double precision, dimension(:,:), allocatable :: blockavs
   double precision, dimension(:), allocatable :: inst, totals, means, stderrs
   integer, dimension(:), allocatable :: i
   character(len=19), dimension(:), allocatable :: propname
   integer :: nprop    ! nprop = number of properties being sampled
   integer :: maxid    ! maxid = maximum value of property id found
   integer :: nblock   ! nblock = total no of blocks at end of run
   integer :: iblock   ! iblock = index of current block
   integer :: nsamp    ! nsamp = no of samples per block
   integer :: kprop    ! kprop = no of timesteps between samples
   integer :: ktseries ! kprop = no of samples between writing tseries
   integer :: iprop    ! iprop = total no of samples taken
   logical :: accumulate_properties_now, do_tseries_now
   character(len=32) :: tseries_file
  end type properties_type
  
  type:: plink
   character(len=18) :: propname
   integer :: pdim, pid
   type(plink), pointer :: next
  end type plink

  character, parameter :: tab = char(9)

  contains
!***********************************************************************
  subroutine accumulate_properties(p)
!***********************************************************************
  type (properties_type), intent(inout) :: p
  integer :: ib

  ib = p%iprop/p%nsamp + 1
  if(ib .le. p%nblock) then
   p%iprop = p%iprop + 1
   p%iblock = ib
   p%blocks(:,p%iblock) = p%blocks(:,p%iblock) + p%inst
   p%totals(:) = p%totals(:) + p%inst
  endif

  end subroutine accumulate_properties
!***********************************************************************
  subroutine add_property(p,pl,name,pdim,propid,isroot)
! Adds a property name to the linked list of properties.
! Note that the linked list is later deleted after being converted into
! an array for more convenient processing. This can only occur at
! run-time, when the number of required properties is known.
!***********************************************************************
  type (properties_type), intent(inout) :: p
  type (plink), pointer :: pl
  character(len=*) :: name      ! property name
  integer :: pdim, propid, error
  logical, optional :: isroot

  if(present(isroot)) then
   if(isroot) allocate(pl, stat=error)
   if(error .ne. 0) stop 'allocation error in add_property'
  else
   allocate(pl%next, stat=error)
   if(error .ne. 0) stop 'allocation error in add_property'
   pl => pl%next
  endif
  if(propid .gt. p%maxid) p%maxid = propid
  pl%pdim = pdim
  pl%propname = name
  if(len(name) .gt. 16) stop &
  & 'add_property: property name must be no more than 16 characters'
  pl%pid = propid
  p%nprop = p%nprop + pdim
  
  end subroutine add_property
!***********************************************************************
  subroutine check_properties_flags(nstep, p)
!***********************************************************************
  type (properties_type), intent(inout) :: p
  integer :: nstep

  if(p%kprop .ne. 0) p%accumulate_properties_now = mod(nstep,p%kprop) .eq. 0
  if(p%ktseries .ne. 0) p%do_tseries_now = mod(nstep,p%ktseries*p%kprop) .eq. 0

  end subroutine check_properties_flags
!***********************************************************************
  subroutine create_property_arrays(p, plroot, steps_this_run)
!***********************************************************************
  type (properties_type), intent(inout) :: p
  type (plink), pointer :: plroot
  type (plink), pointer :: plcur
  integer :: np, i, j, k, steps_this_run, nb

! Create arrays for properties
  np = p%nprop
! Calculate total number of blocks for this run
  if(mod(steps_this_run,p%kprop*p%nsamp) .ne. 0) &
   stop 'properties.create_property_arrays: &
   stop - run length must be a whole number of property averaging blocks'
  p%nblock = steps_this_run/(p%kprop*p%nsamp)
  nb = p%nblock
  call destroy_properties(p)
  allocate(p%inst(np), p%totals(np), p%blocks(np,nb), p%propname(np), &
  & p%i(p%maxid), stat=i)
  if(i .ne. 0) stop 'properties.create_property_arrays: allocation error'
! traverse the list of properties and transfer the name to the array
! of property names, then allocate an index to the property.
  if(.not. associated(plroot)) stop &
  & 'properties.create_property_arrays: property list is empty'
  plcur => plroot
  k = 1
  p%i = 0
  do i = 1, p%maxid
   p%i(plcur%pid) = k
   do j = 1, plcur%pdim
    if(plcur%pdim .gt. 1) then
     write(p%propname(k),'(a16,a1,i2.2)') trim(plcur%propname), '-', j
    else
     write(p%propname(k),'(a19)') trim(plcur%propname)
    endif
    k = k + 1
   enddo
   if(.not. associated(plcur%next)) exit
   plcur => plcur%next
  enddo

  end subroutine create_property_arrays
!***********************************************************************
  subroutine delete_property_list(plroot)
!***********************************************************************
  type (plink), pointer :: plroot
  type (plink), pointer :: plcur

! delete the entire linked list
  plcur => plroot
  do
   if(associated(plcur)) then
    plroot => plcur
    plcur => plcur%next
    nullify(plroot%next)
    nullify(plroot)
   else
    nullify(plcur)
    exit
   endif
  enddo

  end subroutine delete_property_list
!***********************************************************************
  subroutine destroy_properties(p)
!***********************************************************************
  type (properties_type), intent(inout) :: p

! deallocate arrays for properties
   if(allocated(p%totals)) deallocate(p%totals)
   if(allocated(p%inst)) deallocate(p%inst)
   if(allocated(p%blocks)) deallocate(p%blocks)
   if(allocated(p%propname)) deallocate(p%propname)

  end subroutine destroy_properties
!***********************************************************************
  subroutine eav_blocks(ead,p)
!***********************************************************************
  type (properties_type), intent(inout) :: p
  type(eav_data) :: ead
  integer :: j

  do j = 1, p%nblock
   call create_coarray(ead,p%blocks(:,j))
   sync all
   if(ead%im .eq. 1) then
    call co_mean(ead,p%blocks(:,j))
   endif
   sync all
  enddo

  end subroutine eav_blocks
!***********************************************************************
  subroutine name_property(p,name,n)
! Records property label
!***********************************************************************
  type (properties_type), intent(inout) :: p
  character(len=*) :: name      ! property name
  integer :: n

  if(len(name) .gt. 18) stop &
   & 'name_property: property name must be no more than 18 characters'
   
  write (p%propname(n),*) name
  
  end subroutine name_property
!***********************************************************************
  subroutine rd_properties_input(p,header,ifile,iunit)
!***********************************************************************
! Reads property accumulation variables from input file.
! 
!***********************************************************************
  type (properties_type), intent(inout) :: p
  
  integer :: i, j, iunit
  
  character(len=*) :: header
  character(len=32) :: rdstring
  character(len=32) :: ifile        ! input file name

  rdstring = ''
  i = 0
  open(iunit,file=ifile,status='old',form='formatted', &
     & position='rewind')
  do j = 1, 1000
   read(iunit,*, iostat=i) rdstring
   if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 999) exit
  enddo

  if (rdstring .eq. header) then
   read(iunit,*) p%kprop     ! read number of steps per sample
   read(iunit,*) p%ktseries  ! read number of samples per time series write
   read(iunit,*) p%nsamp     ! read number of samples per block
   close(iunit)
  else
   close(iunit)
    stop 'Properties input data not found - stop'
  endif

! initialize logical variables
  p%accumulate_properties_now = .false.
  p%do_tseries_now = .false.

  end subroutine rd_properties_input
!***********************************************************************
  subroutine rd_properties(p,sunit)
! Reads property block totals and totals from a file, assumed to be
! already opened.
!***********************************************************************
  type (properties_type), intent(inout) :: p
  double precision, dimension(:,:), allocatable :: blocks
  integer :: error, sunit, nb

  read(sunit) p%iprop, nb
  allocate(blocks(p%nprop,nb), stat=error)
  if(error .ne. 0) stop 'allocation error in rd_properties'
! read block totals and overall totals from properties file
  read(sunit) blocks, p%totals

! Create enlarged array for block averages
  p%nblock = nb + p%nblock
  if(allocated(p%blocks)) deallocate(p%blocks)
  allocate(p%blocks(p%nprop,p%nblock), stat=error)
  if(error .ne. 0) stop 'allocation error in rd_properties'
  p%blocks = 0.d0

! Put the existing block totals into the block totals array
  p%blocks(:,1:nb) = blocks
  deallocate(blocks)

  end subroutine rd_properties
!***********************************************************************
  subroutine store_properties(p,sunit)
! Stores property averages (actually totals) in a file, assumed to be
! already opened.
!***********************************************************************
  type (properties_type), intent(inout) :: p
  integer :: sunit

  write(sunit) p%iprop, p%iblock
  write(sunit) p%blocks, p%totals
  
  end subroutine store_properties
!***********************************************************************
  subroutine wr_property_averages(p,tstep,ofile,ounit)
! Writes averages and standard errors to the output file.
!***********************************************************************
  type (properties_type), intent(inout) :: p
  double precision :: tstep
  integer :: i, ounit, error
  character(len=40) :: format_string
  character(len=32) :: ofile      !output file name

  write(format_string,*) '(e18.8,', p%nprop, '(a1,e18.8)', ')'

  open(ounit,file=ofile,status='old',form='formatted',position='append')
  
  allocate(p%means(p%nprop), p%stderrs(p%nprop), stat=error)
  if(error .ne. 0) stop 'allocation error in wr_property_averages'
  p%means = 0.d0
  p%stderrs = 0.d0
  
  p%means = sum(p%blockavs, dim=2)/p%nblock
  if(p%nblock .gt. 1) then
   do i = 1, p%nprop
    p%stderrs(i) = sqrt(sum((p%blockavs(i,:) - p%means(i))**2)/ &
    & (dble(p%nblock)*dble((p%nblock-1))))
   enddo
  endif
   
  write (ounit,format_string) &
 & p%kprop*p%iprop*tstep,(tab,p%means(i), i = 1, p%nprop)
 
  write(format_string,*) '(a18,', p%nprop, '(a1,e18.8)', ')'
  
  if(p%nblock .gt. 1) then
   write (ounit,format_string) &
    & 'Std errors', (tab, p%stderrs(i), i = 1, p%nprop)
  else
   write (ounit,*) 'Not enough samples to calculate std errors'
  endif
  close(ounit)
  
  deallocate(p%means,p%stderrs,p%blockavs)

  end subroutine wr_property_averages
!***********************************************************************
  subroutine wr_property_blockaverages(p,tstep,ofile,ounit)
! Writes block averages to the output file.
!***********************************************************************
  type (properties_type), intent(inout) :: p
  double precision :: tstep
  integer :: ounit, error, i, j
  character(len=40) :: format_string
  character(len=32) :: ofile      !output file name

  write(format_string,*) '(e18.8,', p%nprop, '(a1,e18.8)', ')'
  
  allocate(p%blockavs(p%nprop,p%nblock), stat=error)
  if(error .ne. 0) stop 'allocation error in wr_property_blockaverages'

  p%blockavs = p%blocks/p%nsamp

  open(ounit,file=ofile,status='old',form='formatted',position='append')
  do j = 1, p%nblock
   write (ounit,format_string) &
   & j*p%kprop*p%nsamp*tstep,(tab, p%blockavs(i,j), i = 1, p%nprop)
  enddo
  close(ounit)

  end subroutine wr_property_blockaverages
!***********************************************************************
  subroutine wr_property_headers(p,ofile,ounit)
! Writes headers for properties to a file.
!***********************************************************************
  type (properties_type), intent(inout) :: p
  integer :: ounit, i
  character(len=40) :: format_string
  character(len=32) :: ofile      !output file name
  character(len=18) :: time

  time = 'time'
  open(ounit,file=ofile,status='old',form='formatted',position='append')
  write(format_string,*) '(a18,', p%nprop, '(a1, a18)', ')'
  write (ounit,format_string) time,(tab,adjustl(p%propname(i)), i = 1, p%nprop)
  close(ounit)
  
  end subroutine wr_property_headers
!***********************************************************************
  subroutine wr_property_tseries(p,time,ounit)
!***********************************************************************
  type (properties_type), intent(inout) :: p
  double precision :: time
  integer :: ounit, i
  character(len=40) :: format_string

! Write instantaneous property values to time series output file.
  open(ounit,file=p%tseries_file,status='old',form='formatted',position='append')
  write(format_string,*) '(e18.8,', p%nprop, '(a1,e18.8)', ')'
  write (ounit,format_string) time, (tab,p%inst(i), i = 1, p%nprop)
  close(ounit)

  end subroutine wr_property_tseries
!***********************************************************************
  subroutine zero_properties(p)
! Zeroes property totals, block averages, counters
! and instantaneous values.
!***********************************************************************
  type (properties_type), intent(inout) :: p

  p%totals = 0.d0
  p%blocks = 0.d0
  p%inst = 0.d0
  p%iprop = 0

  end subroutine zero_properties
!***********************************************************************
! Functions to calculate properties
!***********************************************************************
  function jqkinetic(patom, eint, massinv)
! Calculates kinetic part of heat flux vector
!***********************************************************************
  double precision, dimension(:,:) :: patom
  double precision, dimension(size(patom,2)) :: jqkinetic
  double precision, dimension(:) :: eint
  double precision, dimension(:), optional :: massinv
  integer :: i, ndim

  ndim = size(patom,2)
  if(present(massinv)) then
  do i = 1, ndim
   jqkinetic(i) = sum(patom(:,i)*eint*massinv)
  enddo
  else
  do i = 1, ndim
   jqkinetic(i) = sum(patom(:,i)*eint)
  enddo
  endif

  end function jqkinetic
!***********************************************************************
  function jqpotential(rij, fij, pi1, pi2, mass1, mass2)
! Calculates potential part of heat flux vector
!***********************************************************************
  double precision, dimension(:,:) :: rij, fij, pi1, pi2
  double precision, dimension(size(rij,2)) :: jqpotential
  double precision, dimension(:) :: mass1, mass2
  integer :: i, j, k, ndim
  
  ndim = size(rij,2)
  jqpotential = 0.d0

  do i = 1, ndim
   do k = 1, ndim
    do j = 1, size(rij,1)
     jqpotential(i) = jqpotential(i) + &
     & rij(j,i)*fij(j,k)*(pi1(j,k)/mass1(j) + pi2(j,k)/mass2(j))
    enddo
   enddo
  enddo

  jqpotential = 0.5d0*jqpotential

  end function jqpotential
!***********************************************************************
  function jqpotential_2(rij, fij, pi, massinv)
! Calculates potential part of heat flux vector
!***********************************************************************
  double precision, dimension(:,:) :: rij, fij, pi
  double precision, dimension(size(rij,2)) :: jqpotential_2
  double precision, dimension(size(rij,1)) :: t1
  double precision, dimension(:), optional :: massinv
  integer :: i, ndim
  
  ndim = size(rij,2)

  t1 = pi(:,1)*fij(:,1)
  do i = 2, ndim
   t1 = t1 + pi(:,i)*fij(:,i)
  enddo

  if(present(massinv)) t1 = t1*massinv

  do i = 1, ndim
   jqpotential_2(i) = dot_product(rij(:,i),t1)
  enddo

  end function jqpotential_2
!***********************************************************************
  function jqpotential_1(rij, fij, pi, massinv)
! Calculates potential part of heat flux vector
!***********************************************************************
  double precision, dimension(:,:) :: rij, fij, pi
  double precision, dimension(size(rij,2)) :: jqpotential_1
  double precision, dimension(:), optional :: massinv
  integer :: i, ndim

  ndim = size(rij,2)
  if(present(massinv)) then
   do i = 1, ndim
    jqpotential_1(i) = sum(rij(:,i)*sum(pi*fij,2)*massinv)
   enddo
  else
   do i = 1, ndim
    jqpotential_1(i) = sum(rij(:,i)*sum(pi*fij,2))
   enddo
  endif

  end function jqpotential_1
!***********************************************************************
  function sum_psq_on_m_1(patom, massinv)
! Calculates kinetic part of pressure tensor
!***********************************************************************
  double precision, dimension(:,:) :: patom
  double precision, dimension(size(patom,2),size(patom,2)) :: sum_psq_on_m_1
  double precision, dimension(:), optional :: massinv
  integer :: i, j, k, np, ndim

  ndim = size(patom,2)
  np = size(patom,1)
  sum_psq_on_m_1 = 0.d0
  if(present(massinv)) then
  do i = 1, ndim
   do j = 1, ndim
   do k = 1, np
    sum_psq_on_m_1(i,j) = sum_psq_on_m_1(i,j) + patom(k,i)*patom(k,j)*massinv(k)
    enddo
   enddo
  enddo
  else
  do i = 1, ndim
   do j = 1, ndim
   do k = 1, np
    sum_psq_on_m_1(i,j) = sum_psq_on_m_1(i,j) + patom(k,i)*patom(k,j)
    enddo
   enddo
  enddo
  endif

  end function sum_psq_on_m_1
!***********************************************************************
  function sum_psq_on_m(patom, massinv)
! Calculates kinetic part of pressure tensor
!***********************************************************************
  double precision, dimension(:,:) :: patom
  double precision, dimension(size(patom,2),size(patom,2)) :: sum_psq_on_m
  double precision, dimension(:), optional :: massinv
  integer :: i, j, ndim

  ndim = size(patom,2)
  if(present(massinv)) then
  do i = 1, ndim
   do j = 1, ndim
    sum_psq_on_m(i,j) = sum(patom(:,i)*patom(:,j)*massinv)
   enddo
  enddo
  else
  do i = 1, ndim
   do j = 1, ndim
    sum_psq_on_m(i,j) = sum(patom(:,i)*patom(:,j))
   enddo
  enddo
  endif

  end function sum_psq_on_m
!***********************************************************************
  function sum_rij_fij(rij, fij)
! Calculates potential part of pressure tensor
!***********************************************************************
  double precision, dimension(:,:) :: rij, fij
  double precision, dimension(size(rij,2),size(rij,2)) :: sum_rij_fij
  integer :: i, j, k, ndim, npair
  
  npair = size(rij,1)
  ndim = size(rij,2)
  sum_rij_fij = 0.d0
  do i = 1, ndim
   do j = 1, ndim
    do k = 1, npair
!    sum_rij_fij(i,j) = sum(rij(:,i)*fij(:,j))
!    sum_rij_fij(i,j) = dot_product(rij(:,i),fij(:,j))
     sum_rij_fij(i,j) = sum_rij_fij(i,j) + rij(k,i)*fij(k,j)
    enddo
   enddo
  enddo

  end function sum_rij_fij
!***********************************************************************
  end module properties
