!***********************************************************************
   module simulation_box
!***********************************************************************
!  Module for simulation box procedures and types.
!
!  Note: box vectors are stored with first index for vector number and
!  second index for component number. (Three row vectors.)
!
!  Note: boxvec0 is the initial box - used by elongational and bulk
!  flow simulations.
!
!  boxveci is the matrix of inverse lattice vectors, represented as
!  three row vectors, as in Nose and Klein, Mol Phys 50, 1055 (1983).
!
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************

  implicit none

  private

  public :: apply_pbcs, pbc_shift, create_box, destroy_box, &
          & init_box, minimum_image, rd_box_state, &
          & store_box_state, volume, matinv3d

  public :: box_data_type

  type box_data_type
   double precision, dimension(:,:), allocatable :: boxvec, boxvec0, boxveci
   double precision :: volume, angle
   integer :: state  ! current box state
                     ! (0=initial, 1=sheared, 2=rotated and elongated)
  end type box_data_type

  contains
!***********************************************************************
  subroutine apply_pbcs(box, ratom, rsh)
! Apply periodic boundary conditions.
! Calculate shift vector for each position vector if rsh is present.
! This version: uses array syntax. Involves creation of large
! temporary arrays as local variables and function arguments.
!***********************************************************************
  implicit none
  type (box_data_type), intent(in) :: box
  double precision, dimension(:,:), intent(inout) :: ratom
  double precision, dimension(:,:), optional, intent(out) :: rsh
  double precision, dimension(size(ratom,1),3) :: rb, shu, sh
  double precision :: r0
  integer :: i, j

   r0 = 0.5d0

   call rboxv(ratom,box%boxvec,box%volume,rb)

  do i = 1, 3
   rb(:,i) = rb(:,i) - r0
   shu(:,i) = int(2.d0*(rb(:,i)) - int(rb(:,i)))
  enddo

   do i = 1, 3
    do j = 1, size(ratom,1)
     sh(j,i) = shu(j,1)*box%boxvec(1,i) + shu(j,2)*box%boxvec(2,i) + shu(j,3)*box%boxvec(3,i)
     ratom(j,i) = ratom(j,i) - sh(j,i)
    enddo
   enddo

  if(present(rsh)) rsh = sh

  end subroutine apply_pbcs
!***********************************************************************
  function pbc_shift(box, ratom)
! Calculate shift vector for each position vector.
! This version: uses array syntax. Involves creation of large
! temporary arrays as local variables and function arguments.
!***********************************************************************
  implicit none
  type (box_data_type), intent(in) :: box
  double precision, dimension(:,:), intent(in) :: ratom
  double precision, dimension(size(ratom,1),size(ratom,2)) :: pbc_shift
  double precision, dimension(size(ratom,1),3) :: rb, shu
  double precision :: r0
  integer :: i

   r0 = 0.5d0

   call rboxv(ratom,box%boxvec,box%volume,rb)

   rb = rb - r0

  shu = int(2.d0*(rb) - int(rb))

  do i = 1, size(ratom,2)
   pbc_shift(:,i) = shu(:,1)*box%boxvec(1,i) + shu(:,2)*box%boxvec(2,i)
   pbc_shift(:,i) = pbc_shift(:,i) + shu(:,3)*box%boxvec(3,i)
  enddo

  end function pbc_shift
!***********************************************************************
 subroutine check_box(box, rcut)
!***********************************************************************
 implicit none
 type (box_data_type) :: box
 double precision :: rcut, temp
 integer :: i, ndim
 
 ndim = size(box%boxvec, 1)
 
 temp = minval((/(box%boxvec(i,i), i = 1, ndim)/))

 if(2.0*rcut .ge. temp) &
 & stop 'simulation_box.check_box: Box is less than 2*rcut'

 end subroutine check_box
!***********************************************************************
 subroutine create_box(box)
! Allocate box arrays
!***********************************************************************
 implicit none
 type (box_data_type) :: box
 integer :: error

! allocate box vector array
 allocate(box%boxvec(3,3), stat=error)
 if(error .ne. 0) stop 'allocation error in create_box'
 allocate(box%boxvec0(3,3), stat=error)
 if(error .ne. 0) stop 'allocation error in create_box'
 allocate(box%boxveci(3,3), stat=error)
 if(error .ne. 0) stop 'allocation error in create_box'

 end subroutine create_box
!***********************************************************************
 subroutine destroy_box(box)
!***********************************************************************
 implicit none
 type (box_data_type) :: box
 integer :: error

! deallocate box vector arrays
  if(allocated(box%boxvec)) deallocate(box%boxvec, box%boxvec0, &
  & box%boxveci, stat=error)

  end subroutine destroy_box
!***********************************************************************
 subroutine init_box(box, lx, ly, lz)
! Initialize box variables
!***********************************************************************
 implicit none
 type (box_data_type) :: box
 double precision :: lx, ly
 double precision, optional :: lz

! initalize box data
  box%boxvec = 0.d0
  box%angle = 0.d0

  box%boxvec(1,1) = lx
  box%boxvec(2,2) = ly
  if(present(lz)) box%boxvec(3,3) = lz

  box%volume = volume(box%boxvec)
  box%boxveci = matinv3d(transpose(box%boxvec))

  end subroutine init_box
!***********************************************************************
  subroutine minimum_image(box, rij)
! Minimum imaging subroutine for an array of pair separation vectors.
!***********************************************************************
  implicit none
  type (box_data_type) :: box
  double precision, dimension(:,:) :: rij
  double precision, dimension(size(rij,1),3) :: shu
  integer :: i, j, k

!  shu = rboxv(rij,box%boxvec,box%volume)
  call rboxv(rij,box%boxvec,box%volume,shu)

  shu = int(2.d0*shu - int(shu))

     do k = 1, 3
      do j = 1, 3
       if(box%boxvec(j,k) .ne. 0) then
        do i = 1, size(rij,1)
         rij(i,k) = rij(i,k) - shu(i,j)*box%boxvec(j,k)
        enddo
       endif
      enddo
     enddo

  end subroutine minimum_image
!***********************************************************************
  subroutine rd_box_state(box,iunit)
!***********************************************************************
  implicit none
  type (box_data_type), intent(inout) :: box
  integer, intent(in) :: iunit

  read(iunit) box%boxvec, box%boxvec0, &
            & box%state, box%angle

  box%volume = volume(box%boxvec)
  
  end subroutine rd_box_state
!***********************************************************************
  subroutine store_box_state(box,iunit)
!***********************************************************************
  implicit none
  type (box_data_type), intent(in) :: box
  integer, intent(in) :: iunit

  write(iunit) box%boxvec, box%boxvec0, &
             & box%state, box%angle
  
  end subroutine store_box_state
!***********************************************************************
  function volume(boxvec)
! Evaluate triple product of box vectors to find box volume.
!***********************************************************************
  implicit none
  double precision, dimension(:,:) :: boxvec
  double precision :: volume
  integer :: ndim

  ndim = size(boxvec,1)

  select case(ndim)
   case(2)
    volume =  boxvec(1,1)*boxvec(2,2) &
           & -boxvec(1,2)*boxvec(2,1)
   case(3)
    volume =  boxvec(1,1)*(boxvec(2,2)*boxvec(3,3) - &
                           & boxvec(2,3)*boxvec(3,2))  &
           & -boxvec(1,2)*(boxvec(2,1)*boxvec(3,3) - &
                           & boxvec(2,3)*boxvec(3,1))  &
           & +boxvec(1,3)*(boxvec(2,1)*boxvec(3,2) - &
                           & boxvec(2,2)*boxvec(3,1))
   case default
    stop 'volume: invalid value of ndim'
  end select
 
  end function volume
!***********************************************************************
  function det_3by3(matrix)
! Evaluate the determinant of a 3 by 3 matrix.
!***********************************************************************
  implicit none
  double precision, dimension(:,:) :: matrix
  double precision :: det_3by3

  det_3by3 =  matrix(1,1)*(matrix(2,2)*matrix(3,3) - &
                           & matrix(2,3)*matrix(3,2))  &
           & -matrix(1,2)*(matrix(2,1)*matrix(3,3) - &
                           & matrix(2,3)*matrix(3,1))  &
           & +matrix(1,3)*(matrix(2,1)*matrix(3,2) - &
                           & matrix(2,2)*matrix(3,1)) 
  end function det_3by3
!***********************************************************************
  function cross(v1,v2)
! Evaluate the cross product of two 3d vectors.
!***********************************************************************
  implicit none
  double precision, dimension(:) :: v1, v2
  double precision, dimension(size(v1)) :: cross

  cross(1) = v1(2)*v2(3) - v1(3)*v2(2)
  cross(2) = v1(3)*v2(1) - v1(1)*v2(3)
  cross(3) = v1(1)*v2(2) - v1(2)*v2(1)

  end function cross
!***********************************************************************
  function matinv3d(matrix)
! Evaluate the inverse of a 3x3 matrix - essentially Cramer's rule.
!***********************************************************************
  implicit none
  double precision, dimension(3,3) :: matrix, matinv3d
!  double precision, dimension(:,:) :: matrix
!  double precision, dimension(size(matrix,1),size(matrix,1)) :: matinv3d
  double precision :: det
  
  det = det_3by3(matrix)

  matinv3d(1,:) = cross(matrix(2,:), matrix(3,:))/det
  matinv3d(2,:) = cross(matrix(3,:), matrix(1,:))/det
  matinv3d(3,:) = cross(matrix(1,:), matrix(2,:))/det

  end function matinv3d
!***********************************************************************
  subroutine rboxv(ratom,boxvec,vol,rb)
! Evaluate position in terms of box vectors.
!***********************************************************************
  implicit none
  double precision, dimension(:,:), intent(in) :: ratom, boxvec
  double precision, dimension(:,:) :: rb
  double precision, intent(in) :: vol
  double precision :: c1, c2, c3, c4
  integer :: i, ndim, np
  
  np = size(ratom,1)

  c1 = boxvec(2,2)*boxvec(3,3)/vol
  c2 = boxvec(2,1)*boxvec(3,3)/vol
  c3 = boxvec(1,2)*boxvec(3,3)/vol
  c4 = boxvec(1,1)*boxvec(3,3)/vol
  do i = 1, np
   rb(i,1) =  ratom(i,1)*c1 - ratom(i,2)*c2
   rb(i,2) = -ratom(i,1)*c3 + ratom(i,2)*c4
   rb(i,3) = ratom(i,3)/boxvec(3,3)
  enddo

  end subroutine rboxv
!***********************************************************************
  end module simulation_box
