  module sllod
!***********************************************************************
! Sllod algorithm for flow: planar shear, planar elongation and bulk
! deformation.
! Note: strain here is Hencky strain = strain rate x time
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************
  use simulation_box

  implicit none
 
  private

! Public procedures
  public :: add_p_grad_u, add_r_grad_u, evolve_box_vectors, &
          & init_pef, init_sllod, rd_sllod_input, rd_sllod_state, &
          & store_sllod_state, wr_sllod

! Public data types
  public :: sllod_data_type

! Note: strain here is Hencky strain = strain rate x time

  type :: sllod_data_type
   double precision :: strain_rate, strain, strain_per_period, &
                     & strain_shift
   integer :: flow_type
  end type sllod_data_type

! named constants for the different flow types
  integer, parameter, public :: equil = 0   ! equilibrium - no flow
  integer, parameter, public :: shear = 1   ! planar shear flow
  integer, parameter, public :: pef   = 2   ! planar elongational flow
  integer, parameter, public :: bulk  = 3   ! bulk flow

  contains
!***********************************************************************
  subroutine add_p_grad_u(eqm, p, sd)
! Adds p gradu term to pdot equation of motion
!***********************************************************************
  implicit none
  type (sllod_data_type), intent(in) :: sd
  double precision, dimension(:,:), intent(in) :: p
  double precision, dimension(:,:), intent(inout) :: eqm

  select case(sd%flow_type)
   case(shear)
    eqm(:,1) = eqm(:,1) - sd%strain_rate*p(:,2)
   case(pef)
    eqm(:,1) = eqm(:,1) - sd%strain_rate*p(:,1)
    eqm(:,2) = eqm(:,2) + sd%strain_rate*p(:,2)
   case(bulk)
    eqm = eqm - sd%strain_rate*p
   case default
  end select

  end subroutine add_p_grad_u
!***********************************************************************
  subroutine add_r_grad_u(eqm, r, sd)
! Adds r gradu term to rdot equation of motion
!***********************************************************************
  implicit none
  type (sllod_data_type), intent(in) :: sd
  double precision, dimension(:,:), intent(in) :: r
  double precision, dimension(:,:), intent(inout) :: eqm

  select case(sd%flow_type)
   case(shear)
    eqm(:,1) = eqm(:,1) + sd%strain_rate*r(:,2)
   case(pef)
    eqm(:,1) = eqm(:,1) + sd%strain_rate*r(:,1)
    eqm(:,2) = eqm(:,2) - sd%strain_rate*r(:,2)
   case(bulk)
    eqm = eqm + sd%strain_rate*r
   case default
  end select

  end subroutine add_r_grad_u
!***********************************************************************
  subroutine evolve_box_vectors(boxvec, boxvec0, sd, delta)
! Evolves box vectors according to the flow type.
! Note: boxvec0 is the box at the beginning of the strain period.
!***********************************************************************
  implicit none
  type (sllod_data_type) :: sd
  double precision, dimension(:,:), intent(in) :: boxvec0
  double precision, dimension(:,:), intent(inout) :: boxvec
  double precision :: delta, dbx, dby

  select case(sd%flow_type)
   case(shear)
    sd%strain = sd%strain + sd%strain_rate*delta
    sd%strain = sd%strain - sd%strain_per_period* &
              & int(sd%strain/sd%strain_per_period - sd%strain_shift)
    boxvec(2,1) = boxvec0(2,1) + boxvec0(2,2)*sd%strain
   case(pef)
    sd%strain = sd%strain + sd%strain_rate*delta
    sd%strain = sd%strain - sd%strain_per_period* &
              & int(sd%strain/sd%strain_per_period - sd%strain_shift)
    dbx = exp(sd%strain)
    dby = 1.0/dbx
    boxvec(1,1) = boxvec0(1,1)*dbx
    boxvec(2,1) = boxvec0(2,1)*dbx
    boxvec(1,2) = boxvec0(1,2)*dby
    boxvec(2,2) = boxvec0(2,2)*dby
   case(bulk)
    sd%strain = sd%strain + sd%strain_rate*delta
    dbx = exp(sd%strain)
    boxvec(1,1) = boxvec0(1,1)*dbx
    boxvec(2,2) = boxvec0(2,2)*dbx
    boxvec(3,3) = boxvec0(3,3)*dbx
   case default
  end select

  end subroutine evolve_box_vectors
!***********************************************************************
  subroutine init_pef(sd, k, N11, theta)
! Initialize planar elongational flow - use this before the first step
! of a pef simulation.
! Usual values of k and N11 are 3 and 2 respectively, because they give
! the least distorted cell after one strain period.
! See Kraynik and Reinelt.
!***********************************************************************
  implicit none
  type (sllod_data_type) :: sd
  double precision :: lambda_p, theta
  integer, intent(in) :: k, N11
  integer :: N12

! calculate hencky strain period

  lambda_p = (k + (k*k - 4)**0.5)/2
  sd%strain_per_period = log(lambda_p)
      
! calculate angle between simulation box and x-axis

  N12 = -nint((N11*(k - N11) - 1)**0.5)
  theta = atan((N11 - lambda_p)/N12)

  end subroutine init_pef
!***********************************************************************
  subroutine init_sllod(sd, boxvec)
! Store initial box and set strain to zero.
!***********************************************************************
  implicit none
  type (sllod_data_type) :: sd
  double precision, dimension(:,:) :: boxvec

  sd%strain_per_period = boxvec(1,1)/boxvec(2,2)

  end subroutine init_sllod
!***********************************************************************
  subroutine rd_sllod_input(sd,header,ifile,iunit)
!***********************************************************************
! Reads SLLOD algorithm variables from input file.
! 
!***********************************************************************
  implicit none
  type (sllod_data_type), intent(inout) :: sd
  
  integer :: i, j, iunit
  
  character(len=*) :: header
  character(len=32) :: rdstring
  character(len=32) :: ifile        ! input file name

  rdstring = ''
  i = 0
  open(iunit,file=ifile,status='old',form='formatted', &
     & position='rewind')
  do j = 1, 1000
   read(iunit,*, iostat=i) rdstring
   if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 999) exit
  enddo

  if (rdstring .eq. header) then
   read(iunit,*) sd%flow_type
   read(iunit,*) sd%strain_rate
   read(iunit,*) sd%strain_shift
   close(iunit)
  else
   close(iunit)
   stop 'SLLOD input data not found - stop'
  endif

  end subroutine rd_sllod_input
!***********************************************************************
  subroutine rd_sllod_state(sd,sunit)
!***********************************************************************  
  implicit none
  type (sllod_data_type), intent(inout) :: sd
  integer :: sunit

  read(sunit) sd%strain, sd%strain_per_period

  end subroutine rd_sllod_state
!***********************************************************************
  subroutine store_sllod_state(sd,sunit)
!***********************************************************************  
  implicit none
  type (sllod_data_type), intent(in) :: sd
  integer :: sunit

  write(sunit) sd%strain, sd%strain_per_period

  end subroutine store_sllod_state
!***********************************************************************
  subroutine wr_sllod(sd,ofile,ounit)
! writes input data to the output file
!***********************************************************************
  implicit none
  type (sllod_data_type), intent(in) :: sd
  integer :: ounit
  character(len=32) :: ofile      !output file name

  open(ounit,file=ofile,status='old',form='formatted',position='append')
  write (ounit,*) 'Initial sllod algorithm values'
  write (ounit,*)
  write (ounit,*) 'Flow type                 = ', sd%flow_type
  write (ounit,*) 'Strain rate               = ', sd%strain_rate
  write (ounit,*) 'Strain shift              = ', sd%strain_shift
  write (ounit,*)
  close(ounit)
  
  end subroutine wr_sllod
!***********************************************************************
  end module sllod
