  module species
!***********************************************************************
! Module providing functionality for species properties.
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************
  use atoms

  implicit none
 
  private

! Public procedures
  public :: create_species, init_ab_interaction_mask, &
          & rd_species_input

! Public data types
  public :: species_set
  public :: sp_pointer

  type sp_pointer
   integer, dimension(:), allocatable :: aid  ! indices of atoms of this species (nasp)
  end type sp_pointer

  type species_set
   integer :: np               ! number of particles
   integer :: nsp              ! number of species
   integer :: npsp             ! number of pairs of species
   integer, dimension(:), allocatable :: nasp ! number of atoms in this
                                              ! species (nsp)
   integer, dimension(:), allocatable :: spid ! species that this atom
                                              ! belongs to (np)
   double precision, dimension(:), allocatable :: mass  ! atom mass (nsp)
   double precision, dimension(:), allocatable :: charge  ! atom charge (nsp)
   type(sp_pointer), dimension(:), allocatable :: sp  ! species pointer (nsp)
   logical, dimension(:), allocatable :: has_mass, has_charge  ! (nsp)
   character(len=2), dimension(:), allocatable :: label
  end type species_set

  contains
!***********************************************************************
  subroutine create_species(spd)
!***********************************************************************
  implicit none
  type(species_set) :: spd
  integer :: error

  allocate(spd%spid(spd%np), spd%nasp(spd%nsp), spd%mass(spd%nsp), &
  & spd%charge(spd%nsp), spd%sp(spd%nsp), spd%has_mass(spd%nsp), &
  & spd%has_charge(spd%nsp), spd%label(spd%nsp), stat = error)
  if(error .ne. 0) stop 'species:create_species: allocation error'

  end subroutine create_species
!***********************************************************************
  subroutine destroy_species(spd)
!***********************************************************************
  implicit none
  type(species_set) :: spd
  integer :: i, error

  do i = 1, spd%nsp
   deallocate(spd%sp(i)%aid, stat = error)
  enddo

  deallocate(spd%spid,spd%nasp, spd%mass, spd%charge, spd%has_mass, &
  & stat = error)
  if(error .ne. 0) stop 'species:destroy_species: deallocation error'

  end subroutine destroy_species
!***********************************************************************
  subroutine init_ab_interaction_mask(spd, imask, ia, ib)
! Set mask to false for those pairs that do not have an ab interaction.
!***********************************************************************
  implicit none
  type(species_set), intent(in) :: spd
  logical, dimension(0:,0:) :: imask
  integer :: i, j, ia, ib

! Default is for all sites to be unable to interact.
! Note: imask(0,:) and imask(:,0) are already initialized to false
! for cell method.

! Set ab and ba interaction pair masks to true

  do i = 1, spd%np
   do j = 1, spd%np
    if(.not. (spd%spid(i) == ia .and. spd%spid(j) == ib &
    & .or. spd%spid(i) == ib .and. spd%spid(j) == ia)) then
     imask(i,j) = .false.
     imask(j,i) = .false.
    endif
   enddo
  enddo

  end subroutine init_ab_interaction_mask
!***********************************************************************
  subroutine rd_species_input(spd,na,header,ifile,iunit)
!***********************************************************************
  implicit none
  type (species_set) :: spd
  
  integer :: i, j, iunit, np, na, error=0
  
  character(len=*) :: header
  character(len=32) :: rdstring
  character(len=32) :: ifile      !input file name

  rdstring = ''
  i = 0
  open(iunit,file=ifile,status='old',form='formatted', &
     & position='rewind')
  do j = 1, 1000
   read(iunit,*, iostat=i) rdstring
   if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 999) exit
  enddo

  if (rdstring .eq. header) then
   read(iunit,*) spd%np, spd%nsp
   spd%npsp = spd%nsp*(spd%nsp+1)/2  ! number of distinct pairs of species
   call create_species(spd)
   np = 0
   do i = 1, spd%nsp
    read(iunit,*) spd%label(i), spd%nasp(i), spd%has_mass(i), spd%mass(i), spd%has_charge(i), spd%charge(i)
    np = np + spd%nasp(i)
   enddo
   close(iunit)
   if(np .ne. spd%np .or. np .ne. na) stop &
     & 'species: number of atoms is inconsistent with species data or atom data'
   do i = 1, spd%nsp
    allocate(spd%sp(i)%aid(spd%nasp(i)), stat = error)
   enddo
   if(error .ne. 0) stop 'species:rd_species_input: allocation error'
  else
   close(iunit)
   stop 'Species input data not found - stop'
  endif

  end subroutine rd_species_input
!***********************************************************************
  end module species
