!***********************************************************************
   module system
!***********************************************************************
!  Module for thermal system procedures and types.
!  Includes temperature and density initialization and control.
!  Includes system initialization - from crystal.
!
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************
  use thermostat
  use molecules
  use simulation_box
  use lattice
  use barostat

  implicit none

  private

  public :: build_system, rd_system_input, &
            rd_config, rd_system_state, store_system_state

  public :: system_data_type

  type system_data_type
   integer :: coordinate_input  ! source of coordinates
   integer :: nms  ! number of molecules in the system
   character(len=32) :: coord_file
   double precision :: din ! initial subsystem density
   double precision :: dsc ! scaled density to allow for unfilled unit cells
   type(lattice_data_type) :: lat
   type(thermostat_data_type) :: td
   type(barostat_data_type) :: prd
  end type system_data_type

! Values of constants
! coordinate_input: 0 = build lattice using built-in functions
!                   1 = import coordinates from a file
  contains
!***********************************************************************
 subroutine build_system(sys, box, rcm)
!***********************************************************************
 implicit none
 type (system_data_type) :: sys
 type (box_data_type) :: box
 double precision, dimension(:,:) :: rcm
 double precision :: lx, ly, lz

! build a homogeneous, periodic system
! for homogeneous system ONLY, allow incomplete unit cells and scale the unit
! cells to get the correct overall density so that non-magic numbers of particles
! can be used (i.e. incomplete lattice for FCC/BCC etc).
  sys%dsc = sys%din*sys%lat%nct*sys%lat%npc/sys%nms
  sys%lat%lsc = (sys%lat%npc/ &
                 & (sys%dsc*sys%lat%vcell))**(1./3.)
  sys%lat%offset = 0.25*sys%lat%lv*sys%lat%lsc
  call make_lattice(sys%lat%ltype, sys%lat%nuc,sys%lat%nct, &
  & sys%lat%npc,sys%lat%lsc,sys%lat%offset, rcm, sys%lat%do_mix)
  lx = sys%lat%lv(1)*sys%lat%lsc(1)*sys%lat%nuc(1)
  ly = sys%lat%lv(2)*sys%lat%lsc(2)*sys%lat%nuc(2)
  lz = sys%lat%lv(3)*sys%lat%lsc(3)*sys%lat%nuc(3)

  call init_box(box, lx, ly, lz)


 end subroutine build_system
!***********************************************************************
  subroutine rd_system_state(sys,iunit)
!***********************************************************************
  implicit none
  type (system_data_type) :: sys
  integer, intent(in) :: iunit

  read(iunit) sys%td%zeta, sys%prd%edot, sys%prd%volume

  end subroutine rd_system_state
!***********************************************************************
  subroutine rd_config(ratom, box, rsc, coord_file, iunit)
! read input coordinate data
!***********************************************************************
  type (box_data_type) :: box
  double precision :: lx, ly, lz, rsc
  double precision, dimension(:,:) :: ratom
  character(len=*) :: coord_file
  character(len=2) :: label
  integer :: iunit, i, j, np

  open (iunit,file=coord_file)
  read(iunit,*) np
  if(np .ne. size(ratom,1)) stop 'system.rd_config: incorrect number of particles'
  read(iunit,*) lx, ly, lz
  do i = 1, size(ratom,1)
   read(iunit,*) label, (ratom(i,j), j = 1, size(ratom,2))
   ratom(i,:) = ratom(i,:)/rsc
  enddo
  close(iunit)

  call init_box(box, lx/rsc, ly/rsc, lz/rsc)

  end subroutine rd_config
!***********************************************************************
  subroutine store_system_state(sys,iunit)
!***********************************************************************
  implicit none
  type (system_data_type) :: sys
  integer, intent(in) :: iunit

  write(iunit) sys%td%zeta, sys%prd%edot, sys%prd%volume

  end subroutine store_system_state
!***********************************************************************
  subroutine rd_system_input(sys,md,header,ifile,iunit)
!***********************************************************************
! Reads system variables from input file.
! 
!***********************************************************************
  implicit none
  
  type(system_data_type) :: sys
  type(mol_set) :: md
  
  integer :: i, j, iunit
  
  character(len=*) :: header
  character(len=32) :: rdstring
  character(len=32) :: ifile      !input file name

  rdstring = ''
  i = 0
  open(iunit,file=ifile,status='old',form='formatted', &
     & position='rewind')
  do j = 1, 1000
   read(iunit,*, iostat=i) rdstring
   if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 999) exit
  enddo

  if(rdstring .eq. header) then
   read(iunit,*) sys%coordinate_input
    read(iunit,*) sys%nms
    read(iunit,*) sys%din
    call rdinput_thermostat(sys%td, iunit)
    call rdinput_barostat(sys%prd, iunit)
    call rdinput_lattice(sys%lat, iunit)
   close(iunit)
  else
   close(iunit)
   stop 'system.rd_system_input: system input data not found'
  endif

  end subroutine rd_system_input
!***********************************************************************
  end module system
