  module tcf
!***********************************************************************
! Module providing functionality for computation of time correlation
! functions
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************
  use ensemble_av

  implicit none

  private

  public :: tcf_data, accum_tcf, &
            store_tcf, write_tcf, &
            set_tcf, create_tcf, destroy_tcf, rdinput_tcf

  type tcf_data
   double precision, dimension(:), allocatable :: cfn ! correlation funcn
   double precision, dimension(:,:), allocatable :: shr ! shift registers
   double precision, dimension(:), allocatable :: nv    ! new values
   double precision :: cfnorm, cfintfact    ! normalisation factors
   integer :: lcf   ! no of points in correlation function
   integer :: nsr   ! no of shift registers per correlation function
   integer :: nsh   ! no of shifts performed
   integer :: ncf   ! no of correlation function accumulations
   integer :: kcf   ! no of timesteps between tcf points
                    ! kcf = 0 means never do tcf accumulations
   integer :: typ   ! type of run: 1=new, 2=continued
   integer :: kcfm  ! multiplier of kcf for second delay time
                    ! kcfm=0 for only a single tcf
! accum_type = type of cf accumulation:
! 1=forward time autocorrelation function <A(t-T)A(t-T+s)>
! 2=forward time cross-correlation function <A(t-T+s)B(t-T)>
! 3=backward time auto-correlation function <A(t)A(t-s)>
! 4=backward time cross-correlation function <B(t)A(t-s)>
! 5=forward time isotropic auto-correlation function of traceless symmetric tensor
! 6=backward time isotropic cross-correlation function of vector
   integer :: accum_type
   character(len=32) :: fname   ! restart file name for tcf data
   character(len=32) :: tfname   ! text file name for tcf results
  end type tcf_data


! define storage order of symmetric traceless stress tensor elements
  integer, parameter :: xx=1,xy=2,xz=3,yy=4,yz=5

  contains
!***********************************************************************
  subroutine update_shr(c)
!***********************************************************************
  implicit none

  type(tcf_data) :: c
  
! update the shift registers
! note that positive shift moves data to the left
  if(c%accum_type .eq. 3 .or. c%accum_type .eq. 4 &
  .or. c%accum_type .eq. 6 ) then
! Backward time correlation functions - current time at left
   c%nsh = c%nsh + 1
   c%shr = eoshift(c%shr,shift=-1,boundary=c%nv,dim=1)
  else
! Forward time correlation functions - current time at right
   c%nsh = c%nsh + 1
   c%shr = eoshift(c%shr,shift=1,boundary=c%nv,dim=1)
  endif

  end subroutine update_shr
!***********************************************************************
  subroutine accum_tcf(c, cval)
!***********************************************************************
  implicit none

  type(tcf_data) :: c
  double precision, dimension(c%nsr), optional :: cval ! current values
  integer :: i, j
  
  call update_shr(c)

! accumulate correlation function

  if (c%nsh .ge. c%lcf) then
   c%ncf = c%ncf + 1
   select case(c%accum_type)
    case (1) ! 1=forward time autocorrelation function <A(t-T)A(t-T+s)>
    do i = 1, c%nsr
     do j = 1, c%lcf
      c%cfn(j) = c%cfn(j) + c%shr(j,i)*c%shr(1,i)
     enddo
    enddo
    case(2) ! 2=forward time cross-correlation function <A(t-T+s)B(t-T)>
     do j = 1, c%lcf
      c%cfn(j) = c%cfn(j) + c%shr(j,1)*c%shr(1,2)
     enddo
    case(3) ! 3=backward time auto-correlation function <A(t)A(t-s)>
    do i = 1, c%nsr
     do j = 1, c%lcf
      c%cfn(j) = c%cfn(j) + c%shr(j,i)*c%shr(1,i)
     enddo
    enddo
    case(4) ! 4=backward time cross-correlation function <B(t)A(t-s)>
     do j = 1, c%lcf
      c%cfn(j) = c%cfn(j) + c%shr(j,1)*cval(1)
     enddo
    case(5) ! 5=forward time isotropic auto-correlation function of traceless symmetric tensor
     call accum_stress_tensor_tcf(c)
    case(6) ! 6=backward time isotropic cross-correlation function of vector
    do i = 1, c%nsr
     do j = 1, c%lcf
      c%cfn(j) = c%cfn(j) + c%shr(j,i)*cval(i)
     enddo
    enddo
    case default
     stop 'Invalid accum_type in corrrelation function data'
    end select
  endif

  end subroutine accum_tcf
!***********************************************************************
  subroutine accum_stress_tensor_tcf(c)
!***********************************************************************
  implicit none

  type(tcf_data), intent(inout) :: c
  integer :: i

  do i = 1, c%lcf
   c%cfn(i) = c%cfn(i) + c%shr(i,xx)*c%shr(1,xx) &
    &                  + c%shr(i,yy)*c%shr(1,yy) &
    &                  + c%shr(i,xx)*c%shr(1,yy)*0.5 &
    &                  + c%shr(i,yy)*c%shr(1,xx)*0.5 &
    &                  + c%shr(i,xy)*c%shr(1,xy) &
    &                  + c%shr(i,xz)*c%shr(1,xz) &
    &                  + c%shr(i,yz)*c%shr(1,yz)
  enddo

  end subroutine accum_stress_tensor_tcf
!***********************************************************************
  subroutine store_tcf(cfd,sunit)
!***********************************************************************
  implicit none
  
  type(tcf_data) :: cfd
  integer :: sunit

   open(unit=sunit,file=cfd%fname,&
  &    form='unformatted',status='unknown',position='rewind')
   write(sunit) cfd%cfn,cfd%shr,cfd%nsh,cfd%ncf
   close(sunit)

  end subroutine store_tcf
!***********************************************************************
  subroutine write_tcf_old(cfd,tstep,text,wunit)
! Write correlation functions and integrals to text file.
!***********************************************************************
  implicit none
  
  type(tcf_data) :: cfd
  double precision, dimension(:), allocatable :: cfint
  double precision :: tstep
  character(len=*) :: text
  character, parameter :: tab=char(9)
  integer :: wunit, j, error
  
  allocate(cfint(cfd%lcf), stat=error)
  if(error .ne. 0) stop 'tcf:write_tcf_old: allocation error'

! write correlation functions and integrals to a text file

  open(unit=wunit,file=cfd%tfname,form='formatted',status='replace')

  write(wunit,*) 'Correlation functions and integrals after ',&
  & cfd%ncf, ' accumulations, '
  write(wunit,*) text

! trapezoid rule integration of correlation function
  cfint = 0.d0

  do j = 2, cfd%lcf
   cfint(j) = cfint(j-1) + cfd%cfn(j-1) + cfd%cfn(j)
  enddo

  cfint = 0.5d0*cfint

! write correlation function and normalized integral to text file

  write(wunit,*) 'Normalised integral is ', &
            (sum(cfd%cfn)-0.5*(cfd%cfn(1) + cfd%cfn(cfd%lcf))) &
             *cfd%cfintfact*cfd%kcf/(cfd%cfnorm*cfd%ncf)

  do j = 1, cfd%lcf
   write(wunit,'(1pe12.5,2(a1,1pe12.5))') &
        (j-1)*tstep*cfd%kcf, tab, cfd%cfn(j)/(cfd%cfnorm*cfd%ncf), &
             tab, cfint(j)*cfd%cfintfact*cfd%kcf/(cfd%cfnorm*cfd%ncf)
  enddo

  deallocate (cfint)

  close(wunit)

  end subroutine write_tcf_old
!***********************************************************************
  subroutine write_tcf(cfd,ead,tstep,text,wunit)
! Write correlation functions and integrals to a text file.
! Modified to allow Fortran2008 coarray ensemble averaging.
!***********************************************************************
  implicit none
  
  type(tcf_data) :: cfd
  type(eav_data) :: ead
  double precision, dimension(cfd%lcf) :: cfint
  double precision, dimension(:), allocatable :: avcf, avcfint
  double precision, dimension(:), allocatable :: stderrcf, stderrcfint
  double precision :: tstep
  character(len=*) :: text
  character, parameter :: tab=char(9)
  integer :: wunit, j, error
  
! trapezoid rule integration of correlation function
  cfint = 0.d0
  do j = 2, cfd%lcf
   cfint(j) = cfint(j-1) + cfd%cfn(j-1) + cfd%cfn(j)
  enddo
  cfint = 0.5d0*cfint

! calculate ensemble averaged correlation function and integral
  if(ead%im .eq. 1) allocate(avcf(cfd%lcf), avcfint(cfd%lcf), stat=error)
  if(error .ne. 0) stop 'tcf:write_tcf: allocation error'
  if(ead%ni .gt. 1 .and. ead%im .eq. 1) &
   allocate(stderrcf(cfd%lcf), stderrcfint(cfd%lcf), stat=error)
   if(error .ne. 0) stop 'tcf:write_tcf: allocation error'

  if(ead%ni .gt. 1) then
   call create_coarray(ead,cfd%cfn)
   sync all
   if(ead%im .eq. 1) then
    call co_mean(ead,avcf)
    call co_stderr(ead,avcf,stderrcf)
   endif
   sync all
   call create_coarray(ead,cfint)
   sync all
   if(ead%im .eq. 1) then
    call co_mean(ead,avcfint)
    call co_stderr(ead,avcfint,stderrcfint)
    avcf = avcf/(cfd%cfnorm*cfd%ncf)
    avcfint = avcfint*cfd%cfintfact*cfd%kcf/(cfd%cfnorm*cfd%ncf)
    stderrcf = stderrcf/(cfd%cfnorm*cfd%ncf)
    stderrcfint = stderrcfint*cfd%cfintfact*cfd%kcf/(cfd%cfnorm*cfd%ncf)
   endif
   sync all
  else
   avcf = cfd%cfn/(cfd%cfnorm*cfd%ncf)
   avcfint = cfint*cfd%cfintfact*cfd%kcf/(cfd%cfnorm*cfd%ncf)
  endif

! write correlation function and normalized integral to text file
  if(ead%im .eq. 1) then
   open(unit=wunit,file=cfd%tfname,form='formatted',status='replace')
   write(wunit,*) 'Correlation functions and integrals after ', &
    cfd%ncf, ' accumulations, '
   write(wunit,*) text
   if(ead%ni .gt. 1) then
    write(wunit,*) 'Ensemble averaged over ', ead%ni, ' images.'
    write(wunit,*) 'Delay time', tab, 'avcf', tab, 'avcferr', tab, &
     'avcfint', tab, 'avcfinterr'
    do j = 1, cfd%lcf
     write(wunit,'(1pe12.5,4(a1,1pe12.5))') &
          (j-1)*tstep*cfd%kcf, tab, avcf(j), tab, stderrcf(j), &
          tab, avcfint(j), tab, stderrcfint(j)
    enddo
    deallocate (avcf, avcfint, stderrcf, stderrcfint)
   else
    write(wunit,*) 'Delay time', tab, 'avcf', tab, 'avcfint'
    do j = 1, cfd%lcf
     write(wunit,'(1pe12.5,2(a1,1pe12.5))') &
      (j-1)*tstep*cfd%kcf, tab, avcf(j), tab, avcfint(j)
    enddo
    deallocate (avcf, avcfint)
   endif
   close(wunit)
  endif

  end subroutine write_tcf
!***********************************************************************
  subroutine set_tcf(cfd,nv,cfnorm,lcf,nsr,nsh,ncf,kcf,typ,accum_type, &
 & fname,tfname)
!***********************************************************************
! Set values of variables in tdf data type.
! 
!***********************************************************************
  implicit none
  
  type(tcf_data), intent(inout) :: cfd
  
  double precision, dimension(:), optional, intent(in) :: nv ! new values
  double precision, optional, intent(in) :: cfnorm
  integer, optional, intent(in) :: lcf   ! no of points in correlation function
  integer, optional, intent(in) :: nsr   ! no of shift registers per correlation function
  integer, optional, intent(in) :: nsh   ! no of shifts performed
  integer, optional, intent(in) :: ncf   ! no of correlation function accumulations
  integer, optional, intent(in) :: kcf   ! no of timesteps between tcf points
  integer, optional, intent(in) :: typ   ! type of run: 1=new, 2=continued
  integer, optional, intent(in) :: accum_type   ! type of correlation function accumulation
  character(len=*), optional, intent(in) :: fname   ! restart file name for tcf data
  character(len=*), optional, intent(in) :: tfname   ! text file name for tcf results

  if(present(nv)) cfd%nv = nv
  if(present(cfnorm)) cfd%cfnorm = cfnorm
  if(present(lcf)) cfd%lcf = lcf
  if(present(nsr)) cfd%nsr = nsr
  if(present(nsh)) cfd%nsh = nsh
  if(present(ncf)) cfd%ncf = ncf
  if(present(kcf)) cfd%kcf = kcf
  if(present(typ)) cfd%typ = typ
  if(present(accum_type)) cfd%accum_type = accum_type
  if(present(fname)) cfd%fname = fname
  if(present(tfname)) cfd%tfname = tfname

  end subroutine set_tcf
!***********************************************************************
  subroutine get_tcf(cfd,cfnorm,lcf,nsr,nsh,ncf,kcf,typ,accum_type, &
 & fname,tfname)
!***********************************************************************
! Get data from correlation function variable of derived type tcf_data.
! 
!***********************************************************************
  implicit none
  
  type(tcf_data), intent(in) :: cfd

  double precision, optional, intent(out) :: cfnorm
  integer, optional, intent(out) :: lcf   ! no of points in correlation function
  integer, optional, intent(out) :: nsr   ! no of shift registers per correlation function
  integer, optional, intent(out) :: nsh   ! no of shifts performed
  integer, optional, intent(out) :: ncf   ! no of correlation function accumulations
  integer, optional, intent(out) :: kcf   ! no of timesteps between tcf points
  integer, optional, intent(out) :: typ   ! type of run: 1=new, 2=continued
  integer, optional, intent(out) :: accum_type   ! type of cf accumulation
  character(len=*), optional, intent(out) :: fname   ! restart file name for tcf data
  character(len=*), optional, intent(out) :: tfname   ! text file name for tcf results

  if(present(cfnorm)) cfnorm = cfd%cfnorm
  if(present(lcf)) lcf = cfd%lcf
  if(present(nsr)) nsr = cfd%nsr
  if(present(nsh)) nsh = cfd%nsh
  if(present(ncf)) ncf = cfd%ncf
  if(present(kcf)) kcf = cfd%kcf
  if(present(typ)) typ = cfd%typ
  if(present(accum_type)) accum_type = cfd%accum_type
  if(present(fname)) fname = cfd%fname
  if(present(tfname)) tfname = cfd%tfname

  end subroutine get_tcf
!***********************************************************************
  subroutine rdinput_tcf(cfd,input_file,header,iunit)
!***********************************************************************
! Read tcf input parameters from text file.
! 
!***********************************************************************
  implicit none
  
  type(tcf_data) :: cfd
  
  integer :: i, j, iunit
  
  character(len=*) :: input_file, header
  character(len=32) :: rdstring

  rdstring = ''
  i = 0
  open(iunit,file=input_file,status='old',form='formatted',&
  & position='rewind')
  do j = 1, 1000
   read(iunit,*, iostat=i) rdstring
   if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 999) exit
  enddo

  if (rdstring .eq. header) then
   read(iunit,*) cfd%lcf,cfd%kcf,cfd%nsr,cfd%kcfm
   close(iunit)
  else
   close(iunit)
    stop 'tcf input data not found - stop'
  endif

  end subroutine rdinput_tcf
!***********************************************************************
  subroutine create_tcf(cfd,error_file,eunit,sunit)
!***********************************************************************
! If beginning a new correlation function, reads tcf parameters from
! input file.
! If continuing a stored correlation function, reads the stored data.
! 
!***********************************************************************
  implicit none
  
  type(tcf_data) :: cfd
  
  integer :: error, eunit, sunit
  
  character(len=*) :: error_file

  allocate(cfd%cfn(cfd%lcf),stat=error)
  if(error .ne. 0) stop 'tcf:create_tcf: allocation error for cfd%cfn'
  allocate(cfd%shr(cfd%lcf,cfd%nsr),stat=error)
  if(error .ne. 0) stop 'tcf:create_tcf: allocation error for cfd%shr'
  allocate(cfd%nv(cfd%nsr),stat=error)
  if(error .ne. 0) stop 'tcf:create_tcf: allocation error for cfd%nv'

  cfd%nv  = 0.d0

  select case (cfd%typ)
  
  case (1)                        ! case 1 = zero values and restart
   cfd%cfn = 0.d0
   cfd%shr = 0
   cfd%ncf = 0
   cfd%nsh = 0

  case (2)                       ! case 2 = read data from file
   open(unit=sunit,file=cfd%fname,&
  &    form='unformatted',status='unknown',position='rewind')
   read(sunit) cfd%cfn,cfd%shr,cfd%nsh,cfd%ncf
   close(sunit)

  case default                    ! unknown case - write error and stop
   open(eunit,file=error_file,status='unknown',form='formatted',&
  & position='append')
    write(eunit,*) 'invalid value of run type in tcf - stop'
   stop

  end select

  end subroutine create_tcf
!***********************************************************************
  subroutine destroy_tcf(cfd,error_file,eunit)
!***********************************************************************
! Deallocate dynamic memory variables.
! 
!***********************************************************************
  implicit none
  
  type(tcf_data) :: cfd
  
  integer :: i, eunit
  
  character(len=*) :: error_file

  deallocate(cfd%cfn,stat=i)
  deallocate(cfd%shr,stat=i)
  deallocate(cfd%nv,stat=i)

  if(i .ne. 0) then
   open(eunit,file=error_file,status='unknown',form='formatted',&
  & position='append')
    write(eunit,*) 'Deallocation of tcf data unsuccessful - stop'
   stop
   endif

  end subroutine destroy_tcf
!***********************************************************************
  end module tcf
!***********************************************************************
