!***********************************************************************
  module thermostat
!***********************************************************************
! Temperature calculation and control functions.
!
! Values of constants:
! thermotype: 0 no thermostat
!             1 Gaussian isokinetic on total molecular cm peculiar kinetic energy
!             2 Nose-Hoover isokinetic on total molecular cm kinetic energy
!
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************
  use ensemble_av

  implicit none

  private

! Public procedures
  public :: kinetic_temperature, gauss_therm_mult, &
          & initialize_momenta, rdinput_thermostat, rescale_momenta

  public :: thermostat_data_type

  type thermostat_data_type
   integer :: thermotype  ! thermostat type
   integer :: ktsc  ! momentum rescaling to reset temperature
   integer :: ndfa  ! number of atomic degrees of freedom for this subsystem
   integer :: ndfm  ! number of molecular degrees of freedom for this subsystem
   integer :: ndfad  ! number of atomic degrees of freedom for
                     ! directional thermostat
   integer :: ndfmd  ! number of molecular degrees of freedom for
                     ! directional thermostat
   double precision :: tin ! input temperature
   double precision :: tfb ! proportional feedback for temperature
   double precision :: tnh ! Nose-Hoover relaxation time
   double precision :: zeta ! Nose-Hoover zeta - thermostat multiplier
  end type thermostat_data_type

  contains
!***********************************************************************
  function gauss_therm_mult(p, f, flowtype, strain_rate, massinv)
! Calculates Gaussian isokinetic thermostat multiplier
!***********************************************************************
  implicit none
  
  double precision, dimension(:,:), intent(in) :: p, f
  double precision, dimension(:), optional, intent(in) :: massinv
  double precision, dimension(size(p,1)) :: massi
  double precision :: gauss_therm_mult, anum, sumpsq
  double precision, optional, intent(in) :: strain_rate
  integer :: i, j, flowtype

  anum = 0.d0
  sumpsq = 0.d0

  if(present(massinv)) then
   massi = massinv
  else
   massi = 1.0
  endif

  do j = 1, size(p,2)
   do i = 1, size(p,1)
    anum = anum + p(i,j)*f(i,j)*massi(i)
    sumpsq = sumpsq + p(i,j)*p(i,j)*massi(i)
   enddo
  enddo

  select case (flowtype)
   case(0)  ! no external field (equilibrium)
    gauss_therm_mult = anum/sumpsq
   case(1)  ! homogeneous planar shear
    gauss_therm_mult = (anum - strain_rate*sum(p(:,1)*p(:,2)*massi))/sumpsq
   case(2)  ! homogeneous planar elongation
    gauss_therm_mult = (anum - strain_rate* &
    & sum((p(:,1)*p(:,1)-p(:,2)*p(:,2))*massi))/sumpsq
   case(3)  ! homogeneous bulk flow
    gauss_therm_mult = anum/sumpsq - strain_rate
   case default
    stop 'gauss_therm_mult: invalid value of flowtype'
  end select

  end function gauss_therm_mult
!***********************************************************************
  subroutine initialize_momenta(patom, input_temp, ndf, massinv, ead)
! Initializes a set of momenta with equal speeds determined by the
! kinetic temperature and random directions. The sum of the momenta is
! then set to zero.
!***********************************************************************
  implicit none

  type(eav_data) :: ead  
  double precision, dimension(:,:), intent(out) :: patom
  double precision, intent(in) :: input_temp
  double precision, dimension(size(patom,1),size(patom,2)) :: prand
  double precision, dimension(:), intent(in), optional :: massinv
  integer :: ndim, np, ndf, i

  np = size(patom,1)
  ndim = size(patom,2)

! initialize random number generator
  call init_seed(set_frac(ead))

! generate random numbers using intrinsic
  call random_number(prand)

! convert to numbers between -0.5 and +0.5
  prand = prand - 0.5d0

! turn the momenta into unit vectors (random directions)
  prand = prand/spread(sqrt(sum(prand**2,2)),2,ndim)

! set total momentum to zero in each direction
  do i = 1, ndim
   prand(:,i) = prand(:,i) - sum(prand(:,i))/np
  enddo

! rescale momenta to give the correct temperature

  if(present(massinv)) then
   call rescale_momenta(prand, input_temp, ndf, massinv)
  else
   call rescale_momenta(prand, input_temp, ndf)
  endif

  patom = prand

  end subroutine initialize_momenta
!***********************************************************************
   subroutine init_seed(frac)
! Subroutine to initialize the seed of the random number generator.
! The value of frac can be used to give different values of the seed,
! for example for different images of a coarray fortran program.
!***********************************************************************
   implicit none
   real :: frac
   integer, dimension(8) :: time
   integer :: i, j, iseed

! use system date and time intrinsic subroutine to set the seed
   call date_and_time(values=time)
! use the scaled seconds and milliseconds fields of the system
! date and scaled value of iseed to 
   i = frac*(real(1000*time(7) + time(8))/real(59999))*huge(iseed)
! find size of seed array using random_seed
   call random_seed(size=j)
! set seed of random number generator
   call random_seed(put=spread(i,1,j))

   end subroutine init_seed
!***********************************************************************
  function kinetic_temperature(p, ndf, massinv)
! Calculates kinetic temperature
!***********************************************************************
  implicit none
  
  double precision, dimension(:,:) :: p
  double precision, dimension(:), optional :: massinv
  double precision :: kinetic_temperature
  integer :: ndf

  if(present(massinv)) then
    kinetic_temperature = sum(sum(p*p,2)*massinv)/ndf
  else
    kinetic_temperature = sum(p*p)/ndf
  endif

  end function kinetic_temperature
!***********************************************************************
  subroutine rdinput_thermostat(td,iunit)
!***********************************************************************
! Reads thermostat variables from input file.
! 
!***********************************************************************
  implicit none
  
  type (thermostat_data_type) :: td
  
  integer :: iunit
  
  read(iunit,*) td%tin
  read(iunit,*) td%thermotype
  read(iunit,*) td%tfb
  read(iunit,*) td%ktsc
  read(iunit,*) td%tnh

  end subroutine rdinput_thermostat
!***********************************************************************
  subroutine rescale_momenta(patom, input_temp, ndf, massinv, sf)
! Rescales momenta to give the required input temperature.
!***********************************************************************
  implicit none
  
  double precision, dimension(:,:) :: patom
  double precision, intent(in) :: input_temp
  double precision, dimension(:), intent(in), optional :: massinv
  double precision :: tactual, s
  double precision, optional :: sf
  integer :: ndf

  if(present(massinv)) then
   tactual = kinetic_temperature(patom, ndf, massinv)
  else
   tactual = kinetic_temperature(patom, ndf)
  endif

  s = sqrt(input_temp/tactual)

  patom = patom*s
  if(present(sf)) sf = s

  end subroutine rescale_momenta
!***********************************************************************
  end module thermostat
